﻿EnableExplicit

;- -- Information.
; Dieses Modul sollte in das Projekt mit >XIncludeFile "..\..\_ PureBasic\UsbTmc-v01.pbi"< eingebunden werden.
;
; Enthält eine Pseudoklasse für die Kommunikation mit einem USB Test & Measurement Class Gerät mit Hilfe des LibUsb-Moduls.
;
; Viele Kommentare stammen aus dem Buch 'USB in der Messtechnik' von Henry Bruhns, 2008 Franzis Verlag.

;- -- Literaturliste zu den Referenzen in den Kommentaren im Quellcode:
; Bruhns      - 'USB in der Messtechnik' von Henry Bruhns,
;               2008 Franzis Verlag GmbH, ISBN 978-3-7723-5509-7
; UsbComplete - USB Complete - Fourth Edition 2009 - Jan Axelson
; Usb20       - Universal Serial Bus Specification Revision 2.0, 2000
; UsbTmc      - Universal Serial Bus Test And Measurement Class Specification
;               (USBTMC) Revision 1.0, 2003
; Usb488      - Universal Serial Bus Test And Measurement Class,
;               Subclass USB488 Specification (USBTMC-USB488) Revision 1.0, 2003
; IEEE4881    - IEEE Standard Digital Interface For Programmable Instrumentation, 1988
; IEEE4882    - IEEE Standard Codes, Formats, Protocols And Common Commands, 1988
; SCPI        - Standard Commands For Programmable Instruments (SCPI), Version 1999.0

;- -- Version 01.
; Status:   In Arbeit, aber was vorhanden ist funktioniert, ansonst siehe die 'Issues'.
; Autor:    Peter Dreisiebner
; E-Mail:   web@dreisiebner.at
; Website:  http://peter.dreisiebner.at
; Geändert: 2017-10-25 - PureBasic v5.61 (x86).
; Benötigt: Common, Macros, DebugView, LibUsb.

;- -- Verwendung.
;{
;  1. GetNewInstance()     - Zuerst muss eine neue Instanz erstellt werden.
;  2. GetTmcDeviceList()   - Geräte dürfen nur aus der abgefragten Liste verwendet werden.
;                            Ausnahme wenn der LibUsb Context der Instanzen identisch ist.
;  3. SetTmcDevice()       - Das ausgewählte Gerät von der abgefragten Liste der Instanz zuweisen.
;  4. GetCapabilities()    - Wenn benötigt, die zusätzlichen Eigenschaften und Fähigkeiten des Geräts abfragen.
;                            Die Capabilities werden automatisch bei SetTmcDevice() in der Instanz gespeichert.
;  5. SetTimeout()         - Die Standardwerte für die Kommunikation ändern:
;     SetTerminationChar()             Standard ist 5 Sekunden Zeitüberschreitung. Als Terminierungszeichen wird $0A verwendet,
;     SetTransferSplitDataLength()     und nicht im MessageHeader mitgesendet (Paramaeter fRequestEnabled = #False).
;                                      Die Länge der Daten für die Transfer Message ist Null, die Daten werden damit in einem
;                                      einzelnen Transfer gesendet.
;  6. PrepareData()        - Die Daten vorbereiten, für das Senden an das Gerät.
;                            Die Daten bleiben erhalten und können mehrmals gesendet werden.
;  7. SendData()           - Die vorbereiteten Daten an das Gerät senden.
;  8. WaitData()           - Warten auf die Daten vom Gerät wenn ein Service Request angefordert wurde, oder das Status Byte
;                            überprüft werden soll.
;  9. RequestData()        - Fordert die Daten an, wenn nach dem Senden eine Antwort vom Gerät erwartet wird.
;                            Das Terminierungszeichen wird von den zurückgegebenen Daten automatisch entfernt.
; 10. DeleteInstance()     - Die Instanz löschen. Die Ressourcen und die LibUsb werden automatisch aufgeräumt.
;
; Die Kommunikation mit dem USB-Gerät, kann auch in einem Thread durchgeführt werden.
; Damit wird das Programm nicht blockiert während der Kommunikation, und es können Ereignisse für den Fortschritt der
; gesendeten und empfangenen Daten ausgelöst werden.
;
; 11. SetRequestProgressDataLength()   - Die Anzahl der zu empfangenden Bytes setzen, wann ein neues Ereignis gesendet werden soll.
;     SetRequestBufferLength()         - Die Größe des Empfangsbuffers in Bytes. Wenn Geräte Daten in einem Transfer senden,
;                                        werden keine Ereignisse für den Fortschritt gesendet. Ist der Buffer kleiner als die
;                                        Daten, sendet das Gerät in mehreren Transfers und Ereignisse können gesendet werden.
; 12. SendRequestThreadData() - Die erforderlichen Daten per Struktur uThreadData übergeben.
;                               Mit BindEvent das Ereignis #eCustomEventUsbTmcThreadData mit der Instanznummer auswerten.
; 13. 'Ereignisprozedur'()    - In der eigenen Ereignisprozedur die stattgefundenen Ereignisse auswerten.
;                               Mit PostEvent() wird die Instanznummer, das Ereignis und die verarbeiteten Bytes gesendet.
;                               EventGadget() ist die Instanz, EventType() ist das Ereignis, EventData() sind die Bytes.
; 14. GetThreadData()         - Bei den Ereignissen Send-, Request- oder ThreadFinished, können die Daten verarbeitet werden.
; 15. SetThreadStatus()       - Für das Abbrechen der Kommunikation, oder Beenden des Threads.
;
; Die zurückgegebenen Daten bzw. der reservierte Speicher dafür, von RequestData() und SendRequestThreadData(), müssen wieder
; freigegeben werden.
;
; -.  InitiateClearDevice()   - Löscht alle vorhandenen Daten in den Bulk-Out und Bulk-In Endpoints im Gerät.
;
; Ein USBTMC-USB488-kompatibles Gerät muss die folgenden Device Request innerhalb 500 ms ausführen.
;  -. SetTimeoutControlTransfer()   - Den Standardwert von 500 ms für die Kommunikation ändern.
;
; USBTMC Device Requests.
; Die erste Ergänzung der Standard Device Requests sind die geräteabhängigen Requests,
; die gemäß dem USBTMC-Standard vorgeschrieben sind. Sie gehören in die Gruppe der Class specific Requests.
;  -. InitiateAbortBulkOut()
;  -. InitiateAbortBulkIn()
;  -. CheckAbortBulkOutStatus()
;  -. CheckAbortBulkInStatus()
;  -. InitiateClear()
;  -. CheckClearStatus()
;  -. IndicatorPulse()
;
; USB488 Subclass Device Requests.
; Diese Unterklasse ist wie die USBTMC Device Requests in die Gruppe der Class specific Requests eingeordnet.
; Diese Requests beziehen sich auf Gerätefunktionen, die sich am Standard IEEE 488.1 orientieren.
; Sie beeinflussen direkt die Schnittstellenfunktion des Test- oder Messgeräts.
;  -. InterruptTransfer()
;  -. ReadStatusByte()
;  -. RemoteEnableControl()
;  -. LocalLockout()
;  -. GoToLocal()
;
;  -. SetTimeoutInterruptTransfer() - Den Standardwert von 5000 ms für die Kommunikation ändern.
;
;}


;-
;- -- DeclareModule UsbTmc. --------------------
;
DeclareModule UsbTmc
   EnableExplicit
   
   ;-
   ;- -- Enumerationen.

   ;- eCustomEvent.
   ; Ereignisse die per PostEvent gesendet werden.
   ; Im Modul Common ist die erste Enumeration definiert, mit Zählbeginn #PB_Event_FirstCustomValue.
   Enumeration eCustomEvent
      #eCustomEventUsbTmcThreadData ; Benachrichtigung wenn Daten zu oder von einem USBTMC-Gerät übertragen wurden.
   EndEnumeration
   
   ;- eThreadEvent.
   ; Ereignisse die per PostEvent für den Fortschritt des Threads gesendet werden.
   Enumeration eThreadEvent
      #eThreadEventMinimum
      #eThreadEventSendProgress
      #eThreadEventSendFinished
      #eThreadEventRequestProgress
      #eThreadEventRequestFinished
      #eThreadEventThreadFinished
      #eThreadEventMaximum
   EndEnumeration
   
   ;- eThreadStatus.
   ; Dient zur Steuerung des Threads. Die Werte werden in der Struktur uInstance im Feld eThreadStatus verwendet.
   Enumeration eThreadStatus
      #eThreadStatusMinimum
      #eThreadStatusNone
      #eThreadStatusRun
      #eThreadStatusCancel
      #eThreadStatusEnd
      #eThreadStatusMaximum
   EndEnumeration

   ;- eMessageId.
   ; Nachrichten-Identifizierer (Message Identifier).
   ; Die folgenden Nachrichtentypen sind für die Klasse USBTMC und die Unterklasse USB488 definiert.
   Enumeration eMessageId
      ; Vom Host an das Gerät.
      #eMessageIdDevDepMsgOut = $01             ; Geräteabhängige USBTMC-Befehlsnachricht.
      #eMessageIdRequestDevDepMsgIn = $02       ; Anforderung einer geräteabhängigen USBTMC Antwort über den Bulk-In-Endpoint.
      #eMessageIdVendorSpecificOut = $7E        ; Anbieterspezifische USBTMC-Befehlsnachricht.
      #eMessageIdRequestVendorSpecificIn = $7F  ; Anforderung einer anbieterspezifischen USBTMC Antwort über den Bulk-In-Endpoint.
      #eMessageIdTrigger = $80                  ; Entspricht dem IEEE 488 GET.
      ; Antwort vom Gerät an den Host.
      #eMessageIdDevDepMsgIn = $02              ; Antwort auf REQUEST_DEV_DEP_MSG_IN.
      #eMessageIdVendorSpecificIn = $7F         ; Antwort auf REQUEST_VENDOR_SPECIFIC_IN.
   EndEnumeration

   ;- eUsbTmcStatus.
   Enumeration eUsbTmcStatus
      #eUsbTmcStatusSuccess = $01
      #eUsbTmcStatusPending = $02
      #eUsbTmcStatusFailed = $80
      #eUsbTmcStatusTransferNotInProgress = $81
      #eUsbTmcStatusSplitNotInProgress = $82
      #eUsbTmcStatusSplitInProgress = $83
      ; USB488.
      #eUsbTmcStatusInterruptInBusy = $20
   EndEnumeration
   
   ;- eUsbTmcStatusByte.
   Enumeration eUsbTmcStatusByte
      #eUsbTmcStatusByteDIO1 = %00000001  ; Diese Bits sind zur freien Verfügung des Geräteherstellers, sie sollen jedoch einer
      #eUsbTmcStatusByteDIO2 = %00000010  ; der beiden Datenstrukturen entsprechen, die in IEEE 488.2 in den Abschnitten
      #eUsbTmcStatusByteDIO3 = %00000100  ; 11.4.2 und 11.4.3 beschrieben sind [IEEE-488.2: 11.2.1.3].
      #eUsbTmcStatusByteDIO4 = %00001000  ;
      #eUsbTmcStatusByteDIO8 = %10000000  ; 
      #eUsbTmcStatusByteMAV  = %00010000  ; Message Available.
      #eUsbTmcStatusByteMSS  = %01000000  ; Master Summary Status.
      #eUsbTmcStatusByteRQS  = %01000000  ; Request Service. MSS und RQS teilen sich dasselbe Bit.
      #eUsbTmcStatusByteESB  = %00100000  ; Standard Event Status Bit.
   EndEnumeration

   ;-
   ;- -- Strukturen.

   ;- uThreadData.
   ; Für das Vorbereiten, Senden und Empfangen der Daten in einem Thread.
   ; Diese Struktur wird der Prozedur SendRequestThreadData() übergeben.
   Structure uThreadData
      iUsbTmcInstance.i             ; Die Nummer der Instanz.
      iEventListenerWindowNumber.i  ; Die Fensternummer für die Benachrichtigung per PostEvent().
      fSendData.i                   ; True wenn Daten gesendet werden sollen.
      sScpi.s                       ; Das SCPI-Kommando.
      *SendData                     ; Die binären Daten die an das SCPI-Kommando angehängt werden.
      iSendDataLength.i             ; Die Länge der binären Daten.
      eSendMessageId.i              ; Die Message-ID von der Enumeration eMessageId für das Senden.
      fRequestData.i                ; True wenn Daten empfangen werden sollen.
      *RequestData                  ; Der Speicherbereich der empfangenen Daten.
      eRequestMessageId.i           ; Die Message-ID von der Enumeration eMessageId für das Empfangen.
      fWaitDataServiceRequest.i     ; True wenn auf einen Service Request Interrupt gewartet werden soll, bevor die Daten
                                    ; empfangen werden.
      fWaitDataStatusByte.i         ; True wenn das MAV-Bit im Status Byte geprüft werden soll, bevor die Daten empfangen werden.
                                    ; Es dürfen nicht beide fWaitData-Flags gleichzeitig gesetzt werden.
      iSendTransferred.i            ; Die Anzahl der gesendeten Bytes.
      iRequestTransferred.i         ; Die Anzahl der empfangenen Bytes.
      fSendFinished.i               ; True wenn die Daten gesendet wurden.
      fSendOk.i                     ; True wenn die Daten erfolgreich gesendet wurden, ansonst False.
      fRequestFinished.i            ; True wenn die Daten empfangen wurden.
      fRequestOk.i                  ; True wenn die Daten erfolgreich empfangen wurden, ansonst False,
      iSendProgessBytes.i           ; Die Anzahl der Bytes die aktuell gesendet wurden.
      iRequestProgressBytes.i       ; Die Anzahl der Bytes die aktuell empfangen wurden.
      sError.s                      ; Die Fehlerbeschreibung bei erfolglosen Senden oder Empfangen.
   EndStructure

   ;- uDevice.
   ; Für die Abfrage der verfügbaren USBTMC-Geräte.
   Structure uDevice
      iVendorId.i                      ; Vendor-ID.
      iProductId.i                     ; Produkt-ID.
      sManufacturer.s                  ; Herstellername.
      sProduct.s                       ; Produktname.
      sSerialnumber.s                  ; Seriennummer.
      iBusNumber.i                     ; Die Busnummer an dem das Gerät angeschlossen ist.
      iPortNumber.i                    ; Die Portnummer an dem das Gerät angeschlossen ist.
      iInterfaceNumber.i               ; Die Nummer des Interfaces, ist für USBTMC-Geräte immer 0.
      iEndpointInterruptIn.i           ; Die Adresse der Interrupt-In Endpoints.
      iEndpointBulkOut.i               ; Die Adresse der Bulk-Out Endpoints.
      iEndpointBulkOutMaxPacketSize.i  ; Die Größe des Endpoints.
      iEndpointBulkIn.i                ; Die Adresse der Bulk-In Endpoints.
      iEndpointBulkInMaxPacketSize.i   ; Die Größe des Endpoints.
      ; Für die Suche und Verwaltung des Geräts.
      fUsbTmc.i                        ; Bei True ist das Gerät ein USBTMC-Gerät.
      *uContext                        ; Zeiger auf den LibUsb-Context.
      *FreeDeviceList                  ; Zeiger auf die LibUsb-Geräteliste.
      *uDevice                         ; Zeiger zur Gerätestruktur.
      *uDeviceHandle                   ; Das Handle zum geöffneten Gerät bzw. Zeiger zur DeviceHandle-Struktur.
      iManufacturer.i                  ; Index für den String-Descriptor
      iProduct.i                       ; Index für den String-Descriptor.
      iSerialnumber.i                  ; Index für den String-Descriptor.
      iConfigurationCount.i            ; Die Anzahl der vorhandenen Konfigurationen.
      iConfigurationValue.i            ; Die Nummer der verwendeten Konfiguration, nicht der Index!.
   EndStructure

   ;- uCapabilities.
   ; Für die Abfrage der zusätzlichen Attribute und Fähigkeiten des USBTMC-USB488-Geräts.
   Structure uCapabilities
      ; USBTMC-Gerät.
      iTmcBcdVersionNumber.i
      fTmcInterfaceAcceptsIndicatorPulseRequest.i
      fTmcInterfaceIsTalkOnly.i
      fTmcInterfaceIsListenOnly.i
      fTmcDeviceSupportsEndingTermChar.i
      ; USB488-Interface.
      i488BcdVersionNumber.i
      f488InterfaceIs4882.i
      f488InterfaceAcceptsRemoteLocalControlRequests.i
      f488InterfaceAcceptsTriggerMessage.i
      f488DeviceUnderstandsAllMandatoryScpiCommands.i
      f488DeviceIsServiceRequestCapable.i
      f488DeviceIsRemoteLocalCapable.i
      f488DeviceIsTriggerCapable.i
   EndStructure
   
   ;-
   #iUsbTmcStructureAlignment = 1

   ;- uCheckAbortBulkOutStatusResponse.
   ; Für den USBTMC Device Requests 'Check Abort Bulk Out Status'.
   ; USBTMC Specification Rev1.0 - PDF Page 29, Table 22 -- CHECK_ABORT_BULK_OUT_STATUS response format.
   Structure uCheckAbortBulkOutStatusResponse Align #iUsbTmcStructureAlignment
      eUsbTmcStatus.a
      iReserved.a[3]
      iNBytesRxd.l
   EndStructure

   ;- uCheckAbortBulkInStatusResponse.
   ; Für den USBTMC Device Requests 'Check Abort Bulk In Status'.
   ; USBTMC Specification Rev1.0 - PDF Page 31, Table 28 -- CHECK_ABORT_BULK_IN_STATUS response format.
   Structure uCheckAbortBulkInStatusResponse Align #iUsbTmcStructureAlignment
      eUsbTmcStatus.a
      iAbortBulkInBitmap.a
      iReserved.a[2]
      iNBytesRxd.l
   EndStructure

   ;- uInitiateAbortBulkOutInResponse.
   ; Für den USBTMC Device Requests 'Initiate Abort Bulk Out/In'.
   ; USBTMC Specification Rev1.0 - PDF Page 30, Table 25 -- INITIATE_ABORT_BULK_IN response format.
   Structure uInitiateAbortBulkOutInResponse Align #iUsbTmcStructureAlignment
      eUsbTmcStatus.a
      iTag.a
   EndStructure

   ;- uInitiateClearResponse.
   ; Für den USBTMC Device Requests 'Initiate Clear'.
   ; USBTMC Specification Rev1.0 - PDF Page 33, Table 31 -- INITIATE_CLEAR response format.
   Structure uInitiateClearResponse Align #iUsbTmcStructureAlignment
      eUsbTmcStatus.a
   EndStructure

   ;- uCheckClearStatusResponse.
   ; Für den USBTMC Device Requests 'Check Clear Status'.
   ; USBTMC Specification Rev1.0 - PDF Page 34, Table 34 -- CHECK_CLEAR_STATUS response format.
   Structure uCheckClearStatusResponse Align #iUsbTmcStructureAlignment
      eUsbTmcStatus.a
      iClearBitmap.a
   EndStructure

   ;- uReadStatusByteResponse.
   ; Für den USB488 Subclass Device Requests 'Read Status Byte'.
   ; USBTMC-USB488 Specification Rev1.0 - PDF Page 19, Table 12/13 -- READ_STATUS_BYTE control endpoint response format.
   Structure uReadStatusByteResponse Align #iUsbTmcStructureAlignment
      eUsbTmcStatus.a
      iTag.a
      iStatusByte.a
   EndStructure

   ;- uInterruptTransferResponse.
   ; Für die Antworten der USB488 Subclass Device Interrupt Transfers.
   ; USBTMC-USB488 Specification Rev1.0 - PDF Page 15
   ; Table 6 -- USB488 Interrupt-IN packet sent due to an SRQ condition,
   ; Table 7 -- USB488 Interrupt-IN packet sent due to READ_STATUS_BYTE request.
   Structure uInterruptTransferResponse Align #iUsbTmcStructureAlignment
      iNotify1.a
      iNotify2.a
   EndStructure

   ;-
   ;- -- Konstanten.

   ;-
   ;- -- Declares.

   Declare.i NewInstance(*uContext, *sError.String)
   Declare.i DeleteInstance(iInstanceNumber.i, *sError.String)

   Declare.i GetTmcDeviceList(iInstanceNumber.i, List TmcDeviceList.uDevice(), *iDeviceCount.Integer, *sError.String)
   Declare.i SetTmcDevice(iInstanceNumber.i, *uDevice.uDevice, *sError.String)
   Declare.i GetCapabilities(iInstanceNumber.i, *uCapabilities.uCapabilities, *sError.String)
   Declare.i GetTimeout(iInstanceNumber.i, *iTimeoutMilliseconds.Integer, *sError.String)
   Declare.i SetTimeout(iInstanceNumber.i, iTimeoutMilliseconds.i, *sError.String)
   Declare.i GetTimeoutControlTransfer(iInstanceNumber.i, *iTimeoutMilliseconds.Integer, *sError.String)
   Declare.i SetTimeoutControlTransfer(iInstanceNumber.i, iTimeoutMilliseconds.i, *sError.String)
   Declare.i GetTerminationChar(iInstanceNumber.i, *iTerminationChar.Integer, *fRequestEnabled.Integer, *sError.String)
   Declare.i SetTerminationChar(iInstanceNumber.i, iTerminationChar.i, fRequestEnabled.i, *sError.String)
   Declare.i GetTransferSplitDataLength(iInstanceNumber.i, *iTransferSplitDataLength.Integer, *sError.String)
   Declare.i SetTransferSplitDataLength(iInstanceNumber.i, iTransferSplitDataLength.i, *sError.String)
   
   Declare.i PrepareData(iInstanceNumber.i, sStringData.s, *BinaryData, iLengthBinaryData.i, *sError.String)
   Declare.i SendData(iInstanceNumber.i, eMessageId.i, *iTransferred.Integer, *sError.String)
   Declare.i RequestData(iInstanceNumber.i, *Data, *iLength.Integer, eMessageId.i, *sError.String)
   Declare.i WaitData(iInstanceNumber.i, *fMessageAvailable.Integer, *iStatusByte.Integer, *sError.String)
      
   Declare.i GetThreadData(iInstanceNumber.i, *uData.uThreadData, *sError.String)
   Declare.i SetThreadStatus(iInstanceNumber.i, eThreadStatus.i, *sError.String)
   Declare.i GetRequestProgressDataLength(iInstanceNumber.i, *iRequestProgressDataLength.Integer, *sError.String)
   Declare.i SetRequestProgressDataLength(iInstanceNumber.i, iRequestProgressDataLength.i, *sError.String)
   Declare.i GetRequestBufferLength(iInstanceNumber.i, *iRequestBufferLength.Integer, *sError.String)
   Declare.i SetRequestBufferLength(iInstanceNumber.i, iRequestBufferLength.i, *sError.String)
   Declare.i SendRequestThreadData(iInstanceNumber.i, *uData.uThreadData, *sError.String)

;  Declare.i ControlTransfer(iInstanceNumber.i, eRequestType.i, eDeviceRequest.i, iValue.i, iIndex.i,
;                            *Data.Integer, *iTransferred.Integer, *sError.String)
   Declare.i InterruptTransfer(iInstanceNumber.i, *uData.uInterruptTransferResponse, *sError.String)

   Declare.i InitiateAbortBulkOut(iInstanceNumber.i, iTransferIdentifier.i, *uData.uInitiateAbortBulkOutInResponse, *sError.String)
   Declare.i InitiateAbortBulkIn(iInstanceNumber.i, iTransferIdentifier.i, *uData.uInitiateAbortBulkOutInResponse, *sError.String)
   Declare.i CheckAbortBulkOutStatus(iInstanceNumber.i, *uData.uCheckAbortBulkOutStatusResponse, *sError.String)
   Declare.i CheckAbortBulkInStatus(iInstanceNumber.i, *uData.uCheckAbortBulkInStatusResponse, *sError.String)
   Declare.i InitiateClear(iInstanceNumber.i, *eUsbTmcStatus.Integer, *sError.String)
   Declare.i CheckClearStatus(iInstanceNumber.i, *uData.uCheckClearStatusResponse, *sError.String)
   Declare.i IndicatorPulse(iInstanceNumber.i, *Data.Integer, *sError.String)

   Declare.i ReadStatusByte(iInstanceNumber.i, iTag.i, *uData.uReadStatusByteResponse, *sError.String)
   Declare.i RemoteEnableControl(iInstanceNumber.i, fEnable.i, *eUsbTmcStatus.Integer, *sError.String)
   Declare.i LocalLockout(iInstanceNumber.i, *eUsbTmcStatus.Integer, *sError.String)
   Declare.i GoToLocal(iInstanceNumber.i, *eUsbTmcStatus.Integer, *sError.String)

EndDeclareModule  ; UsbTmc.


;-
;- -- Module UsbTmc. --------------------
;
Module UsbTmc
   EnableExplicit
   UseModule Common
   UseModule Macros
   DebugOutUsbTmc

   ;-
   ;- -- Macros.

   ;-
   ;- -- Enumerationen.

   ;- eDeviceRequest für Control Transfers.
   Enumeration eDeviceRequest
      ; USBTMC Device Requests.
      #eInitiateAbortBulkOut = 1
      #eCheckAbortBulkOutStatus = 2
      #eInitiateAbortBulkIn = 3
      #eCheckAbortBulkInStatus = 4
      #eInitiateClear = 5
      #eCheckClearStatus = 6
      #eGetCapabilities = 7
      #eIndicatorPulse = 64
      ; USB488 Subclass Device Requests.
      #eReadStatusByte = 128
      #eRenControl = 160
      #eGoToLocal = 161
      #eLocalLockout = 162
   EndEnumeration

   ;-
   ;- -- Konstanten.

   #iInterfaceClass = $FE                              ; Die Interface-Klasse die ein TMC-Gerät besitzen muss.
   #iInterfaceSubClass = 3                             ; Die Interface-Subklasse die ein TMC-Gerät besitzen muss.
   #iInterfaceProtocol = 1                             ; Das Interface-Protkoll das ein TMC-Gerät besitzen muss.
   #iDefaultTimeoutMilliseconds = 5000                 ; Der Rigol DSA815 benötigt 4 Sekunden bei einer Bildschirmkopie.
   #iDefaultTimeoutControlTransferMilliseconds = 500   ; Ein USBTMC-USB488-Gerät muss Standard Requests in dieser Zeit beantworten.
   #iDefaultTimeoutInterruptTransferMilliseconds = 5000  ; Die Zeit die auf eine Antwort gewartet wird.
   #iDefaultTerminationChar = $0A                      ; Ist Standard bei 488.2 USB488-Geräten.
   #fDefaultTerminationCharRequestEnabled = #False     ; Das Terminierungszeichen nicht im MessageHeader mitsenden.
   #iDefaultTransferSplitDataLength = #Null            ; Die gesamten Daten in einer einzelnen Transfer Message senden.
   #iDefaultRequestProgessDataLength = $19000          ; Die Länge der Daten die neu empfangen werden müssen, damit ein Ereignis
                                                       ; für den Fortschritt gesendet wird.
   #iDefaultRequestBufferLength = $120000              ; Die Größe des Empfangsbuffers in Bytes.
   
   #iTransferAttributeEndOfTransfer = %00000001        ; Bitmap-Mask für 'End Of Message' (EOF) im MessageHeader.
   #iTransferAttributeTermChar = %00000010             ; Bitmap-Mask für das TermChar im MessageHeader.

   ;-
   ;- -- Strukturen.

   ;- uInstance.
   ; Für die Verwaltung der Instanzen der Pseudoklasse.
   Structure uInstance
      iInstanceNumber.i                      ; Instanznummer, wird automatisch beim Erstellen der Pseudoklasse vergeben.
      iTimeoutMilliseconds.i                 ; Die Zeitspanne, die auf eine Antwort vom USB-Gerät gewartet wird.
      iTimeoutControlTransferMilliseconds.i  ; Die Zeitspanne, die auf eine Antwort vom USB-Gerät gewartet wird.
      iTimeoutInterruptTransferMilliseconds.i; Die Zeitspanne, die auf eine Antwort vom USB-Gerät gewartet wird.
      iTerminationChar.a                     ; Das Terminierungszeichen das die gesendeten und empfangenen Daten abschließt.
      fTerminationCharRequestEnabled.i       ; True wenn das TermChar bei einem Request vom Gerät beachtet werden muss.
      iTransferSplitDataLength.i             ; Die Länge der Daten, in die sie beim Senden in Transfer Messages aufgeteilt werden.
      iRequestProgressDataLength.i           ; Die Länge der neu empfangenen Daten, bei der ein Ereignis für den Fortschritt
                                             ; gesendet wird. Wird nur innerhalb eines Threads verwendet.
      iRequestBufferLength.i                 ; Die Größe des Empfangsbuffers in Bytes.
      *PreparedData                          ; Die MemoryID der zum Senden vorbereiteten Daten.
      iPreparedDataLength.i                  ; Die Länge der Daten ohne der Füllbytes.
      iTransferIdentifier.i                  ; Das Tag für die Headerdaten.
      iControlTransferIdentifier.i           ; Das Tag für den Control Tansfer.
      fLibUsbInitialized.i                   ; True wenn die LibUsb initialisiert ist, False wenn z.B. ein Context übergeben wurde.
      *uContext                              ; Zeiger auf den LibUsb-Context.
      iMutex.i                               ; Mutex für das Sperren des Zugriffs auf das USB-Gerät.
      uDevice.uDevice                        ; Die Daten des ausgewählten USBTMC-Geräts.
      uCapabilities.uCapabilities            ; Die zusätzlichen Attribute und Fähigkeiten des USB488-Geräts.
      uThreadData.uThreadData                ; Für das Vorbereiten, Senden und Empfangen der Daten in einem Thread.
      eThreadStatus.i                        ; Für das Steuern des Threads, ein Wert von der Enumeration eThreadStatus.
      iThreadNumber.i                        ; Die Nummer des Threads.
      fThreadRunning.i                       ; True wenn der Thread läuft und nicht pausiert.
   EndStructure

   ;- uGetCapabilities.
   ; Für die zusätzlichen Attribute und Fähigkeiten des USBTMC Gerätes.
   ; Information:
   ; USBTMC Specification Revision 1.0, April 14 2003, Page 29 Table 37 - GET_CAPABILITIES response format.
   ; USBTMC USB488 Subclass Specification Revision 1.0, April 14 2003, Page 10 Table 8 - GET_CAPABILITIES response format.
   ; Henry Bruhns, USB in der Messtechnik, 2008, Seite 207 - 7.11.7 GET_CAPABILITIES.
   Structure uGetCapabilities
      iUsbTmcStatus.a
      iReserved.a
      iBcdUsbTmc.u
      iInterfaceBitmap.a
      iDeviceBitmap.a
      iReservedUsbTmc.a[6]
      ; 488
      iBcdUsb488.u
      iInterface488Bitmap.a
      iDevice488Bitmap.a
      iReservedUsb488.a[8]
   EndStructure

   ;- uBulkHeader für Bulk-Out- und Bulk-In-MessageHeader für Send und Request.
   ; Information:
   ; USBTMC Specification Revision 1.0, April 14 2003, Page 5 Table 1 - 3.2 Bulk-OUT endpoint.
   Structure uBulkHeader
      iMsgId.a
      iTag.a
      iTagInverse.a
      iReserved1.a
      iTransferSize.l
      iTransferAttributes.a
      iTermChar.a
      iReserved2.a[2]
   EndStructure

   ;-
   ;- -- Variablen.

   Global miInstanceCounter.i, miInstanceCounterMax.i ; Zähler für die Instanznummern der Verwaltung.
   Global NewMap mInstanceMap.uInstance()             ; Map für die Verwaltung der Instanzen.
   Global miMutex.i                                   ; Mutex für die Verwaltung in Threads.
   Global miThreadCounter.i, miThreadCounterMax.i     ; Zähler für die Threadnummern der Verwaltung.
   Threaded miThreadedThreadCounter.i                 ; Zähler für die Threads und die Debugausgabe.

   ;-

   Procedure.i LockInstance(iInstanceNumber.i, *sError.String, fNewInstance = #False)
      ; Reserviert die Instanz für die Verwaltung.
      ; Es muss immer UnlockInstance() danach aufgerufen werden um den Mutex freizugeben.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- sError:            Fehlerbeschreibung.
      ; -> fNewInstance:      Bei True wird keine Fehlermeldung ausgegeben wenn die Instanz nicht gefunden wurde.
      ; Rückgabe:             True wenn die Instanz in der Verwaltung gefunden wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, sKey.s

      If (Not miMutex)
         DebugView::Error("Es ist kein Mutex vorhanden.", iInstanceNumber, miThreadedThreadCounter)
      Else
         LockMutex(miMutex)
      EndIf

      sKey = Str(iInstanceNumber)
      If FindMapElement(mInstanceMap(), sKey)
         fReturn = #True
      EndIf

      If Not (fReturn Or fNewInstance)
         *sError\s = CurrentProcedureNameInstance + "Diese Instanz existiert nicht."
         DebugView::Error("Diese Instanz existiert nicht.", iInstanceNumber, miThreadedThreadCounter)
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure UnlockInstance(iInstanceNumber.i)
      ; Gibt die mit LockInstance() reservierte Instanz wieder frei.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)

      If (Not miMutex)
         DebugView::Error("Es ist kein Mutex vorhanden.", iInstanceNumber, miThreadedThreadCounter)
      Else
         UnlockMutex(miMutex)
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
   EndProcedure


   Procedure.i LockDevice(iInstanceNumber.i, *sError.String)
      ; Sperrt den Zugriff auf das Gerät.
      ; Es muss immer UnlockDevice() danach aufgerufen werden um das Gerät bzw. den Mutex freizugeben.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn das Gerät bzw.der Mutex gesperrt wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            ; Der Mutex wird beim Erstellen der Instanz angelegt, aber beim Beenden eines Threads möglicherweise gelöscht.
            If (Not \iMutex)
               \iMutex = CreateMutex()
            EndIf
            If \iMutex
               fReturn = Bool(TryLockMutex(\iMutex))
               If (Not fReturn)
                  *sError\s = CurrentProcedureNameInstance + "Das Gerät konnte nicht gesperrt werden."
                  DebugView::Error("Das Gerät konnte nicht gesperrt werden.", iInstanceNumber, miThreadedThreadCounter)
               EndIf
            EndIf
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure UnlockDevice(iInstanceNumber.i)
      ; Gibt den Zugriff auf das Gerät, das mit LockDevice() gesperrt wurde, wieder frei.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected sError.String

      If LockInstance(iInstanceNumber, sError)
         With mInstanceMap()
            If \iMutex
               UnlockMutex(\iMutex)
            EndIf
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
   EndProcedure


   Procedure.i StorePreparedData(iInstanceNumber.i, *Data, iLength, *sError.String)
      ; Speichert die MemoryId bzw. Speicheradresse der vorbereiteten Daten für das Senden.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> iData:             Die MemoryId.
      ; -> iLength:           Die Länge der Daten ohne den Füllbytes, wird für den MessageHeader benötigt.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die MemoryId gespeichert wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            \PreparedData = *Data
            \iPreparedDataLength = iLength
         EndWith
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i FreePreparedData(iInstanceNumber.i, *sError.String)
      ; Gibt die vorhandenen vorbereiteten Daten bzw. den reservierten Speicher wieder frei.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn der Speicher freigegeben wurden, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            If \PreparedData
               FreeMemory(\PreparedData)
               \PreparedData = #Null
            EndIf
         EndWith
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i CalculateFillBytes(iInstanceNumber.i, iLength.i, iDivider.i = 4)
      ; Berechnet die Anzahl der benötigten Füllbytes, damit die Daten ohne Rest teilbar sind.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> iLength:           Die Länge der vorhandenen Daten.
      ; -> iDivider:          Die Zahl durch der die Daten ohne Rest teilbar sein müssen.
      ; Rückgabe:             Die Anzahl der benötigten Füllbytes.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected iReturn.i

      If iLength
         iLength = iLength % iDivider
         If iLength
            iReturn = iDivider - iLength
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn iReturn
   EndProcedure


   Procedure.i IncrementControlTransferIdentifier(iInstanceNumber.i, iTag.i)
      ; Erhöht den Control Transfer Identifier (Tag).
      ; Es sind nur Werte von 2 bis 127 erlaubt.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> iTag:              Den Control Transfer Identifier.
      ; Rückgabe:             Den Control Transfer Identifier.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)

      iTag + 1
      If (iTag <= 1)
         iTag = 2
      ElseIf (iTag > 127)
         iTag = 2
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn iTag
   EndProcedure


   Procedure.i IncrementTransferIdentifier(iInstanceNumber.i, iTag.i)
      ; Erhöht den Transfer Identifier (Tag) für den Message Header um 1.
      ; Es sind nur Werte von 1 bis 255 erlaubt.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> iTag:              Den Transfer Identifier.
      ; Rückgabe:             Den Transfer Identifier.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)

      iTag + 1
      If (iTag <= 0)
         iTag = 1
      ElseIf (iTag > 255)
         iTag = 1
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn iTag
   EndProcedure


   Procedure.i CreateSendBulkOutHeaderWithData(iInstanceNumber.i, *BulkOutData.Integer, *Data, iLengthData.i,
                                               eMessageId.i, fEndOfMessage.i, *sError.String)
      ; Erstellt den USBTMC BulkOut Header inklusive der Nutzerdaten.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *BulkOutData:      Ein Zeiger auf die erstellten BulkOut-Daten.
      ; -> *Data:             Ein Zeiger auf die Nutzerdaten die gesendet werden.
      ; -> iLengthData:       Die Länge der Daten ohne Terminierungszeichen und ohne Füllbytes.
      ; -> eMessageId:        Die USBTMC MessageID für den Type des BulkOut-Headers.
      ; -> fEndOfMessage:     True für die letzte Message, betrifft das Transfer Attribute.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die BulkOut-Daten erstellt wurden, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fError.i, *uBulkOutHeader.uBulkHeader
      Protected iTag.i, iTermChar.a, iLengthTermChar.i, iLengthFillBytes.i

      ; Den Transfer Identifier ermitteln.
      ; Die Existenz der Instanz wurde schon geprüft, darum gibt es hier keine Fehlerüberprüfung.
      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            iTag = IncrementTransferIdentifier(iInstanceNumber, \iTransferIdentifier)
            \iTransferIdentifier = iTag
            iTermChar = \iTerminationChar
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)
      
      ; Das Terminierungszeichen und die Füllbytes ermitteln.
      If fEndOfMessage
         If iTermChar
            iLengthTermChar = 1
         EndIf
         iLengthFillBytes = CalculateFillBytes(iInstanceNumber, iLengthData + iLengthTermChar)
      EndIf
      
      ; Speicher für die Daten reservieren.
      If fEndOfMessage
         *BulkOutData\i = AllocateMemory(SizeOf(uBulkHeader) + iLengthData + iLengthTermChar + iLengthFillBytes)
      Else
         *BulkOutData\i = AllocateMemory(SizeOf(uBulkHeader) + iLengthData)
      EndIf
      
      If (Not *BulkOutData\i)
         *sError\s = CurrentProcedureNameInstance + "Der BulkOut-Speicher konnte nicht alloziert werden."
         DebugView::Error("Der BulkOut-Speicher konnte nicht alloziert werden.", iInstanceNumber, miThreadedThreadCounter)
         fError = #True
      EndIf

      If (Not fError)
         ; Den BulkOut-Header füllen.
         *uBulkOutHeader = *BulkOutData\i
         With *uBulkOutHeader
            \iMsgId = eMessageId
            \iTag = iTag
            \iTagInverse = (iTag ! $FF)
            ; Nur die Nutzdaten ohne Message-Header und Füllbytes.
            \iTransferSize = iLengthData + iLengthTermChar
            ; Bit 0 für 'End Of Message' (EOF) setzen, aber nicht für Vendor Specific.
            If (fEndOfMessage And (eMessageId = #eMessageIdDevDepMsgOut))
               \iTransferAttributes = #iTransferAttributeEndOfTransfer
            Else
               \iTransferAttributes = #Null
            EndIf
         EndWith

         ; Die Nutzerdaten und Terminierungszeichen hinter den BulkOut-Header kopieren.
         If *Data
            CopyMemory(*Data, *BulkOutData\i + SizeOf(uBulkHeader), iLengthData)
            If (fEndOfMessage And iTermChar)
               PokeA(*BulkOutData\i + SizeOf(uBulkHeader) + iLengthData, iTermChar)
            EndIf
         EndIf

         fReturn = #True
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i CreateRequestBulkOutHeader(iInstanceNumber.i, *uBulkOutHeader.uBulkHeader, iLengthData.i,
                                          eMessageId.i, *sError.String)
      ; Erstellt den USBTMC BulkOut Header.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *uBulkOutHeader:   Ein Zeiger auf die Request-BulkOut-Header-Struktur.
      ; -> iLengthData:       Die Länge der Daten die empfangen werden können.
      ; -> eMessageId:        Die USBTMC MessageID für den Type des BulkOut-Headers.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die BulkOut-Daten erstellt wurden, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fError.i, iTag.i, iTermChar.a, fTerminationCharRequestEnabled.i

      If (Not *uBulkOutHeader)
         *sError\s = CurrentProcedureNameInstance + "Der Zeiger auf die BulkOut-Header-Struktur ist Null. *uBulkOutHeader = 0"
         DebugView::Error("Der Zeiger auf die BulkOut-Header-Struktur ist Null. *uBulkOutHeader = 0",
                          iInstanceNumber, miThreadedThreadCounter)

      Else
         ; Den Transfer Identifier und das Terminierungszeichen ermitteln.
         ; Die Existenz der Instanz wurde schon geprüft, darum gibt es hier keine Fehlerüberprüfung.
         If LockInstance(iInstanceNumber, *sError)
            With mInstanceMap()
               iTag = IncrementTransferIdentifier(iInstanceNumber, \iTransferIdentifier)
               \iTransferIdentifier = iTag
               iTermChar = \iTerminationChar
               fTerminationCharRequestEnabled = \fTerminationCharRequestEnabled
            EndWith
         EndIf
         UnlockInstance(iInstanceNumber)

         ; Den Request-BulkOut-Header füllen.
         With *uBulkOutHeader
            \iMsgId = eMessageId
            \iTag = iTag
            \iTagInverse = (iTag ! $FF)
            \iTransferSize = iLengthData
            If fTerminationCharRequestEnabled
               \iTransferAttributes = #iTransferAttributeTermChar ; Bit 1 für 'TermChar' setzen.
               \iTermChar = iTermChar
            Else
               \iTransferAttributes = #Null  ; Bit 1 für 'TermChar' nicht setzen.
               \iTermChar = 0
            EndIf
         EndWith

         fReturn = #True
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i Constructor(iInstanceNumber.i, *uContext, fInitializedIntern.i, *sError.String)
      ; Fügt eine neue Instanz der internen Verwaltung hinzu.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> *uContext:            Der Zeiger auf die LibUsb-Context-Struktur.
      ; -> fInitializedIntern:   True wenn die LibUsb von der Instanz initialisiert wurde, und kein Context übergeben wurde.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True wenn die Instanz erstellt wurde, ansonst False.
      DebugView::ProcIn(miInstanceCounterMax)
      Protected fReturn.i, sKey.s

      If (Not miMutex)
         miMutex = CreateMutex()
         If (Not miMutex)
            *sError\s = CurrentProcedureNameInstance + "Der Mutex konnte nicht erstellt werden."
            DebugView::Error("Der Mutex konnte nicht erstellt werden.", iInstanceNumber)
         EndIf
      EndIf

      If LockInstance(iInstanceNumber, *sError, #True)
         *sError\s = CurrentProcedureNameInstance + "Diese Instanz ist schon vorhanden."
         DebugView::Error("Diese Instanz ist schon vorhanden.", iInstanceNumber)

      Else
         sKey = Str(iInstanceNumber)
         AddMapElement(mInstanceMap(), sKey, #PB_Map_NoElementCheck)

         With mInstanceMap()
            \iInstanceNumber = iInstanceNumber
            \iTimeoutMilliseconds = #iDefaultTimeoutMilliseconds
            \iTimeoutControlTransferMilliseconds = #iDefaultTimeoutControlTransferMilliseconds
            \iTimeoutInterruptTransferMilliseconds = #iDefaultTimeoutInterruptTransferMilliseconds
            \iTerminationChar = #iDefaultTerminationChar
            \fTerminationCharRequestEnabled = #fDefaultTerminationCharRequestEnabled
            \iTransferSplitDataLength = #iDefaultTransferSplitDataLength
            \iRequestProgressDataLength = #iDefaultRequestProgessDataLength
            \iRequestBufferLength = #iDefaultRequestBufferLength
            \fLibUsbInitialized = fInitializedIntern
            \uContext = *uContext
            \iMutex = CreateMutex()
         EndWith

         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i Destructor(iInstanceNumber.i, *sError.String)
      ; Entfernt diese Instanz aus der internen Verwaltung.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die Instanz entfernt wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber)
      Protected fReturn.i

      ; Vorhandene Daten freigeben.
      If (Not FreePreparedData(iInstanceNumber, *sError))
         ; Fehler, die Instanz ist nicht vorhanden.
      EndIf

      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            If \iMutex
               FreeMutex(\iMutex)
            EndIf
         EndWith

         DeleteMapElement(mInstanceMap())
         miInstanceCounter - 1
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      ; Ressourcen freigeben.
      If (miInstanceCounter = #Null)
         If miMutex
            FreeMutex(miMutex)
            miMutex = #Null
         EndIf
         ClearMap(mInstanceMap())
      EndIf

      DebugView::Info("miInstanceCounter = " + miInstanceCounter, iInstanceNumber)
      DebugView::ProcOut(iInstanceNumber)
      ProcedureReturn fReturn
   EndProcedure

   ;- --

   Procedure.i NewInstance(*uContext, *sError.String)
      ; Erstellt eine neue Instanz.
      ; Beim Löschen der Instanz wird die LibUsb mit dem Context geschlossen, siehe DeleteInstance.
      ; -> *uContext:   Der Zeiger auf eine LibUsb-Context-Struktur, oder #Null für den Standard-Context.
      ; <- sError:      Fehlerbeschreibung.
      ; Rückgabe:       Die Instanznummer oder Null bei einem Fehler.
      miInstanceCounterMax + 1
      miInstanceCounter + 1
      DebugView::ProcIn(miInstanceCounterMax)
      Protected iReturn.i, iInstanceNumber.i = miInstanceCounterMax
      Protected fInitialized.i, fInitializedIntern.i

      ; Die LibUsb mit dem Standard-Context öffnen wenn kein Context übergeben wurde.
      If *uContext
         fInitialized = #True
         fInitializedIntern = #False
      Else
         fInitialized = LibUsb::Init(*uContext, *sError)
         fInitializedIntern = #True
      EndIf

      ; Die neue Instanz erstellen.
      If fInitialized
         If (Not Constructor(iInstanceNumber, *uContext, fInitializedIntern, *sError))
            ; Die Instanz konnte nicht erstellt werden.
            miInstanceCounterMax - 1
            miInstanceCounter - 1
         Else
            iReturn = iInstanceNumber
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber)
      ProcedureReturn iReturn
   EndProcedure


   Procedure.i DeleteInstance(iInstanceNumber.i, *sError.String)
      ; Löscht diese Instanz.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die Instanz gelöscht wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber)
      Protected fReturn.i, iThreadNumber.i

      ; Einen vorhandenen Thread beenden.
      SetThreadStatus(iInstanceNumber, #eThreadStatusEnd, *sError)

      ; Ein geöffnetes Gerät und die LibUsb schließen.
      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            ; Ein geöffnetes Gerät schließen.
            If (\uDevice\uDeviceHandle)
               ; Die Schnittstelle freigeben.
               If (Not LibUsb::ClaimInterface(\uDevice\uDeviceHandle, \uDevice\iInterfaceNumber, *sError))
                  DebugView::Error("Die Schnittstelle konnte nicht freigegeben werden.", iInstanceNumber)
               EndIf
               ; Das Gerät schließen.
               LibUsb::Close(\uDevice\uDeviceHandle)
            EndIf
            ; Eine vorhandene Geräteliste freigeben.
            If \uDevice\FreeDeviceList
               LibUsb::FreeDeviceList(\uDevice\FreeDeviceList, #True)
            EndIf
            ; Die LibUsb schließen.
            If \fLibUsbInitialized
               LibUsb::Exit(\uContext)
            EndIf
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)

      ; Die Instanz löschen.
      fReturn = Destructor(iInstanceNumber, *sError)

      DebugView::ProcOut(iInstanceNumber)
      ProcedureReturn fReturn
   EndProcedure

   ;- --

   Procedure.i GetTmcDeviceList(iInstanceNumber.i, List TmcDeviceList.uDevice(), *iDeviceCount.Integer, *sError.String)
      ; Ermittelt alle verfügbaren USBTMC-Geräte mit installiertem WinUSB-Treiber.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- TmcDeviceList():   Die Liste mit den verfügbaren USB-Geräten.
      ; <- *iDeviceCount:     Die Anzahl der verfügbaren USBTMC-Geräte, kann auch Null sein.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die Liste erstellt wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      NewList *LibusbDeviceList.LibUsb::uDevice()
      NewList CheckDeviceList.uDevice()
      Protected fReturn.i, fError.i, fResult.i, fFreeDeviceList.i, sString.String, sError.String
      Protected *uContext, *FreeDeviceList, *uDevice.LibUsb::uDevice, *uDeviceHandle
      Protected *uCheckDevice.uDevice, iDevicesCount.i
      Protected uDeviceDescriptor.LibUsb::uDeviceDescriptor
      Protected *uConfigDescriptor.LibUsb::uConfigDescriptor
      Protected *uInterfaceDescriptor.LibUsb::uInterfaceDescriptor
      Protected *uEndpoints, *uEndpoint.LibUsb::uEndpointDescriptor
      Protected iAttributes.i, iTransferType.i, iNumber.i
      Protected c.i, k.i, i.i, cIndex.i

      ;{ Den Context einlesen.
      If LockInstance(iInstanceNumber, *sError)
         *uContext = mInstanceMap()\uContext
         fResult = #True

         ; Eine vorhandene Geräteliste freigeben.
         If mInstanceMap()\uDevice\FreeDeviceList
            LibUsb::FreeDeviceList(mInstanceMap()\uDevice\FreeDeviceList, #True)
            mInstanceMap()\uDevice\FreeDeviceList = #Null
         EndIf
      EndIf
      UnlockInstance(iInstanceNumber)  ;}

      ;{ Eine Liste mit den USB-Geräten ermitteln.
      If fResult
         fFreeDeviceList = LibUsb::GetDeviceList(*uContext, *LibusbDeviceList(), @iDevicesCount, @*FreeDeviceList, sError)

         ; Den Zeiger auf die Geräteliste sichern.
         If fFreeDeviceList
            If LockInstance(iInstanceNumber, *sError)
               mInstanceMap()\uDevice\FreeDeviceList = *FreeDeviceList
            EndIf
            UnlockInstance(iInstanceNumber)
         EndIf
      EndIf ;}

      ;{ Von allen vorhandenen USB-Geräten die DeviceDescriptoren ermitteln.
      If (Not fFreeDeviceList)
         fError = #True
      Else
         ForEach *LibusbDeviceList()
            If *LibusbDeviceList()
               If LibUsb::GetDeviceDescriptor(*LibusbDeviceList(), uDeviceDescriptor, sError)

                  ; Geräte mit der Class, SubClass und Protocol 0 in die temporäre Liste zur weiteren Prüfung aufnehmen.
                  With uDeviceDescriptor
                     If ((\iDeviceClass = 0) And (\iDeviceSubClass = 0) And (\iDeviceProtocol = 0))
                        *uCheckDevice = AddElement(CheckDeviceList())
                        If *uCheckDevice
                           *uCheckDevice\uDevice = *LibusbDeviceList()
                           ; Der Context des Geräts muss mit dem Context der Instanz später bei der Gerätezuweisung übereinstimmen.
                           *uCheckDevice\uContext = *uContext
                           *uCheckDevice\iVendorId = \iVendorId
                           *uCheckDevice\iProductId = \iProductId
                           ; Die Bus- und Portnummern sichern, um Geräte mit gleicher Seriennummer unterscheiden zu können.
                           LibUsb::GetBusNumber(*LibusbDeviceList(), @iNumber, sError)
                           *uCheckDevice\iBusNumber = iNumber
                           iNumber = #Null
                           LibUsb::GetPortNumber(*LibusbDeviceList(), @iNumber, sError)
                           *uCheckDevice\iPortNumber = iNumber
                           iNumber = #Null
                           ; Die Indexe der Sting-Descriptoren für Manufacturer, Product und Serialnumber sichern.
                           *uCheckDevice\iManufacturer = \iManufacturer
                           *uCheckDevice\iProduct = \iProduct
                           *uCheckDevice\iSerialnumber = \iSerialNumber
                           ; Die Anzahl der Konfigurationen sichern.
                           *uCheckDevice\iConfigurationCount = \iNumConfigurations
                        EndIf
                     EndIf
                  EndWith

               EndIf
            EndIf
         Next
      EndIf ;}

      ; Versuchen die Geräte zu öffnen und einige String-Descriptoren einzulesen.
      ; Das funktioniert teilweise auch bei Geräten, für die kein WinUSB-Treiber installiert ist.
      ForEach CheckDeviceList()
         ;{
         *uCheckDevice = CheckDeviceList()
         If *uCheckDevice\uDevice
            If LibUsb::Open(*uCheckDevice\uDevice, @*uDeviceHandle, sError)

               ; Die erste Konfiguration setzen.
               LibUsb::SetConfiguration(*uDeviceHandle, 1, sError)

               ; Die String-Descriptoren für Manufacturer, Product und Serialnumber ermitteln.
               With *uCheckDevice
                  If \iManufacturer
                     If LibUsb::GetStringDescriptor(*uDeviceHandle, \iManufacturer, #Null, sString, sError)
                        \sManufacturer = sString\s
                     EndIf
                  EndIf
                  If \iProduct
                     If LibUsb::GetStringDescriptor(*uDeviceHandle, \iProduct, #Null, sString, sError)
                        \sProduct = sString\s
                     EndIf
                  EndIf
                  If \iSerialnumber
                     If LibUsb::GetStringDescriptor(*uDeviceHandle, \iSerialnumber, #Null, sString, sError)
                        \sSerialnumber = sString\s
                     EndIf
                  EndIf
               EndWith

               ; Das Gerät wieder schließen.
               LibUsb::Close(*uDeviceHandle)
               *uDeviceHandle = #Null
            EndIf ;}

            ; Das erste USBTMC-Interface verwenden, das in den Konfigurationen gefunden wird.
            For cIndex = 0 To *uCheckDevice\iConfigurationCount

               ; Von der Konfiguration das Interface 0 und die Adressen der zugehörigen Endpoints ermitteln.
               If LibUsb::GetConfigDescriptor(*uCheckDevice\uDevice, cIndex, @*uConfigDescriptor, sError)
                  ; Das erste Interfaces ermitteln.
                  ; Ein USBTMC-Gerät muss ein Interface mit der Nummer 0 besitzen.
                  If *uConfigDescriptor\uInterface
                     *uInterfaceDescriptor = *uConfigDescriptor\uInterface\uAltSetting
                     If *uInterfaceDescriptor

                        ;{ Die Interface-Klasse auf ein TMC-Gerät prüfen.
                        With *uInterfaceDescriptor
                           If (\iInterfaceClass = #iInterfaceClass)
                              If (\iInterfaceSubClass = #iInterfaceSubClass)
                                 If (\iInterfaceProtocol = #iInterfaceProtocol)
                                    ; Das Gerät ist ein USBTMC-Gerät.
                                    *uCheckDevice\fUsbTmc = #True
                                    *uCheckDevice\iInterfaceNumber = \iInterfaceNumber
                                    *uCheckDevice\iConfigurationValue = *uConfigDescriptor\iConfigurationValue
                                 EndIf
                              EndIf
                           EndIf
                        EndWith  ;}

                        ;{ Die Adressen der Endpoints ermitteln.
                        If *uCheckDevice\fUsbTmc
                           *uEndpoints = *uInterfaceDescriptor\uEndpoint
                           If *uEndpoints
                              For i = 1 To *uInterfaceDescriptor\iNumEndpoints
                                 *uEndpoint = (*uEndpoints + ((i - 1) * SizeOf(LibUsb::uEndpointDescriptor)))
                                 If *uEndpoint

                                    ; Für das Unterscheiden zwischen Interrupt- und Bulk-Transfer Endpoint.
                                    iAttributes = (*uEndpoint\iAttributes & LibUsb::#eLIBUSB_TRANSFER_TYPE_INTERRUPT)
                                    ; Für das Unterscheiden zwischen In- und Out-Endpoint.
                                    iTransferType = (*uEndpoint\iEndpointAddress & LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST)

                                    ; Interrupt-Endpoint.
                                    If (iAttributes = LibUsb::#eLIBUSB_TRANSFER_TYPE_INTERRUPT)
                                       ; Prüfen ob In- oder Out-Endpoint.
                                       ; Ein USBTMC-USB488-Gerät kann optional einen Interrupt-In Endpoint besitzen.
                                       If (iTransferType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST)
                                          *uCheckDevice\iEndpointInterruptIn = *uEndpoint\iEndpointAddress
                                       EndIf

                                    ; Bulk-Endpoint.
                                    ElseIf (iAttributes = LibUsb::#eLIBUSB_TRANSFER_TYPE_BULK)
                                       ; Prüfen ob In- oder Out-Endpoint.
                                       If (iTransferType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST)
                                          *uCheckDevice\iEndpointBulkIn = *uEndpoint\iEndpointAddress
                                          *uCheckDevice\iEndpointBulkInMaxPacketSize = *uEndpoint\iMaxPacketSize
                                       ElseIf (iTransferType = LibUsb::#eLIBUSB_REQUEST_HOSTTODEVICE)
                                          *uCheckDevice\iEndpointBulkOut = *uEndpoint\iEndpointAddress
                                          *uCheckDevice\iEndpointBulkOutMaxPacketSize = *uEndpoint\iMaxPacketSize
                                       EndIf
                                    EndIf

                                 EndIf
                              Next i
                           EndIf

                           Break ; For cIndex. USBTMC-Interface gefunden.

                        EndIf ;}
                     EndIf
                  EndIf
               EndIf
            Next cIndex

            ; Die Konfiguration wieder freigeben.
            If *uConfigDescriptor
               LibUsb::FreeConfigDescriptor(*uConfigDescriptor)
               *uConfigDescriptor = #Null
            EndIf
         EndIf

      Next  ; ForEach CheckDeviceList()).

      ;{ Nur die USBTMC-Geräte in die Liste lassen, alle anderen davon entfernen.
      ForEach CheckDeviceList()
         *uCheckDevice = CheckDeviceList()
         If *uCheckDevice
            If (Not *uCheckDevice\fUsbTmc)
               DeleteElement(CheckDeviceList())
            EndIf
         EndIf
      Next  ;}

      ;{ Erfolg zurückgeben wenn Geräte verfügbar sind, und die USBTMC-Geräte für die Rückgabe kopieren.
      If (Not fError)
         fReturn = #True
         *iDeviceCount\i = ListSize(CheckDeviceList())
         If *iDeviceCount\i
            CopyList(CheckDeviceList(), TmcDeviceList())
         EndIf
      EndIf ;}

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SetTmcDevice(iInstanceNumber.i, *uDevice.uDevice, *sError.String)
      ; Das übergebene USBTMC-Gerät verwenden und öffnen.
      ; Die Daten stammen aus der Liste von der Funktion GetTmcDeviceList().
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> *uDevice:          Das ausgewählte Gerät.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn das Gerät geöffnet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fError.i, *uInstance.uInstance, *uDeviceHandle

      If (Not *uDevice)
         fError = #True
         *sError\s = CurrentProcedureNameInstance + "Es wurde kein Gerät ausgewählt. *uDevice = 0"
         DebugView::Error("Es wurde kein Gerät ausgewählt. *uDevice = 0", iInstanceNumber, miThreadedThreadCounter)

      Else
         If LockInstance(iInstanceNumber, *sError)
            With mInstanceMap()
               ; Prüfen ob der Context der Instanz mit dem Context des Geräts übereinstimmt.
               If Not (\uContext = *uDevice\uContext)
                  fError = #True
                  *sError\s = CurrentProcedureNameInstance +
                              "Der Context der Instanz stimmt nicht mit dem Context des Geräts überein. " +
                              "mInstanceMap()\uContext = " + \uContext + " - *uDevice\uContext = " + *uDevice\uContext
                  DebugView::Error("Der Context der Instanz stimmt nicht mit dem Context des Geräts überein. " +
                              "mInstanceMap()\uContext = " + \uContext + " - *uDevice\uContext = " + *uDevice\uContext,
                              iInstanceNumber, miThreadedThreadCounter)
               EndIf

               ; Ein schon vorhandenes geöffnetes Gerät schließen und die Daten des neuen Geräts kopieren.
               If (Not fError)
                  If (\uDevice\uDeviceHandle)
                     ; Die Schnittstelle freigeben.
                     If (Not LibUsb::ClaimInterface(\uDevice\uDeviceHandle, \uDevice\iInterfaceNumber, *sError))
                        DebugView::Error("Die Schnittstelle konnte nicht freigegeben werden.",
                                         iInstanceNumber, miThreadedThreadCounter)
                     EndIf
                     ; Das Gerät schließen.
                     LibUsb::Close(\uDevice\uDeviceHandle)
                  EndIf

                  ; Die Daten vom ausgewählten Gerät kopieren.
                  CopyStructure(*uDevice, \uDevice, uDevice)
                  \uDevice\uDeviceHandle = #Null
                  *uInstance = mInstanceMap()
               EndIf
            EndWith
         EndIf
         UnlockInstance(iInstanceNumber)
      EndIf

      ; Das Gerät öffnen, die Schnittstelle reservieren und das DeviceHandle sichern.
      If (Not fError)
         With *uInstance
            If (Not \uDevice\uDevice)
               *sError\s = CurrentProcedureNameInstance + "Das Gerät ist nicht vorhanden. *uInstance\uDevice\uDevice = 0"
               DebugView::Error("Das Gerät ist nicht vorhanden. *uInstance\uDevice\uDevice = 0",
                                iInstanceNumber, miThreadedThreadCounter)

            Else
               ; Das Gerät öffnen.
               If ( Not LibUsb::Open(\uDevice\uDevice, @*uDeviceHandle, *sError))
                  DebugView::Error("Das Gerät konnte nicht geöffnet werden.", iInstanceNumber, miThreadedThreadCounter)

               Else
                  ; Die Konfiguration mit der USBTMC-Schnittstelle setzen.
                  If (Not LibUsb::SetConfiguration(*uDeviceHandle, \uDevice\iConfigurationValue, *sError))
                     DebugView::Error("Die Konfiguration " + \uDevice\iConfigurationValue +
                                      " konnte nicht gesetzt werden.", iInstanceNumber, miThreadedThreadCounter)

                  Else
                  ; Die Schnittstelle reservieren.
                     If (Not LibUsb::ClaimInterface(*uDeviceHandle, \uDevice\iInterfaceNumber, *sError))
                        DebugView::Error("Die Schnittstelle konnte nicht reserviert werden.",
                                         iInstanceNumber, miThreadedThreadCounter)

                     Else
                        ; Das Handle zum geöffneten Gerät sichern.
                        If LockInstance(iInstanceNumber, *sError)
                           mInstanceMap()\uDevice\uDeviceHandle = *uDeviceHandle
                        EndIf
                        UnlockInstance(iInstanceNumber)

                        ; Die zusätzlichen Attribute und Fähigkeiten des USBTMC-Gerätes ermitteln.
                        ; Mögliche Fehler ignorieren.
                        If (Not GetCapabilities(iInstanceNumber, #Null, *sError))
                           DebugView::Error(*sError\s)
                        EndIf

                        fReturn = #True
                     EndIf
                  EndIf
               EndIf
            EndIf
         EndWith
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i GetTimeout(iInstanceNumber.i, *iTimeoutMilliseconds.Integer, *sError.String)
      ; Gibt die Zeitspanne zurück, die auf eine Antwort vom USB-Gerät gewartet wird.
      ; -> iInstanceNumber:         Die Nummer der Instanz.
      ; <- *iTimeoutMilliseconds:   Die Zeit in Millisekunden.
      ; <- sError:                  Fehlerbeschreibung.
      ; Rückgabe:                   True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         If *iTimeoutMilliseconds
            *iTimeoutMilliseconds\i = mInstanceMap()\iTimeoutMilliseconds
            fReturn = #True
         EndIf
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SetTimeout(iInstanceNumber.i, iTimeoutMilliseconds.i, *sError.String)
      ; Setzt die Zeitspanne die auf eine Antwort vom USB-Gerät gewartet wird.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> iTimeoutMilliseconds: Die Zeit in Millisekunden.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         mInstanceMap()\iTimeoutMilliseconds = iTimeoutMilliseconds
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i GetTimeoutControlTransfer(iInstanceNumber.i, *iTimeoutMilliseconds.Integer, *sError.String)
      ; Gibt die Zeitspanne zurück, die auf eine Antwort eines Control Transfers vom USB-Gerät gewartet wird.
      ; -> iInstanceNumber:         Die Nummer der Instanz.
      ; <- *iTimeoutMilliseconds:   Die Zeit in Millisekunden.
      ; <- sError:                  Fehlerbeschreibung.
      ; Rückgabe:                   True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         If *iTimeoutMilliseconds
            *iTimeoutMilliseconds\i = mInstanceMap()\iTimeoutControlTransferMilliseconds
            fReturn = #True
         EndIf
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SetTimeoutControlTransfer(iInstanceNumber.i, iTimeoutMilliseconds.i, *sError.String)
      ; Setzt die Zeitspanne die auf eine Antwort eines Control Transfers vom USB-Gerät gewartet wird.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> iTimeoutMilliseconds: Die Zeit in Millisekunden.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         mInstanceMap()\iTimeoutControlTransferMilliseconds = iTimeoutMilliseconds
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i GetTimeoutInterruptTransfer(iInstanceNumber.i, *iTimeoutMilliseconds.Integer, *sError.String)
      ; Gibt die Zeitspanne zurück, die auf eine Antwort eines Interrupt Transfers vom USB-Gerät gewartet wird.
      ; -> iInstanceNumber:         Die Nummer der Instanz.
      ; <- *iTimeoutMilliseconds:   Die Zeit in Millisekunden.
      ; <- sError:                  Fehlerbeschreibung.
      ; Rückgabe:                   True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         If *iTimeoutMilliseconds
            *iTimeoutMilliseconds\i = mInstanceMap()\iTimeoutInterruptTransferMilliseconds
            fReturn = #True
         EndIf
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SetTimeoutInterruptTransfer(iInstanceNumber.i, iTimeoutMilliseconds.i, *sError.String)
      ; Setzt die Zeitspanne die auf eine Antwort eines Interrupt Transfers vom USB-Gerät gewartet wird.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> iTimeoutMilliseconds: Die Zeit in Millisekunden.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         mInstanceMap()\iTimeoutInterruptTransferMilliseconds = iTimeoutMilliseconds
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i GetTerminationChar(iInstanceNumber.i, *iTerminationChar.Integer, *fRequestEnabled.Integer, *sError.String)
      ; Gibt das Terminierungszeichen zurück, mit dem die gesendeten und empfangenen Daten abgeschlossen werden.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *iTerminationChar: Das Terminierungszeichen oder Null wenn es nicht verwendet wird.
      ; <- *fRequestEnabled:  True wenn das Terminierungszeichen beim Request vom Gerät beachtet werden muss.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            If (*iTerminationChar And *fRequestEnabled)
               *iTerminationChar\i = \iTerminationChar
               *fRequestEnabled\i = \fTerminationCharRequestEnabled
               fReturn = #True
            EndIf
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SetTerminationChar(iInstanceNumber.i, iTerminationChar.i, fRequestEnabled.i, *sError.String)
      ; Setzt das Terminierungszeichen, mit dem die gesendeten und empfangenen Daten abgeschlossen werden.
      ; Wenn das Terminierungszeichen nicht für die Daten verwendet werden soll, muss Null übergeben werden.
      ; Wenn das Terminierungszeichen bei einem Request im MessageHeader verwendet werden soll,
      ; muss für fRequestEnabled #True übergeben werden.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> iTerminationChar:  Das Terminierungszeichen oder Null wenn es nicht verwendet wird.
      ; -> fRequestEnabled:   True wenn das Terminierungszeichen beim Request vom Gerät beachtet werden muss.
      ;                       Das TermChar wird dann im MessageHeader mitgesendet, und das entsprechende Flag dafür gesetzt.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            \iTerminationChar = iTerminationChar
            \fTerminationCharRequestEnabled = fRequestEnabled
         EndWith
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i GetTransferSplitDataLength(iInstanceNumber.i, *iTransferSplitDataLength.Integer, *sError.String)
      ; Gibt die Länge der Daten zurück, in die sie beim Senden in einzelne Transfer Messages aufgeteilt werden.
      ; -> iInstanceNumber:            Die Nummer der Instanz.
      ; <- *iTransferSplitDataLength:  Bei Null werden die Datein in einer einzelnen Transfer Message gesendet.
      ; <- sError:                     Fehlerbeschreibung.
      ; Rückgabe:                      True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            If *iTransferSplitDataLength
               *iTransferSplitDataLength\i = \iTransferSplitDataLength
               fReturn = #True
            EndIf
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SetTransferSplitDataLength(iInstanceNumber.i, iTransferSplitDataLength.i, *sError.String)
      ; Setzt die Länge der Daten, in die sie beim Senden in mehrere Transfer Messages aufgeteilt werden.
      ; Die Länge muss ein Vielfaches von 4 für die Übertragung sein. Ein falscher Wert wird automatisch abgerundet.
      ; Wenn Null übergeben wird, werden die Daten in einer einzelnen Transfer Message gesendet.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> iTransferDataLength:  Die Länge der Daten, Null oder ein Vielfaches von 4.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True bei Erfolg oder False bei einem Fehler.
      ;
      ; Hinweis:  UsbTmc - Seite 6 (PDF 12) - 3.2 Bulk-OUT endpoint, Regel Nr 5.
      ; The Host must send a complete USBTMC command message with a single transfer. This is illustrated
      ; below in Figure 2. If the Host fails to do so, the device must halt the Bulk-OUT endpoint. The only
      ; exception is if, in the specification of a particular USBTMC command message, explicit permission is
      ; given to send the command message with multiple transfers.
      ;
      ; Die Rigol Geräte empfangen einzelne Kommandos mit mehreren Transfers von nur wenigen Bytes ohne Probleme.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, iRest.i
      
      ; Die Länge auf ein Vielfaches von 4 abrunden.
      iRest = iTransferSplitDataLength % 4
      If iRest
         iTransferSplitDataLength - iRest
      EndIf
      
      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            \iTransferSplitDataLength = iTransferSplitDataLength
         EndWith
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)
      
      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure

   ;- --
   
   Procedure.i ReadShortPacketFromBulkInEndpoint(iInstanceNumber.i, iTimeout.i, *sError.String)
      ; Liest Daten vom Bulk-In-Endpoint ein, bis ein 'Short Packet' empfangen wird.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> iTimeout:          Die Zeit in Millisekunden für die Zeitüberschreitung. Bei Null wird die Standardzeit verwendet.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, iEndpointBulkIn.i, iMaxPacketSize.i, iTransferred.i, iRepeatCounter.i
      Protected *RequestBuffer, *uDeviceHandle
      #iAbortRequestBufferSize = $120000
      #iRepeatCounterMax = 10
      
      ; Das Gerät ist schon gesperrt per LockDevice().
      
      ; Den Bulk-In-Endpoint und seine Größe ermitteln.
      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            *uDeviceHandle = \uDevice\uDeviceHandle
            iEndpointBulkIn = \uDevice\iEndpointBulkIn
            If (iTimeout = #Null)
               iTimeout = \iTimeoutControlTransferMilliseconds
            EndIf
            iMaxPacketSize = \uDevice\iEndpointBulkInMaxPacketSize
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)
      
      ; Speicher reservieren.
      *RequestBuffer = AllocateMemory(#iAbortRequestBufferSize)
      fResult = Bool(*RequestBuffer)
      If (Not fResult)
         ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
         DebugView::Error("Der Speicher konnte nicht reserviert werden.", iInstanceNumber, miThreadedThreadCounter)
         
      Else
         ; Die Daten einlesen bis ein 'Short Packet' empfangen wurde.
         ; Die empfangenen Daten müssen weniger sein, als der Bulk-In-Endpoint groß ist.
         Repeat
            fResult = LibUsb::BulkTransfer(*uDeviceHandle, iEndpointBulkIn, *RequestBuffer,
                                           #iAbortRequestBufferSize, @iTransferred, iTimeout, *sError)
            If (Not fResult)
               *sError\s = CurrentProcedureNameInstance + "Es konnte kein 'Short Packet' eingelesen werden."
               DebugView::Error("Es konnte kein 'Short Packet' eingelesen werden.",
                                iInstanceNumber, miThreadedThreadCounter)
               iTransferred = 0  ; Repeat-Schleife beenden.
            Else
               fReturn = #True   ; Erfolg zurückgeben.
            EndIf
         Until (iTransferred < iMaxPacketSize)
         
         FreeMemory(*RequestBuffer)
      EndIf
  
      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure
     
   
   Procedure.i InitiateClearDevice(iInstanceNumber.i, *sError.String)
      ; Die Buffer der Bulk-Out und Bulk-In Endpoints im Gerät löschen.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True bei Erfolg oder False bei einem Fehler.
      ;
      ; Ablauf für das Löschen, siehe USBTMC Specification Rev1.0 - PDF Page 32, 4.2.1.6 INITIATE_CLEAR.
      ;   INITIATE_CLEAR
      ;    	 - STATUS_SUCCESS
      ;         -> CHECK_CLEAR_STATUS
      ;    		    - STATUS_PENDING
      ;    			    -> If bmClear.D0 = 1, the Host should Read from the Bulk-IN endpoint Until a short packet is received.
      ;    			    -> The Host must send CHECK_CLEAR_STATUS at a later time.
      ;    			    -> CHECK_CLEAR_STATUS wiederholen.
      ;    
      ;    		    - STATUS_SUCCESS
      ;    			    -> The Host must send a CLEAR_FEATURE request To clear the Bulk-OUT Halt.
      ;    
      ;    		    - All other values
      ;    			    -> The Host must send a CLEAR_FEATURE request To clear the Bulk-OUT Halt.
      ;          
      ;      - All other values (The device must Not Halt the Bulk-OUT endpoint.)
      ;    	    -> Fertig - The Host must Not send CHECK_CLEAR_STATUS.
      ;
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, fDeviceLocked.i, iRepeatCounter.i
      Protected uInitiateClearResponse.uInitiateClearResponse
      Protected uCheckClearStatusResponse.uCheckClearStatusResponse
      Protected iEndpointBulkOut.i, iTimeout.i
      Protected *uDeviceHandle
      #iRepeatCounterMax = 10
      #iPendingWaitTime = 50  ; Millisekunden.
      #iTimeoutReadShortPacket = 500   ; Millisekunden.
      
      ; INFO - Mit dem Rigol MSO2000 bleibt diese Prozedur beim Aufruf von CheckClearStatus() im Modul LibUsb bei
      ; INFO - libusb_control_transfer() hängen. Der Timeout hat keinen Einfluss.
      ; INFO - Getestet mit Firmware v00.03.05.03.03.
      
      ; Versuchen das Gerät zu sperren.
      fDeviceLocked = LockDevice(iInstanceNumber, *sError)
      
      ; Die Buffer der Bulk-Out und Bulk-In Endpoints löschen.
      fResult = InitiateClear(iInstanceNumber, @uInitiateClearResponse, *sError)
      If (Not fResult)
         ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
         DebugView::Error("Die Buffer der Bulk-Out und Bulk-In Endpoints konnten nicht gelöscht werden.",
                          iInstanceNumber, miThreadedThreadCounter)
      EndIf 
      
      If fResult
         fResult = Bool(uInitiateClearResponse\eUsbTmcStatus = #eUsbTmcStatusSuccess)
         If (Not fResult)
            ; Ein unbekannter Fehler ist aufgetreten.
            ; Siehe dazu USBTMC Specification Rev1.0 - PDF Page 25, Table 16 -- USBTMC_status values.
            *sError\s = CurrentProcedureNameInstance + "Es ist ein unbekannter Fehler aufgetreten. " +
                        "uInitiateClearResponse\eUsbTmcStatus = " + uInitiateClearResponse\eUsbTmcStatus
            DebugView::Error("Es ist ein unbekannter Fehler aufgetreten. uInitiateClearResponse\eUsbTmcStatus = " +
                             uInitiateClearResponse\eUsbTmcStatus, iInstanceNumber, miThreadedThreadCounter)
         EndIf
      EndIf
      
      If fResult
         ; Prüfen ob das Gerät das Löschen der Endpoints bearbeitet hat.
         iRepeatCounter = 0
         Repeat
            fResult = CheckClearStatus(iInstanceNumber, @uCheckClearStatusResponse, *sError)
            If (Not fResult)
               ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
               DebugView::Error("Der Clear Status konnte nicht ermittelt werden.",
                                iInstanceNumber, miThreadedThreadCounter)
               ; Repeat-Schleife beenden.
            EndIf
               
            If fResult   
               If (uCheckClearStatusResponse\eUsbTmcStatus = #eUsbTmcStatusPending)
                  ; Den Status erneut abfragen.
                  iRepeatCounter + 1
                  If (iRepeatCounter = #iRepeatCounterMax)
                     *sError\s = CurrentProcedureNameInstance +
                                 "Der Status des Löschen der Endpoints wurde nicht vom Gerät bestätigt (Status Pending)."
                     DebugView::Error("Der Status des Löschen der Endpoints wurde nicht vom Gerät bestätigt (Status Pending).",
                                      iInstanceNumber, miThreadedThreadCounter)
                     fResult = #False  ; Repeat-Schleife beenden.   
                     
                  Else
                     ; Wenn noch Daten im Bulk-In-Endpoint sind, die Daten einlesen.
                     ; Bit D0 ist dann gesetzt, die anderen Bits müssen immer gelöscht sein.
                     If (uCheckClearStatusResponse\iClearBitmap & %00000001)
                        fResult = ReadShortPacketFromBulkInEndpoint(iInstanceNumber, #iTimeoutReadShortPacket, *sError)
                     Else   
                        Delay(#iPendingWaitTime)   ; Kurze Zeit warten.
                     EndIf
                     ; Repeat-Schleife wiederholen.   
                  EndIf
                  
               ElseIf (uCheckClearStatusResponse\eUsbTmcStatus = #eUsbTmcStatusSuccess)
                  fReturn = #True   ; Erfolg zurückgeben.
                  fResult = #False  ; Repeat-Schleife beenden.
                  
               Else
                  ; Ein unbekannter Fehler ist aufgetreten.
                  ; Siehe dazu USBTMC Specification Rev1.0 - PDF Page 25, Table 16 -- USBTMC_status values.
                  *sError\s = CurrentProcedureNameInstance + "Es ist ein unbekannter Fehler aufgetreten. " +
                              "uCheckClearStatusResponse\eUsbTmcStatus = " + uCheckClearStatusResponse\eUsbTmcStatus
                  DebugView::Error("Es ist ein unbekannter Fehler aufgetreten. uCheckClearStatusResponse\eUsbTmcStatus = " +
                                   uCheckClearStatusResponse\eUsbTmcStatus, iInstanceNumber, miThreadedThreadCounter)
                  fResult = #False  ; Repeat-Schleife beenden.
               EndIf
            EndIf
         Until (fResult = #False)
         
         ; Einen CLEAR_FEATURE Request senden, damit der Halt-Status des Bulk-Out-Endpoint gelöscht wird.
         ; Die Bulk-Out-Endpoint ermitteln.
         If LockInstance(iInstanceNumber, *sError)
            With mInstanceMap()
               *uDeviceHandle = \uDevice\uDeviceHandle
               iEndpointBulkOut = \uDevice\iEndpointBulkOut
               iTimeout = \iTimeoutControlTransferMilliseconds
            EndWith
         EndIf
         UnlockInstance(iInstanceNumber)
         
         fResult = LibUsb::ClearHalt(*uDeviceHandle, iEndpointBulkOut, *sError)
         If (Not fResult)
            ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
            DebugView::Error("Der Halt-Status des Bulk-Out-Endpoint konnte nicht gelöscht werden.",
                             iInstanceNumber, miThreadedThreadCounter) 
            fReturn = #False  ; Fehler zurückgeben.
         EndIf
      EndIf
         
      ; Gerät wieder freigeben.
      If fDeviceLocked
         UnlockDevice(iInstanceNumber)
      EndIf
      
      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure
 
   
   Procedure.i AbortSendData(iInstanceNumber.i, iTransferIdentifier.i, *sError.String)
      ; Bricht die aktuelle USBTMC Command Message ab.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> iTransferIdentifier:  Der Transfer Identifier der aktuellen Transfer Message die abgebrochen werden soll.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True bei Erfolg oder False bei einem Fehler.
      ;
      ; Ablauf für das Abbrechen, siehe USBTMC Specification Rev1.0 - PDF Page 27, 4.2.1.2 INITIATE_ABORT_BULK_OUT.
      ;   INITIATE_ABORT_BULK_OUT
      ;    	- STATUS_SUCCESS (The device must Halt the Bulk-OUT endpoint.)
      ;    		 -> CHECK_ABORT_BULK_OUT_STATUS
      ;    			 - STATUS_PENDING
      ;    				 -> CHECK_ABORT_BULK_OUT_STATUS wiederholen.
      ;    
      ;    			 - STATUS_SUCCESS
      ;    				 -> CLEAR_FEATURE senden.
      ;    
      ;    			 - All other values
      ;    				 -> CLEAR_FEATURE senden.
      ; 
      ; 	   - STATUS_TRANSFER_NOT_IN_PROGRESS (The device must not Halt the Bulk-OUT endpoint.)
      ; 		   -> INITIATE_ABORT_BULK_OUT wiederholen
      ; 
      ; 	   - STATUS_FAILED (The device must not Halt the Bulk-OUT endpoint. - Problem bei MSO2000?)
      ; 		   -> Fertig - CHECK_ABORT_BULK_OUT_STATUS nicht senden!
      ; 
      ; 	   - All other values
      ; 		   ->Fertig - CHECK_ABORT_BULK_OUT_STATUS nicht senden!
      ;
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, fDeviceLocked.i, iRepeatCounter.i
      Protected fCheckAbortBulkOutStatus.i
      Protected uInitiateAbortBulkOutInResponse.uInitiateAbortBulkOutInResponse
      Protected uCheckAbortBulkOutStatusResponse.uCheckAbortBulkOutStatusResponse
      Protected iEndpointBulkOut.i, iTimeout.i
      Protected *uDeviceHandle
      #iRepeatCounterMax = 10
      #iPendingWaitTime = 50  ; Millisekunden.
      #iTimeoutReadShortPacket = 500   ; Millisekunden.
      
      ; Versuchen das Gerät zu sperren.
      fDeviceLocked = LockDevice(iInstanceNumber, *sError)
      
      iRepeatCounter = 0
      Repeat
         ; Den Bulk-Out-Tansfer abbrechen.
         fResult = InitiateAbortBulkOut(iInstanceNumber, iTransferIdentifier, @uInitiateAbortBulkOutInResponse, *sError)
         If (Not fResult)
            ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
            DebugView::Error("Der Bulk-Out-Transfer konnte nicht abgebrochen werden.", iInstanceNumber, miThreadedThreadCounter)
         EndIf 
         
         If fResult
            If (uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusFailed)
               ; Es ist kein Transfer am Laufen und der Bulk-Out-Endpoint ist leer.
               DebugView::Info("Es ist kein Transfer am Laufen und der Bulk-Out-Endpoint ist leer. " +
                               "uInitiateAbortBulkOutInResponse\iTag = " + uInitiateAbortBulkOutInResponse\iTag,
                               iInstanceNumber, miThreadedThreadCounter)
               fResult = #False
               fReturn = #True   ; Erfolg zurückgeben.
               
            ElseIf (uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusTransferNotInProgress)
               ; Es ist ein Transfer mit einem anderen Transfer Identifier am Laufen.
               ; Es ist kein Transfer mit dem Transfer Identifier am Laufen, aber der Bulk-Out-Endpoint ist nicht leer.
               DebugView::Info("Es ist ein Transfer mit einem anderen Transfer Identifier am Laufen, oder es ist kein",
                               iInstanceNumber, miThreadedThreadCounter)
               DebugView::Info("Transfer mit dem Transfer Identifier am Laufen, aber der Bulk-Out-Endpoint ist nicht leer. " +
                               "uInitiateAbortBulkOutInResponse\iTag = " + uInitiateAbortBulkOutInResponse\iTag,
                               iInstanceNumber, miThreadedThreadCounter)
               
               iRepeatCounter + 1
               If (iRepeatCounter = #iRepeatCounterMax)
                  *sError\s = CurrentProcedureNameInstance +
                              "Der Status des Abbruchs des Bulk-Out-Transfers wurde nicht vom Gerät bestätigt."
                  DebugView::Error("Der Status des Abbruchs des Bulk-Out-Transfers wurde nicht vom Gerät bestätigt. " +
                                   "uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusTransferNotInProgress",
                                   iInstanceNumber, miThreadedThreadCounter)
                  fResult = #False  ; Repeat-Schleife beenden.   
               Else
                  Delay(#iPendingWaitTime)   ; Kurze Zeit warten.
               EndIf
               
            ElseIf (uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusSuccess)
               ; Der Transfer wird abgebrochen, weiter mit CheckAbortBulkOutStatus().
               DebugView::Info("Der Transfer wird abgebrochen, weiter mit CheckAbortBulkOutStatus(). " +
                               "uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusSuccess",
                               iInstanceNumber, miThreadedThreadCounter)
               fResult = #False
               fCheckAbortBulkOutStatus = #True
               
            Else
               ; Ein unbekannter Fehler ist aufgetreten.
               ; Siehe dazu USBTMC Specification Rev1.0 - PDF Page 25, Table 16 -- USBTMC_status values.
               *sError\s = CurrentProcedureNameInstance + "Es ist ein unbekannter Fehler aufgetreten. " +
                           "uInitiateAbortBulkOutInResponse\eUsbTmcStatus = " + uInitiateAbortBulkOutInResponse\eUsbTmcStatus
               DebugView::Error("Es ist ein unbekannter Fehler aufgetreten. uInitiateAbortBulkOutInResponse\eUsbTmcStatus = " +
                                uInitiateAbortBulkOutInResponse\eUsbTmcStatus, iInstanceNumber, miThreadedThreadCounter)
               fResult = #False
               
            EndIf
         EndIf
      Until (fResult = #False)
      
      ; Prüfen ob das Gerät den Abbruch des Transfers bearbeitet hat.
      If fCheckAbortBulkOutStatus
         iRepeatCounter = 0
         Repeat
            fResult = CheckAbortBulkOutStatus(iInstanceNumber, @uCheckAbortBulkOutStatusResponse, *sError)
            If (Not fResult)
               ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
               DebugView::Error("Der Status des Abbruchs des Bulk-Out-Transfers konnte nicht ermittelt werden.",
                                iInstanceNumber, miThreadedThreadCounter)
               ; Repeat-Schleife beenden.
            EndIf
            
            If fResult
               If (uCheckAbortBulkOutStatusResponse\eUsbTmcStatus = #eUsbTmcStatusPending)
                  ; Der Transfer wurde noch nicht abgebrochen.
                  DebugView::Info("Der Transfer wurde noch nicht abgebrochen. " +
                                  "uCheckAbortBulkOutStatusResponse\eUsbTmcStatus = #eUsbTmcStatusPending",
                                   iInstanceNumber, miThreadedThreadCounter)
                  ; Den Status erneut abfragen.
                  iRepeatCounter + 1
                  If (iRepeatCounter = #iRepeatCounterMax)
                     *sError\s = CurrentProcedureNameInstance +
                                 "Der Status des Abbruchs des Bulk-Out-Transfers wurde nicht vom Gerät bestätigt."
                     DebugView::Error("Der Status des Abbruchs des Bulk-Out-Transfers wurde nicht vom Gerät bestätigt. " +
                                      "uCheckAbortBulkOutStatusResponse\eUsbTmcStatus = #eUsbTmcStatusPending",
                                      iInstanceNumber, miThreadedThreadCounter)
                     fResult = #False  ; Repeat-Schleife beenden.   
                  Else
                     Delay(#iPendingWaitTime)   ; Kurze Zeit warten.
                  EndIf
                  
               ElseIf (uCheckAbortBulkOutStatusResponse\eUsbTmcStatus = #eUsbTmcStatusSuccess)
                  fResult = #False  ; Repeat-Schleife beenden.
                  fReturn = #True   ; Erfolg zurückgeben.
                  
               Else   
                  ; Ein unbekannter Fehler ist aufgetreten.
                  ; Siehe dazu USBTMC Specification Rev1.0 - PDF Page 25, Table 16 -- USBTMC_status values.
                  *sError\s = CurrentProcedureNameInstance + "Es ist ein unbekannter Fehler aufgetreten. " +
                              "uCheckAbortBulkOutStatusResponse\eUsbTmcStatus = " + uCheckAbortBulkOutStatusResponse\eUsbTmcStatus
                  DebugView::Error("Es ist ein unbekannter Fehler aufgetreten. uCheckAbortBulkOutStatusResponse\eUsbTmcStatus = " +
                                   uCheckAbortBulkOutStatusResponse\eUsbTmcStatus, iInstanceNumber, miThreadedThreadCounter)
                  fResult = #False  ; Repeat-Schleife beenden.
               EndIf
            EndIf
         Until (fResult = #False)
              
         ; Einen CLEAR_FEATURE Request senden, damit der Halt-Status des Bulk-Out-Endpoint gelöscht wird.
         ; Die Bulk-Out-Endpoint ermitteln.
         If LockInstance(iInstanceNumber, *sError)
            With mInstanceMap()
               *uDeviceHandle = \uDevice\uDeviceHandle
               iEndpointBulkOut = \uDevice\iEndpointBulkOut
               iTimeout = \iTimeoutControlTransferMilliseconds
            EndWith
         EndIf
         UnlockInstance(iInstanceNumber)
         
         fResult = LibUsb::ClearHalt(*uDeviceHandle, iEndpointBulkOut, *sError)
         If (Not fResult)
            ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
            DebugView::Error("Der Halt-Status des Bulk-Out-Endpoint konnte nicht gelöscht werden.",
                             iInstanceNumber, miThreadedThreadCounter) 
            fReturn = #False  ; Fehler zurückgeben.
         EndIf
      EndIf

      ; Gerät wieder freigeben.
      If fDeviceLocked
         UnlockDevice(iInstanceNumber)
      EndIf
      
      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure
   
   
   Procedure.i AbortRequestData(iInstanceNumber.i, iTransferIdentifier.i, *sError.String)
      ; Bricht den aktuellen Bulk-In-Transfer ab.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> iTransferIdentifier:  Der Transfer Identifier der aktuellen Transfer Message die abgebrochen werden soll.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True bei Erfolg oder False bei einem Fehler.
      ;
      ; Ablauf für das Abbrechen, siehe USBTMC Specification Rev1.0 - PDF Page 29, 4.2.1.4 INITIATE_ABORT_BULK_IN.
      ;   INITIATE_ABORT_BULK_IN
      ;    	- STATUS_SUCCESS
      ;    	   -> The Host should Continue reading from the Bulk-IN endpoint Until a short packet is received.
      ;    	   -> The Host, after a short packet is received, must send CHECK_ABORT_BULK_IN_STATUS.
      ;    			- STATUS_PENDING
      ;    			   -> If bmAbortBulkIn.D0 = 1, the Host should Read from the Bulk-IN endpoint Until a short packet is received. 
      ;    				-> CHECK_ABORT_BULK_IN_STATUS wiederholen.
      ;    
      ;    			- STATUS_SUCCESS
      ;    				-> Fertig - CHECK_ABORT_BULK_IN_STATUS nicht senden!
      ;    
      ;    			- All other values
      ;    				-> Fertig - CHECK_ABORT_BULK_IN_STATUS nicht senden!
      ;    
      ;    	- STATUS_TRANSFER_NOT_IN_PROGRESS (The Host must Not send CHECK_ABORT_BULK_IN_STATUS.)
      ;        -> INITIATE_ABORT_BULK_IN wiederholen.
      ;          
      ;    	- STATUS_FAILED
      ;    	   -> Fertig - CHECK_ABORT_BULK_IN_STATUS nicht senden!
      ;    
      ;    	- All other values
      ;    	   -> Fertig - CHECK_ABORT_BULK_IN_STATUS nicht senden!
      
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, fDeviceLocked.i, fCheckAbortBulkInStatus.i, iRepeatCounter.i
      Protected uInitiateAbortBulkOutInResponse.uInitiateAbortBulkOutInResponse
      Protected uCheckAbortBulkInStatusResponse.uCheckAbortBulkInStatusResponse
      #iRepeatCounterMax = 10
      #iPendingWaitTime = 50  ; Millisekunden.
      #iTimeoutReadShortPacket = 500   ; Millisekunden.

      ; Versuchen das Gerät zu sperren.
      fDeviceLocked = LockDevice(iInstanceNumber, *sError)
      
      iRepeatCounter = 0
      Repeat
         ; Den Bulk-In-Tansfer abbrechen.
         fResult = InitiateAbortBulkIn(iInstanceNumber, iTransferIdentifier, @uInitiateAbortBulkOutInResponse, *sError)
         If (Not fResult)
            ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
            DebugView::Error("Der Bulk-In-Transfer konnte nicht abgebrochen werden.", iInstanceNumber, miThreadedThreadCounter)
         EndIf
      
         If fResult
            If (uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusFailed)
               ; Es ist kein Transfer am Laufen und der Bulk-Out-Endpoint ist leer.
               DebugView::Info("Es ist kein Transfer am Laufen und der Bulk-Out-Endpoint ist leer.",
                               iInstanceNumber, miThreadedThreadCounter)
               fReturn = #True   ; Erfolg zurückgeben.
               fResult = #False
               
            ElseIf (uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusTransferNotInProgress)
               ; Es ist ein Transfer mit einem anderen Transfer Identifier am Laufen.
               ; Es ist kein Transfer mit dem Transfer Identifier am Laufen, aber der Bulk-Out-Endpoint ist nicht leer.
               DebugView::Info("Es ist ein Transfer mit einem anderen Transfer Identifier am Laufen, oder es ist kein",
                                iInstanceNumber, miThreadedThreadCounter)
               DebugView::Info("Transfer mit dem Transfer Identifier am Laufen, aber der Bulk-Out-Endpoint ist nicht leer.",
                                iInstanceNumber, miThreadedThreadCounter)
               
               iRepeatCounter + 1
               If (iRepeatCounter = #iRepeatCounterMax)
                  *sError\s = CurrentProcedureNameInstance +
                              "Der Status des Abbruchs des Bulk-In-Transfers wurde nicht vom Gerät bestätigt."
                  DebugView::Error("Der Status des Abbruchs des Bulk-In-Transfers wurde nicht vom Gerät bestätigt. " +
                                   "uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusTransferNotInProgress",
                                   iInstanceNumber, miThreadedThreadCounter)
                  fResult = #False  ; Repeat-Schleife beenden.   
               Else
                  Delay(#iPendingWaitTime)   ; Kurze Zeit warten.
               EndIf
               
            ElseIf (uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusSuccess)
               ; Der Transfer wird abgebrochen, weiter mit ReadShortPacketFromBulkInEndpoint() und CheckAbortBulkInStatus().
               DebugView::Info("Der Transfer wird abgebrochen, weiter mit ReadShortPacketFromBulkInEndpoint() und " +
                               "CheckAbortBulkInStatus(). uInitiateAbortBulkOutInResponse\eUsbTmcStatus = #eUsbTmcStatusSuccess",
                               iInstanceNumber, miThreadedThreadCounter)
               fResult = #False
               fCheckAbortBulkInStatus = #True
               
            Else
               ; Ein unbekannter Fehler ist aufgetreten.
               ; Siehe dazu USBTMC Specification Rev1.0 - PDF Page 25, Table 16 -- USBTMC_status values.
               *sError\s = CurrentProcedureNameInstance + "Es ist ein unbekannter Fehler aufgetreten. " +
                           "uInitiateAbortBulkOutInResponse\eUsbTmcStatus = " + uInitiateAbortBulkOutInResponse\eUsbTmcStatus
               DebugView::Error("Es ist ein unbekannter Fehler aufgetreten. uInitiateAbortBulkOutInResponse\eUsbTmcStatus = " +
                                uInitiateAbortBulkOutInResponse\eUsbTmcStatus, iInstanceNumber, miThreadedThreadCounter)
               fResult = #False
            EndIf
         EndIf
      Until (fResult = #False)
      
      ; Prüfen ob das Gerät den Abbruch des Transfers bearbeitet hat.
      If fCheckAbortBulkInStatus
         ; Vorher noch vorhandene Daten einlesen bis ein 'Short Packet' empfangen wurde.
         If fResult
            fResult = ReadShortPacketFromBulkInEndpoint(iInstanceNumber, #iTimeoutReadShortPacket, *sError)
         EndIf
         
         iRepeatCounter = 0
         Repeat
            fResult = CheckAbortBulkInStatus(iInstanceNumber, @uCheckAbortBulkInStatusResponse, *sError)
            If (Not fResult)
               ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
               DebugView::Error("Der Status des Abbruchs des Bulk-In-Transfers konnte nicht ermittelt werden.",
                                iInstanceNumber, miThreadedThreadCounter)
               ; Repeat-Schleife beenden.
            EndIf
            
            If fResult
               If (uCheckAbortBulkInStatusResponse\eUsbTmcStatus = #eUsbTmcStatusPending)
                  ; Wenn noch Daten im Bulk-In-Endpoint sind, die Daten einlesen.
                  ; Bit D0 ist dann gesetzt, die anderen Bits müssen immer gelöscht sein.
                  If (uCheckAbortBulkInStatusResponse\iAbortBulkInBitmap & %00000001)
                     fResult = ReadShortPacketFromBulkInEndpoint(iInstanceNumber, #iTimeoutReadShortPacket, *sError)
                  EndIf
                  
                  ; Den Status nocheinmal abfragen.
                  iRepeatCounter + 1
                  If (iRepeatCounter = #iRepeatCounterMax)
                     *sError\s = CurrentProcedureNameInstance +
                                 "Der Status des Abbruchs des Bulk-In-Transfers wurde nicht vom Gerät bestätigt."
                     DebugView::Error("Der Status des Abbruchs des Bulk-In-Transfers wurde nicht vom Gerät bestätigt. " + 
                                      "uCheckAbortBulkInStatusResponse\eUsbTmcStatus = #eUsbTmcStatusPending",
                                      iInstanceNumber, miThreadedThreadCounter)
                     fResult = #False  ; Repeat-Schleife beenden.
                  Else
                     Delay(#iPendingWaitTime)   ; Kurze Zeit warten.
                  EndIf
                  
               ElseIf (uCheckAbortBulkInStatusResponse\eUsbTmcStatus = #eUsbTmcStatusSuccess)
                  ; Fertig und Erfolg zurückgeben.
                  fReturn = #True
                  fResult = #False  ; Repeat-Schleife beenden.
                  
               Else
                  ; Ein unbekannter Fehler ist aufgetreten.
                  ; Siehe dazu USBTMC Specification Rev1.0 - PDF Page 25, Table 16 -- USBTMC_status values.
                  *sError\s = CurrentProcedureNameInstance + "Es ist ein unbekannter Fehler aufgetreten. " +
                              "uCheckAbortBulkInStatusResponse\eUsbTmcStatus = " + uCheckAbortBulkInStatusResponse\eUsbTmcStatus
                  DebugView::Error("Es ist ein unbekannter Fehler aufgetreten. uCheckAbortBulkInStatusResponse\eUsbTmcStatus = " +
                                   uCheckAbortBulkInStatusResponse\eUsbTmcStatus, iInstanceNumber, miThreadedThreadCounter)
                  fResult = #False
               EndIf
            EndIf
         Until (fResult = #False)
      EndIf

      ; Gerät wieder freigeben.
      If fDeviceLocked
         UnlockDevice(iInstanceNumber)
      EndIf
      
      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure
   
   
   Procedure.i PrepareData(iInstanceNumber.i, sStringData.s, *BinaryData, iLengthBinaryData.i, *sError.String)
      ; Bereitet die Daten für das Senden vor. Die Daten werden kopiert und die Originale nicht verändert.
      ; Die binären Daten werden direkt an die Stringdaten angefügt.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> *sStringData:      Die Stringdaten wie SCPI-Kommandos.
      ; -> *BinaryData:       Ein Zeiger auf die binären Daten die gesendet werden.
      ; -> iLengthBinaryData: Die Länge der binären Daten.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die Daten vorbereitet wurden, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fError.i, iLength.i, iLengthStringData.i
      Protected *Data1, *Data2

      ; Vorhandene Daten freigeben.
      If (Not FreePreparedData(iInstanceNumber, *sError))
         ; Fehler, die Instanz ist nicht vorhanden.

      Else
         ; Daten prüfen.
         iLengthStringData = Len(sStringData)
         If Not ((iLengthStringData + iLengthBinaryData) > #Null)
            *sError\s = CurrentProcedureNameInstance + "Es sind keine Daten vorhanden."
            DebugView::Error("Es sind keine Daten vorhanden.", iInstanceNumber, miThreadedThreadCounter)

         Else
            ; Die String-Daten kopieren.
            ; Die Zeiger *Data1 und *Data2 werden abwechselnd verwendet beim Kopieren der Daten.
            If iLengthStringData
               *Data1 = Ascii(sStringData)
               ; Null-Zeichen am Ende des Strings entfernen.
               If *Data1
                  *Data2 = ReAllocateMemory(*Data1, MemoryStringLength(*Data1, #PB_Ascii))
                  If (Not *Data2)
                     fError = #True
                     *sError\s = CurrentProcedureNameInstance + "Der String-Speicher konnte nicht realloziert werden."
                     DebugView::Error("Der String-Speicher konnte nicht realloziert werden.",
                                      iInstanceNumber, miThreadedThreadCounter)
                     FreeMemory(*Data1)
                  Else
                     iLength = MemorySize(*Data2)
                  EndIf
               EndIf
            EndIf

            ; Die binären Daten hinter die String-Daten kopieren.
            If (Not fError)
               If (iLengthBinaryData And *BinaryData)
                  *Data1 = ReAllocateMemory(*Data2, iLength + iLengthBinaryData, #PB_Memory_NoClear)
                  If (Not *Data1)
                     fError = #True
                     *sError\s = CurrentProcedureNameInstance + "Der Binary-Speicher konnte nicht realloziert werden."
                     DebugView::Error("Der Binary-Speicher konnte nicht realloziert werden.",
                                      iInstanceNumber, miThreadedThreadCounter)
                     FreeMemory(*Data2)
                  Else
                     CopyMemory(*BinaryData, *Data1 + iLength, iLengthBinaryData)
                  EndIf
               EndIf
            EndIf

            ; Die vorbereiteten Daten sichern.
            If (Not fError)
               If (Not StorePreparedData(iInstanceNumber, *Data2, iLength + iLengthBinaryData, *sError))
                  FreeMemory(*Data2)
               Else
                  ; Daten vorbereitet und gesichert, Erfolg zurückgeben.
                  fReturn = #True
               EndIf
            EndIf
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SendData(iInstanceNumber.i, eMessageId.i, *iTransferred.Integer, *sError.String)
      ; Sendet die Daten, die zuvor mit der Funktion PrepareData() übergeben wurden, an das USB-Gerät.
      ; Die Daten werden in einer einzigen oder mehreren Transfer Messages gesendet, jenachdem ob mit der Prozedur
      ; SetTransferDataLength() ein Länge für die Datenpakete angegeben wurde.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> eMessageId:        Die USBTMC Message-ID für den Type des BulkOut-Headers.
      ; <- *iTransferred:     Die Anzahl der übertragenen Bytes.
      ;                       Wenn die Information nicht benötigt wird, kann hier Null übergeben werden.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die Daten gesendet wurden, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, fError.i, fDeviceLocked.i, iLength.i, iTransferred.i
      Protected fMultipleTransfers.i, fLastTransfer.i, *PreparedData, iTransferMessageCounter.i, iPreparedDataLength.i
      Protected *uInstance.uInstance, *BulkOutData, *uBulkOutHeader.uBulkHeader, iTransferMessageTag.i
      Protected *uDeviceHandle, iEventSendData.i
      
      ; Versuchen das Gerät zu sperren.
      fDeviceLocked = LockDevice(iInstanceNumber, *sError)
      fError = Bool(Not fDeviceLocked)
      
      ; Die vorbereiteten Daten lesen.
      If (Not fError)
         If (Not LockInstance(iInstanceNumber, *sError))
            ; Fehler, diese Instanz existiert nicht.
            fError = #True
            
         Else
            *uInstance = @mInstanceMap()
         EndIf
         UnlockInstance(iInstanceNumber)
      EndIf
      
      ; Prüfen ob das Gerät geöffnet ist.
      If (Not fError)
         With *uInstance
            If (Not \uDevice\uDeviceHandle)
               fError = #True
               *sError\s = CurrentProcedureNameInstance +
                           "Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0"
               DebugView::Error("Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0",
                                iInstanceNumber, miThreadedThreadCounter)
            EndIf
            
            ; Prüfen ob die Daten auf mehrere Transfer Messages aufgeteilt werden müssen.
            fMultipleTransfers = Bool(\iTransferSplitDataLength)
            fLastTransfer = Bool(Not fMultipleTransfers)
            
            ; Den USBTMC BulkOut-Header füllen und die vorbereiteten Nutzdaten anhängen.
            If (Not fError)
               Repeat
                  ; Die Daten aufteilen.
                  If fMultipleTransfers
                     ; Den Zeiger auf die vorbereiteten Daten für den nächsten Transfer anpassen.
                     *PreparedData = \PreparedData + (iTransferMessageCounter * \iTransferSplitDataLength)
                     iTransferMessageCounter + 1
                     ; Die Länge der gesamten Daten prüfen, ob ein Transfer ausreicht.
                     If (\iPreparedDataLength <= \iTransferSplitDataLength)
                        iPreparedDataLength = \iPreparedDataLength
                        fLastTransfer = #True
                     Else
                        ; Prüfen ob noch verbleibende Daten vorhanden sind.
                        If (\iPreparedDataLength <= (iTransferMessageCounter * \iTransferSplitDataLength))
                           iPreparedDataLength = \iPreparedDataLength - ((iTransferMessageCounter - 1) * \iTransferSplitDataLength)
                           fLastTransfer = #True
                         Else
                           iPreparedDataLength = \iTransferSplitDataLength
                        EndIf
                     EndIf
                     ; Den Message Header für einen weiteren oder den letzten Transfer erstellen.
                     CreateSendBulkOutHeaderWithData(iInstanceNumber, @*BulkOutData, *PreparedData, iPreparedDataLength,
                                                     eMessageId, fLastTransfer, *sError)
                     ; Den Tag der Message sichern falls die Kommunikation abgebrochen werden soll.
                     *uBulkOutHeader = *BulkOutData
                     iTransferMessageTag = *uBulkOutHeader\iTag
                     ; Anzahl der gesendeten Bytes sichern.
                     iEventSendData + iPreparedDataLength
                     
                  Else
                     ; Den Messge Header für einen einzelnen Transfer erstellen.
                     CreateSendBulkOutHeaderWithData(iInstanceNumber, @*BulkOutData, \PreparedData, \iPreparedDataLength,
                                                     eMessageId, #True, *sError)
                     ; Anzahl der gesendeten Bytes sichern.
                     iEventSendData + \iPreparedDataLength
                  EndIf
                  
                  ; Die Daten an das Gerät senden.
                  If (Not *BulkOutData)
                     *sError\s = CurrentProcedureNameInstance + "Der Zeiger auf die zu sendenden Daten ist Null. *BulkOutData = 0"
                     DebugView::Error("Der Zeiger auf die zu sendenden Daten ist Null. *BulkOutData = 0",
                                      iInstanceNumber, miThreadedThreadCounter)
                     fReturn = #False
                     fLastTransfer = #True
                     
                  Else
                     iLength = MemorySize(*BulkOutData)
                     fReturn = LibUsb::BulkTransfer(\uDevice\uDeviceHandle, \uDevice\iEndpointBulkOut, *BulkOutData, iLength,
                                                    *iTransferred, \iTimeoutMilliseconds, *sError)
                     ; Bei einem Fehler abbrechen.
                     If (Not fReturn)
                        fLastTransfer = #True
                     EndIf
                     ; Die gesendeten Daten wieder freigeben.
                     FreeMemory(*BulkOutData)
                     
                     ; Nur innerhalb eines Threads auf das Abbrechen prüfen und den Fortschritt senden.
                     If *uInstance\iThreadNumber
                        ; Nach dem Senden der Daten, prüfen ob die Kommunikation abgebrochen werden soll.
                        If (Not fLastTransfer)
                           If (\eThreadStatus = #eThreadStatusCancel) Or (\eThreadStatus = #eThreadStatusEnd)
                              fLastTransfer = #True
                              If (Not AbortSendData(iInstanceNumber, iTransferMessageTag, *sError.String))
                                 ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
                                 DebugView::Error("Es ist ein Fehler beim Abbrechen des Sendens aufgetreten.",
                                                  iInstanceNumber, miThreadedThreadCounter)
                              EndIf
                              *sError\s = CurrentProcedureNameInstance + "Der Thread für das Senden wurde abgebrochen." 
                              DebugView::Error("Der Thread für das Senden wurde abgebrochen.",
                                               iInstanceNumber, miThreadedThreadCounter)
                              fReturn = #False  ; Fehler für den Abbruch zurückgeben.
                           EndIf
                        EndIf
                        
                        ; Ein Ereignis für den Fortschritt der gesendeten Daten auslösen.
                        ; Das Terminierungszeichen wird bei der Anzahl der gesendeten Daten (iEventSendData) nicht mitgezählt.
                        If \fThreadRunning
                           PostEvent(#eCustomEventUsbTmcThreadData, \uThreadData\iEventListenerWindowNumber,
                                     iInstanceNumber, #eThreadEventSendProgress, iEventSendData)
                        EndIf
                     EndIf
                  EndIf
               Until (fLastTransfer = #True)
            EndIf
         EndWith
      EndIf
      
      ; Gerät wieder freigeben.
      If fDeviceLocked
         UnlockDevice(iInstanceNumber)
      EndIf
      
      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i RequestData_ALT_fuer_Mac(iInstanceNumber.i, *Data.Integer, *iTransferred.Integer, eMessageId.i, *sError.String)
      ; Holt Daten vom USB-Gerät ab.
      ; Die Daten werden ohne dem optionalen Terminierungszeichen zurückgegeben.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *Data:             Ein Zeiger auf die empfangenen Daten. Die Daten müssen wieder freigegeben werden.
      ; <- *iTransferred:     Die Anzahl der übertragenen Bytes.
      ; -> eMessageId:        Die USBTMC Message-ID für den Type des Request-BulkOut-Headers.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn Daten empfangen wurden, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, fError.i, fExit.i, fDeviceLocked.i
      Protected fStopRequest.i, fSendRequest.i, fResendRequest.i, fEndOfMessage.i, fMessageHeaderReceived.i
      Protected iDataSize.i, iDataWritePosition.i, iTransferred.i, iTransferredData.i, iPendingData.i, iReceivedData.i
      Protected iEndpointBulkOut, iEndpointBulkIn, iTimeout.i, iMessageIdAndTag.i, iTermCharLength.i
      Protected iDataOffset.i, iRequestBufferLength.i
      Protected *RequestBuffer, *uDeviceHandle, *DataReAllocate
      Protected uBulkOutHeader.uBulkHeader, *uBulkInHeader.uBulkHeader
      Protected iEventDataSize.i, iEventReceivedData.i, iEventListenerWindowNumber.i, fThread.i, fEventProgress.i
      ; Buffergröße ist etwas größer als die Bilddaten einer Bildschirmkopie der Rigol-Geräte, 1,1 MB.
      ; Größe um den der Empfangsspeicher jeweils vergrößert wird, 33 MB.
      #iDataSizeReAllocate = $2100000

      ; Versuchen das Gerät zu sperren.
      fDeviceLocked = LockDevice(iInstanceNumber, *sError)
      fError = Bool(Not fDeviceLocked)

      If (Not fError)
         If (Not *Data)
            fError = #True
            *sError\s = CurrentProcedureNameInstance + "Der Zeiger für die empfangenen Daten ist Null. *Data = 0"
            DebugView::Error("Der Zeiger für die empfangenen Daten ist Null. *Data = 0",
                             iInstanceNumber, miThreadedThreadCounter)
         ElseIf (Not *iTransferred)
            fError = #True
            *sError\s = CurrentProcedureNameInstance +
                        "Der Zeiger für die Anzahl der empfangenen Daten ist Null. *iTransferred = 0"
            DebugView::Error("Der Zeiger für die Anzahl der empfangenen Daten ist Null. *iTransferred = 0",
                             iInstanceNumber, miThreadedThreadCounter)
         EndIf

         ; Die benötigten Daten zum geöffneten Gerät einlesen.
         If (Not LockInstance(iInstanceNumber, *sError))
            fError = #True
            DebugView::Error("Diese Instanz existiert nicht.", iInstanceNumber, miThreadedThreadCounter)

         Else
            With mInstanceMap()
               *uDeviceHandle = \uDevice\uDeviceHandle
               iEndpointBulkOut = \uDevice\iEndpointBulkOut
               iEndpointBulkIn = \uDevice\iEndpointBulkIn
               iTimeout = \iTimeoutMilliseconds
               iTermCharLength = Bool(\iTerminationChar) ; Bei ungleich Null gibt Bool 1 für True zurück.
               fThread = Bool(\iThreadNumber)
               ; Um ein Ereignis für den Fortschritt auszulösen, muss diese Anzahl an Daten empfangen werden.
               ; Normalerweise wird aber immer die Buffergröße #iRequestBufferSize auf einmal empfangen.
               iEventDataSize = \iRequestProgressDataLength
               iRequestBufferLength = \iRequestBufferLength
            EndWith
         EndIf
         UnlockInstance(iInstanceNumber)

         ; Das Gerät muss geöffnet sein.
         If (Not *uDeviceHandle)
            fError = #True
            *sError\s = CurrentProcedureNameInstance +
                        "Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0"
            DebugView::Error("Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0",
                             iInstanceNumber, miThreadedThreadCounter)
         EndIf

         ; Den Speicher für die gesamten zu empfangenden Daten vorbereiten.
         If (Not fError)
            iDataSize = iRequestBufferLength
            iDataWritePosition = 0
            *Data\i = AllocateMemory(iDataSize, #PB_Memory_NoClear)
            If (Not *Data\i)
               fError = #True
               *sError\s = CurrentProcedureNameInstance +
                           "Der Speicher für die zu empfangenden Daten konnte nicht alloziert werden."
               DebugView::Error("Der Speicher für die zu empfangenden Daten konnte nicht alloziert werden.",
                                iInstanceNumber, miThreadedThreadCounter)
            EndIf
         EndIf

         ; Den Request-Speicher vorbereiten.
         If (Not fError)
            *RequestBuffer = AllocateMemory(SizeOf(uBulkHeader) + iRequestBufferLength, #PB_Memory_NoClear)
            If (Not *RequestBuffer)
               fError = #True
               *sError\s = CurrentProcedureNameInstance +
                           "Der Request-Speicher für die zu empfangenden Daten konnte nicht alloziert werden."
               DebugView::Error("Der Request-Speicher für die zu empfangenden Daten konnte nicht alloziert werden.",
                                iInstanceNumber, miThreadedThreadCounter)
            EndIf
         EndIf
      EndIf

      If (Not fError)
         fStopRequest = #False   ; Lesen vom Gerät noch nicht beenden.
         fSendRequest = #True    ; Eine Request-Message an das Gerät senden.

         ; Die Request-Bulk-In-Message wiederholt an das Gerät senden, bis alle Daten übertragen wurden.
         While (Not fStopRequest)

            ; Den USBTMC Request-BulkOut-Header erstellen.
            ; Es wird für jede Transfer Message wird im MessageHeader der Transfer Identifier inkrementiert.
            fResult = CreateRequestBulkOutHeader(iInstanceNumber, @uBulkOutHeader, iRequestBufferLength, eMessageId, *sError)
            If (Not fResult)
               fStopRequest = #True
               DebugView::Error("Der Request-Bulk-Out-Header konnte nicht erstellt werden.",
                                iInstanceNumber, miThreadedThreadCounter)

            Else
               ; Für den Vergleich ob der richtige MessageHeader vom Gerät empfangen wurde. Es wird damit auch geprüft
               ; ob überhaupt ein MessageHeader oder sonstige Daten empfangen wurden. Das ist leider je nach Gerät und
               ; Betriebssystem unterschiedlich. z.B. bei einem Rigol DS1000Z Oszilloskop werden unter Mac OS X einzelne
               ; 64 Byte Pakete empfangen. Unter Windows werden die ganzen Daten auf einmal empfangen.
               ;
               ; Die ersten vier Bytes vom MessageHeader sichern.
               ; Der zurückgesendete MessageHeader vom Gerät (Message-ID, Tag und Reserved) muss identisch sein.
               iMessageIdAndTag = PeekI(@uBulkOutHeader)
               CopyMemory(@uBulkOutHeader, *RequestBuffer, SizeOf(uBulkHeader))
            EndIf

            ; Die Request-Bulk-In-Message an das Gerät senden.
            If fResult
               fResult = LibUsb::BulkTransfer(*uDeviceHandle, iEndpointBulkOut, @uBulkOutHeader, SizeOf(uBulkHeader),
                                              #Null, iTimeout, *sError)
            EndIf
            ; Bei einem Fehler abbrechen.
            If (Not fResult)
               fStopRequest = #True
               DebugView::Error("Der Request-Bulk-In-Message konnte nicht gesendet werden.",
                                iInstanceNumber, miThreadedThreadCounter)
            EndIf

            ; Die Antwort vom Gerät einlesen.
            If fResult
               fResult = LibUsb::BulkTransfer(*uDeviceHandle, iEndpointBulkIn, *RequestBuffer,
                                              SizeOf(uBulkHeader) + iRequestBufferLength, @iTransferred, iTimeout, *sError)
               ; Bei einem Fehler abbrechen.
               If (Not fResult)
                  fStopRequest = #True
                  DebugView::Error("Die Antwort für die Request-Bulk-In-Message konnte nicht vom Gerät empfangen werden.",
                                   iInstanceNumber, miThreadedThreadCounter)
               EndIf

               ; Prüfen ob der richtige MessageHeader empfangen wurde, oder sonstige Daten.
               If fResult
                  If iTransferred
                     ; Die Message-ID nur prüfen wenn überhaupt ein MessageHeader vorhanden sein muss.
                     If fSendRequest
                        fMessageHeaderReceived = Bool(iMessageIdAndTag = PeekI(*RequestBuffer))
                        ; Wenn die IDs identisch sind, kann eine neue Request-BulkIn-Message an das Gerät gesendet werden,
                        ; wenn noch Daten ausständig sind. Ansonst wurden wahrscheinlich einzelne 64 Byte Pakete empfangen.
                        fSendRequest = fMessageHeaderReceived
                     Else
                        ; Der MessageHeader wurde nicht erwartet, die Message-ID braucht somit nicht überprüft werden.
                        fMessageHeaderReceived = #False
                     EndIf

                  Else
                     ; Die Message-IDs können nicht identisch sein, da keine Daten empfangen wurden.
                     fMessageHeaderReceived = #False
                     fSendRequest = #False

                     ; Es sollten alle Daten vorhanden sein, da ein leeres Transfer Message Packet empfangen wurde.
                     fReturn = #True
                     fStopRequest = #True
                  EndIf
               EndIf

               ; Die empfangenen Daten von dem MessageHeader trennen, und zu den schon empfangenen Daten kopieren.
               ; Den Datenspeicher gegebenfalls vergrößern und die benötigen Zähler der Daten berechnen.
               If fResult

                  ; Die Größe der empfangenen Nutzdaten aus dem MessageHeader auslesen.
                  If fMessageHeaderReceived
                     ; Mit Hilfe der uBulkHeader-Struktur auf die empfangenen Daten zugreifen.
                     *uBulkInHeader = *RequestBuffer
                     iTransferredData = *uBulkInHeader\iTransferSize
                     ; Die angegebene Datengröße der Daten die vom Gerät gesendet werden, für die weitere Berechnung sichern.
                     iPendingData = iTransferredData

                     ; Prüfen ob die letzte Transfer Message empfangen wurde. Das Bit 0 'End Of Message' (EOM) ist dann gesetzt.
                     fEndOfMessage = Bool(*uBulkInHeader\iTransferAttributes & #iTransferAttributeEndOfTransfer)

                     ; Prüfen ob weniger Daten empfangen wurden, als im MessageHeader angegeben ist.
                     If (iTransferredData > (iTransferred - SizeOf(uBulkHeader)))
                        ; Die tatsächliche Anzahl der vorhandenen Daten berechnen.
                        iTransferredData = (iTransferred - SizeOf(uBulkHeader))
                        ; Jetzt müssen die noch ausstehenden Daten, ohne das Senden einer neuen Request-Bulk-In-Message
                        ; eingelesen werden.
                        If (eMessageId = #eMessageIdDevDepMsgIn)
                           fSendRequest = #False
                        EndIf
                        
                     ElseIf (iTransferredData = (iTransferred - SizeOf(uBulkHeader)))
                        ; Eine Vendor Specific In Message besitzt kein 'End Of Message' Bit.
                        ; Wenn es nur eine Transfer Message gibt, sollten jetzt alle Daten vorhanden sein.
                        If (eMessageId = #eMessageIdVendorSpecificIn)
                           fEndOfMessage = #True
                           fReturn = #True
                           fStopRequest = #True
                        EndIf
                     EndIf

                  Else
                     ; Es ist kein MessageHeader vorhanden, es wurde anscheinend ein einzelnes 64 Byte Bulk-Transfer Packet
                     ; empfangen. Alles was empfangen wurde gehört zu den Daten.
                     iTransferredData = iTransferred
                  EndIf

                  ; Die noch verbleibenden Daten berechnen.
                  iPendingData - iTransferredData

                  ; Prüfen ob schon alle Daten empfangen wurden.
                  If (iPendingData <= 0)
                     ; Es wurden alle Daten empfangen, nur mehr die fehlenden Daten kopieren.
                     iTransferredData - Abs(iPendingData)
                  EndIf

                  ; Die aktuell empfangene Datengröße berechnen für z.B. die Weitergabe zur Statusanzeige.
                  iReceivedData + iTransferredData

                  ; Prüfen ob der Datenspeicher für die zusätzlichen Daten noch ausreicht.
                  If (fMessageHeaderReceived And (iTransferredData > 0))
                     If ((iDataWritePosition + iTransferredData) > iDataSize)
                        ; Speicherplatz für die Daten vergrößern.
                        *DataReAllocate = ReAllocateMemory(*Data\i, iDataSize + #iDataSizeReAllocate, #PB_Memory_NoClear)
                        If (Not *DataReAllocate)
                           ; Keine Daten zusammenkopieren, Empfang der Daten abbrechen.
                           fError = #True
                           fReturn = #False
                           fStopRequest = #True
                           *sError\s = CurrentProcedureNameInstance +
                                       "Der Speicher für die zu empfangenden Daten konnte nicht realloziert werden."
                           DebugView::Error("Der Speicher für die zu empfangenden Daten konnte nicht realloziert werden.",
                                            iInstanceNumber, miThreadedThreadCounter)
                           If *Data\i
                              FreeMemory(*Data\i)
                              *Data\i = #Null
                           EndIf
                        Else
                           iDataSize + #iDataSizeReAllocate
                           *Data\i = *DataReAllocate
                        EndIf
                     EndIf
                  EndIf

                  ; Die neuen Daten jetzt hinten an die vorhandenen Daten kopieren.
                  If ((Not fError) And (iTransferredData > 0))
                     ; Wenn ein MessageHeader vorhanden ist, diesen nicht mitkopieren.
                     If fMessageHeaderReceived
                        iDataOffset = SizeOf(uBulkHeader)
                     Else
                        iDataOffset = 0
                     EndIf
                     If (iTransferredData > 0)
                        CopyMemory(*RequestBuffer + iDataOffset, *Data\i + iDataWritePosition, iTransferredData)
                        ; Die neue Position für das Schreiben in den Datenspeicher berechnen.
                        iDataWritePosition + iTransferredData
                     EndIf
                  EndIf

                  ; Wenn weitere Daten erwartet werden, dass abschließende Short Packet vom Gerät ohne Daten einlesen.
                  If (Not fEndOfMessage)
                     If (Not fSendRequest)
                        ; Das abschließende Short Packet vom Gerät einlesen und verwerfen.
                        If (Not LibUsb::BulkTransfer(*uDeviceHandle, iEndpointBulkIn, *RequestBuffer, iRequestBufferLength,
                                                     #Null, iTimeout, *sError))
                           DebugView::Error("Das abschließende 'Short Packet' konnte nicht empfangen werden.",
                                            iInstanceNumber, miThreadedThreadCounter)
                        EndIf
                     EndIf
                     ; Für die noch ausstehenden Daten eine neue Request-BulkIn-Message senden.
                     fResendRequest = #True
                  Else
                     ; Keine weiteren Daten mehr empfangen.
                     fReturn = #True
                     fStopRequest = #True
                  EndIf

                  ; Prüfen ob alle Daten empfangen wurden.
                  If (fSendRequest And fEndOfMessage)
                     ; Wenn eine neue Request-BulkIn-Message gesendet werden soll, aber schon die letzte Message
                     ; empfangen wurde, sollten alle Daten vorhanden sein.
                     fReturn = #True
                     fStopRequest = #True
                  EndIf

                  ; Nur innerhalb eines Threads auf das Abbrechen prüfen und den Fortschritt senden.
                  If fThread
                     ; Nach dem Empfang der Daten, prüfen ob die Kommunikation abgebrochen werden soll.
                     ; Ansonst ein Ereignis für die Anzahl der empfangenen Daten senden.
                     If LockInstance(iInstanceNumber, *sError)
                        With mInstanceMap()
                           fExit = Bool((\eThreadStatus = #eThreadStatusCancel) Or (\eThreadStatus = #eThreadStatusEnd))
                           iEventListenerWindowNumber = \uThreadData\iEventListenerWindowNumber
                           ; Ereignisdaten für PostEvent.
                           If Not (fStopRequest Or fExit)
                              fEventProgress = Bool((iReceivedData - iEventReceivedData) >= iEventDataSize)
                           Else
                              fEventProgress = #False
                           EndIf
                        EndWith
                     EndIf
                     UnlockInstance(iInstanceNumber)

                     ; Kommunikation beenden wenn notwendig.
                     If (fExit And (Not fStopRequest))
                        fReturn = #False  ; Fehler für den Abbruch zurückgeben.
                        fStopRequest = #True
                        If (Not AbortRequestData(iInstanceNumber, uBulkOutHeader\iTag, *sError.String))
                           ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
                           DebugView::Error("Es ist ein Fehler beim Abbrechen des Empfangens aufgetreten.",
                                            iInstanceNumber, miThreadedThreadCounter)
                        EndIf
                        *sError\s = CurrentProcedureNameInstance + "Der Thread für das Empfangen wurde abgebrochen." 
                        DebugView::Error("Der Thread für das Empfangen wurde abgebrochen.",
                                         iInstanceNumber, miThreadedThreadCounter)
                        
                     ; Ereignis für den Fortschritt der empfangenen Daten auslösen.
                     ElseIf (Not fStopRequest)
                        If (fThread And fEventProgress)
                           iEventReceivedData = iReceivedData
                           PostEvent(#eCustomEventUsbTmcThreadData, iEventListenerWindowNumber,
                                     iInstanceNumber, #eThreadEventRequestProgress, iReceivedData)
                        EndIf
                     EndIf
                  EndIf
                     
               EndIf ; fResult. ; Die empfangenen Daten zusammenkopieren.

               ; Eine neue Request-BulkOut-Message, für die weiteren Daten an das Gerät senden wenn notwendig.
               If fResendRequest
                  fResendRequest = #False
                  fSendRequest = #True
               EndIf

            EndIf ; fResult. ; Die Anwort vom Gerät einlesen.
         Wend  ;  fStopRequest.
      EndIf ; (Not fError). ; Die Transfer Messages senden und empfangen bis alle Daten übertragen wurden.

      ; Den Datenspeicher an die empfangenen Daten anpassen und verkleinern.
      ; Das optionale Terminierungszeichen wird ebenfalls entfernt.
      If fReturn
         *DataReAllocate = ReAllocateMemory(*Data\i, iDataWritePosition - iTermCharLength, #PB_Memory_NoClear)
         If (Not *DataReAllocate)
            fReturn = #False
            *sError\s = CurrentProcedureNameInstance +
                        "Der Speicher mit den empfangenen Daten konnte nicht realloziert werden."
            DebugView::Error("Der Speicher mit den empfangenen Daten konnte nicht realloziert werden.",
                             iInstanceNumber, miThreadedThreadCounter)
            If *Data\i
               FreeMemory(*Data\i)
               *Data\i = #Null
            EndIf
         Else
            *Data\i = *DataReAllocate
            *iTransferred\i = iDataWritePosition - iTermCharLength
         EndIf
      EndIf

      ; Den Request-Speicher freigeben.
      If *RequestBuffer
         FreeMemory(*RequestBuffer)
      EndIf

      ; Gerät wieder freigeben.
      If fDeviceLocked
         UnlockDevice(iInstanceNumber)
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure

   
   Procedure.i RequestData(iInstanceNumber.i, *Data.Integer, *iTransferred.Integer, eMessageId.i, *sError.String)
      ; Holt Daten vom USB-Gerät ab.
      ; Die Daten werden ohne dem optionalen Terminierungszeichen zurückgegeben.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *Data:             Ein Zeiger auf die empfangenen Daten. Die Daten müssen wieder freigegeben werden.
      ; <- *iTransferred:     Die Anzahl der übertragenen Bytes.
      ; -> eMessageId:        Die USBTMC Message-ID für den Type des Request-BulkOut-Headers.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn Daten empfangen wurden, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, fError.i, fExit.i, fDeviceLocked.i
      Protected fStopRequest.i, fSendRequest.i, fResendRequest.i, fEndOfMessage.i, fMessageHeaderReceived.i
      Protected iDataSize.i, iDataWritePosition.i, iTransferred.i, iTransferredData.i, iPendingData.i, iReceivedData.i
      Protected iEndpointBulkOut, iEndpointBulkIn, iTimeout.i, iMessageIdAndTag.i, iTermCharLength.i
      Protected iDataOffset.i, iRequestBufferLength.i
      Protected *RequestBuffer, *uDeviceHandle, *DataReAllocate
      Protected uBulkOutHeader.uBulkHeader, *uBulkInHeader.uBulkHeader
      Protected iEventDataSize.i, iEventReceivedData.i, iEventListenerWindowNumber.i, fThread.i, fEventProgress.i
      ; Buffergröße ist etwas größer als die Bilddaten einer Bildschirmkopie der Rigol-Geräte, 1,1 MB.
      ; Größe um den der Empfangsspeicher jeweils vergrößert wird, 33 MB.
      #iDataSizeReAllocate = $2100000

      ; Versuchen das Gerät zu sperren.
      fDeviceLocked = LockDevice(iInstanceNumber, *sError)
      fError = Bool(Not fDeviceLocked)

      If (Not fError)
         If (Not *Data)
            fError = #True
            *sError\s = CurrentProcedureNameInstance + "Der Zeiger für die empfangenen Daten ist Null. *Data = 0"
            DebugView::Error("Der Zeiger für die empfangenen Daten ist Null. *Data = 0",
                             iInstanceNumber, miThreadedThreadCounter)
         ElseIf (Not *iTransferred)
            fError = #True
            *sError\s = CurrentProcedureNameInstance +
                        "Der Zeiger für die Anzahl der empfangenen Daten ist Null. *iTransferred = 0"
            DebugView::Error("Der Zeiger für die Anzahl der empfangenen Daten ist Null. *iTransferred = 0",
                             iInstanceNumber, miThreadedThreadCounter)
         EndIf

         ; Die benötigten Daten zum geöffneten Gerät einlesen.
         If (Not LockInstance(iInstanceNumber, *sError))
            fError = #True
            DebugView::Error("Diese Instanz existiert nicht.", iInstanceNumber, miThreadedThreadCounter)

         Else
            With mInstanceMap()
               *uDeviceHandle = \uDevice\uDeviceHandle
               iEndpointBulkOut = \uDevice\iEndpointBulkOut
               iEndpointBulkIn = \uDevice\iEndpointBulkIn
               iTimeout = \iTimeoutMilliseconds
               iTermCharLength = Bool(\iTerminationChar) ; Bei ungleich Null gibt Bool 1 für True zurück.
               fThread = Bool(\iThreadNumber)
               ; Um ein Ereignis für den Fortschritt auszulösen, muss diese Anzahl an Daten empfangen werden.
               ; Normalerweise wird aber immer die Buffergröße #iRequestBufferSize auf einmal empfangen.
               iEventDataSize = \iRequestProgressDataLength
               iRequestBufferLength = \iRequestBufferLength
            EndWith
         EndIf
         UnlockInstance(iInstanceNumber)

         ; Das Gerät muss geöffnet sein.
         If (Not *uDeviceHandle)
            fError = #True
            *sError\s = CurrentProcedureNameInstance +
                        "Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0"
            DebugView::Error("Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0",
                             iInstanceNumber, miThreadedThreadCounter)
         EndIf

         ; Den Speicher für die gesamten zu empfangenden Daten vorbereiten.
         If (Not fError)
            iDataSize = iRequestBufferLength
            iDataWritePosition = 0
            *Data\i = AllocateMemory(iDataSize, #PB_Memory_NoClear)
            If (Not *Data\i)
               fError = #True
               *sError\s = CurrentProcedureNameInstance +
                           "Der Speicher für die zu empfangenden Daten konnte nicht alloziert werden."
               DebugView::Error("Der Speicher für die zu empfangenden Daten konnte nicht alloziert werden.",
                                iInstanceNumber, miThreadedThreadCounter)
            EndIf
         EndIf

         ; Den Request-Speicher vorbereiten.
         If (Not fError)
            *RequestBuffer = AllocateMemory(SizeOf(uBulkHeader) + iRequestBufferLength, #PB_Memory_NoClear)
            If (Not *RequestBuffer)
               fError = #True
               *sError\s = CurrentProcedureNameInstance +
                           "Der Request-Speicher für die zu empfangenden Daten konnte nicht alloziert werden."
               DebugView::Error("Der Request-Speicher für die zu empfangenden Daten konnte nicht alloziert werden.",
                                iInstanceNumber, miThreadedThreadCounter)
            EndIf
         EndIf
      EndIf

      If (Not fError)
         fStopRequest = #False   ; Lesen vom Gerät noch nicht beenden.
         fSendRequest = #True    ; Eine Request-Message an das Gerät senden.

         ; Die Request-Bulk-In-Message wiederholt an das Gerät senden, bis alle Daten übertragen wurden.
         While (Not fStopRequest)

            ; Den USBTMC Request-BulkOut-Header erstellen.
            ; Es wird für jede Transfer Message wird im MessageHeader der Transfer Identifier inkrementiert.
            fResult = CreateRequestBulkOutHeader(iInstanceNumber, @uBulkOutHeader, iRequestBufferLength, eMessageId, *sError)
            If (Not fResult)
               fStopRequest = #True
               DebugView::Error("Der Request-Bulk-Out-Header konnte nicht erstellt werden.",
                                iInstanceNumber, miThreadedThreadCounter)

            Else
               ; Für den Vergleich ob der richtige MessageHeader vom Gerät empfangen wurde. Es wird damit auch geprüft
               ; ob überhaupt ein MessageHeader oder sonstige Daten empfangen wurden. Das ist leider je nach Gerät und
               ; Betriebssystem unterschiedlich. z.B. bei einem Rigol DS1000Z Oszilloskop werden unter Mac OS X einzelne
               ; 64 Byte Pakete empfangen. Unter Windows werden die ganzen Daten auf einmal empfangen.
               ;
               ; Die ersten vier Bytes vom MessageHeader sichern.
               ; Der zurückgesendete MessageHeader vom Gerät (Message-ID, Tag und Reserved) muss identisch sein.
               iMessageIdAndTag = PeekI(@uBulkOutHeader)
               CopyMemory(@uBulkOutHeader, *RequestBuffer, SizeOf(uBulkHeader))
            EndIf

            ; Die Request-Bulk-In-Message an das Gerät senden.
            If fResult
               fResult = LibUsb::BulkTransfer(*uDeviceHandle, iEndpointBulkOut, @uBulkOutHeader, SizeOf(uBulkHeader),
                                              #Null, iTimeout, *sError)
            EndIf
            ; Bei einem Fehler abbrechen.
            If (Not fResult)
               fStopRequest = #True
               DebugView::Error("Der Request-Bulk-In-Message konnte nicht gesendet werden.",
                                iInstanceNumber, miThreadedThreadCounter)
            EndIf

            ; Die Antwort vom Gerät einlesen.
            If fResult
               fResult = LibUsb::BulkTransfer(*uDeviceHandle, iEndpointBulkIn, *RequestBuffer,
                                              SizeOf(uBulkHeader) + iRequestBufferLength, @iTransferred, iTimeout, *sError)
               ; Bei einem Fehler abbrechen.
               If (Not fResult)
                  fStopRequest = #True
                  DebugView::Error("Die Antwort für die Request-Bulk-In-Message konnte nicht vom Gerät empfangen werden.",
                                   iInstanceNumber, miThreadedThreadCounter)
               EndIf

               ; Prüfen ob der richtige MessageHeader empfangen wurde, oder sonstige Daten.
               If fResult
                  If iTransferred
                     ; Die Message-ID nur prüfen wenn überhaupt ein MessageHeader vorhanden sein muss.
                     If fSendRequest
                        fMessageHeaderReceived = Bool(iMessageIdAndTag = PeekI(*RequestBuffer))
                        ; Wenn die IDs identisch sind, kann eine neue Request-BulkIn-Message an das Gerät gesendet werden,
                        ; wenn noch Daten ausständig sind. Ansonst wurden wahrscheinlich einzelne 64 Byte Pakete empfangen.
                        fSendRequest = fMessageHeaderReceived
                     Else
                        ; Der MessageHeader wurde nicht erwartet, die Message-ID braucht somit nicht überprüft werden.
                        fMessageHeaderReceived = #False
                     EndIf

                  Else
                     ; Die Message-IDs können nicht identisch sein, da keine Daten empfangen wurden.
                     fMessageHeaderReceived = #False
                     fSendRequest = #False

                     ; Es sollten alle Daten vorhanden sein, da ein leeres Transfer Message Packet empfangen wurde.
                     fReturn = #True
                     fStopRequest = #True
                  EndIf
               EndIf

               ; Die empfangenen Daten von dem MessageHeader trennen, und zu den schon empfangenen Daten kopieren.
               ; Den Datenspeicher gegebenfalls vergrößern und die benötigen Zähler der Daten berechnen.
               If fResult

                  ; Die Größe der empfangenen Nutzdaten aus dem MessageHeader auslesen.
                  If fMessageHeaderReceived
                     ; Mit Hilfe der uBulkHeader-Struktur auf die empfangenen Daten zugreifen.
                     *uBulkInHeader = *RequestBuffer
                     iTransferredData = *uBulkInHeader\iTransferSize
                     ; Die angegebene Datengröße der Daten die vom Gerät gesendet werden, für die weitere Berechnung sichern.
                     iPendingData = iTransferredData

                     ; Prüfen ob die letzte Transfer Message empfangen wurde. Das Bit 0 'End Of Message' (EOM) ist dann gesetzt.
                     fEndOfMessage = Bool(*uBulkInHeader\iTransferAttributes & #iTransferAttributeEndOfTransfer)

                     ; Prüfen ob weniger Daten empfangen wurden, als im MessageHeader angegeben ist.
                     If (iTransferredData > (iTransferred - SizeOf(uBulkHeader)))
                        ; Die tatsächliche Anzahl der vorhandenen Daten berechnen.
                        iTransferredData = (iTransferred - SizeOf(uBulkHeader))
                        ; Jetzt müssen die noch ausstehenden Daten, ohne das Senden einer neuen Request-Bulk-In-Message
                        ; eingelesen werden.
                        If (eMessageId = #eMessageIdDevDepMsgIn)
                           fSendRequest = #False
                        EndIf
                        
                     ElseIf (iTransferredData = (iTransferred - SizeOf(uBulkHeader)))
                        ; Eine Vendor Specific In Message besitzt kein 'End Of Message' Bit.
                        ; Wenn es nur eine Transfer Message gibt, sollten jetzt alle Daten vorhanden sein.
                        If (eMessageId = #eMessageIdVendorSpecificIn)
                           fEndOfMessage = #True
                        EndIf
                     EndIf

                  Else
                     ; Es ist kein MessageHeader vorhanden, es wurde anscheinend ein einzelnes 64 Byte Bulk-Transfer Packet
                     ; empfangen. Alles was empfangen wurde gehört zu den Daten.
                     iTransferredData = iTransferred
                  EndIf

                  ; Die noch verbleibenden Daten berechnen.
                  iPendingData - iTransferredData

                  ; Prüfen ob schon alle Daten empfangen wurden.
                  If (iPendingData <= 0)
                     ; Es wurden alle Daten empfangen, nur mehr die fehlenden Daten kopieren.
                     iTransferredData - Abs(iPendingData)
                  EndIf

                  ; Die aktuell empfangene Datengröße berechnen für z.B. die Weitergabe zur Statusanzeige.
                  iReceivedData + iTransferredData

                  ; Prüfen ob der Datenspeicher für die zusätzlichen Daten noch ausreicht.
                  If (fMessageHeaderReceived And (iTransferredData > 0))
                     If ((iDataWritePosition + iTransferredData) > iDataSize)
                        ; Speicherplatz für die Daten vergrößern.
                        *DataReAllocate = ReAllocateMemory(*Data\i, iDataSize + #iDataSizeReAllocate, #PB_Memory_NoClear)
                        If (Not *DataReAllocate)
                           ; Keine Daten zusammenkopieren, Empfang der Daten abbrechen.
                           fError = #True
                           fReturn = #False
                           fStopRequest = #True
                           *sError\s = CurrentProcedureNameInstance +
                                       "Der Speicher für die zu empfangenden Daten konnte nicht realloziert werden."
                           DebugView::Error("Der Speicher für die zu empfangenden Daten konnte nicht realloziert werden.",
                                            iInstanceNumber, miThreadedThreadCounter)
                           If *Data\i
                              FreeMemory(*Data\i)
                              *Data\i = #Null
                           EndIf
                        Else
                           iDataSize + #iDataSizeReAllocate
                           *Data\i = *DataReAllocate
                        EndIf
                     EndIf
                  EndIf

                  ; Die neuen Daten jetzt hinten an die vorhandenen Daten kopieren.
                  If ((Not fError) And (iTransferredData > 0))
                     ; Wenn ein MessageHeader vorhanden ist, diesen nicht mitkopieren.
                     If fMessageHeaderReceived
                        iDataOffset = SizeOf(uBulkHeader)
                     Else
                        iDataOffset = 0
                     EndIf
                     If (iTransferredData > 0)
                        CopyMemory(*RequestBuffer + iDataOffset, *Data\i + iDataWritePosition, iTransferredData)
                        ; Die neue Position für das Schreiben in den Datenspeicher berechnen.
                        iDataWritePosition + iTransferredData
                     EndIf
                  EndIf

                  ; Wenn weitere Daten erwartet werden, dass abschließende Short Packet vom Gerät ohne Daten einlesen.
                  If (Not fEndOfMessage)
                     If (Not fSendRequest)
                        ; Das abschließende Short Packet vom Gerät einlesen und verwerfen.
                        If (Not LibUsb::BulkTransfer(*uDeviceHandle, iEndpointBulkIn, *RequestBuffer, iRequestBufferLength,
                                                     #Null, iTimeout, *sError))
                           DebugView::Error("Das abschließende 'Short Packet' konnte nicht empfangen werden.",
                                            iInstanceNumber, miThreadedThreadCounter)
                        EndIf
                     EndIf
                     ; Für die noch ausstehenden Daten eine neue Request-BulkIn-Message senden.
                     fResendRequest = #True
                  Else
                     ; Keine weiteren Daten mehr empfangen.
                     fReturn = #True
                     fStopRequest = #True
                  EndIf

                  ; Prüfen ob alle Daten empfangen wurden.
                  If (fSendRequest And fEndOfMessage)
                     ; Wenn eine neue Request-BulkIn-Message gesendet werden soll, aber schon die letzte Message
                     ; empfangen wurde, sollten alle Daten vorhanden sein.
                     fReturn = #True
                     fStopRequest = #True
                  EndIf

                  ; Nur innerhalb eines Threads auf das Abbrechen prüfen und den Fortschritt senden.
                  If fThread
                     ; Nach dem Empfang der Daten, prüfen ob die Kommunikation abgebrochen werden soll.
                     ; Ansonst ein Ereignis für die Anzahl der empfangenen Daten senden.
                     If LockInstance(iInstanceNumber, *sError)
                        With mInstanceMap()
                           fExit = Bool((\eThreadStatus = #eThreadStatusCancel) Or (\eThreadStatus = #eThreadStatusEnd))
                           iEventListenerWindowNumber = \uThreadData\iEventListenerWindowNumber
                           ; Ereignisdaten für PostEvent.
                           If Not (fStopRequest Or fExit)
                              fEventProgress = Bool((iReceivedData - iEventReceivedData) >= iEventDataSize)
                           Else
                              fEventProgress = #False
                           EndIf
                        EndWith
                     EndIf
                     UnlockInstance(iInstanceNumber)

                     ; Kommunikation beenden wenn notwendig.
                     If (fExit And (Not fStopRequest))
                        fReturn = #False  ; Fehler für den Abbruch zurückgeben.
                        fStopRequest = #True
                        If (Not AbortRequestData(iInstanceNumber, uBulkOutHeader\iTag, *sError.String))
                           ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
                           DebugView::Error("Es ist ein Fehler beim Abbrechen des Empfangens aufgetreten.",
                                            iInstanceNumber, miThreadedThreadCounter)
                        EndIf
                        *sError\s = CurrentProcedureNameInstance + "Der Thread für das Empfangen wurde abgebrochen." 
                        DebugView::Error("Der Thread für das Empfangen wurde abgebrochen.",
                                         iInstanceNumber, miThreadedThreadCounter)
                        
                     ; Ereignis für den Fortschritt der empfangenen Daten auslösen.
                     ElseIf (Not fStopRequest)
                        If (fThread And fEventProgress)
                           iEventReceivedData = iReceivedData
                           PostEvent(#eCustomEventUsbTmcThreadData, iEventListenerWindowNumber,
                                     iInstanceNumber, #eThreadEventRequestProgress, iReceivedData)
                        EndIf
                     EndIf
                  EndIf
                     
               EndIf ; fResult. ; Die empfangenen Daten zusammenkopieren.

               ; Eine neue Request-BulkOut-Message, für die weiteren Daten an das Gerät senden wenn notwendig.
               If fResendRequest
                  fResendRequest = #False
                  fSendRequest = #True
               EndIf

            EndIf ; fResult. ; Die Anwort vom Gerät einlesen.
         Wend  ;  fStopRequest.
      EndIf ; (Not fError). ; Die Transfer Messages senden und empfangen bis alle Daten übertragen wurden.

      ; Den Datenspeicher an die empfangenen Daten anpassen und verkleinern.
      ; Das optionale Terminierungszeichen wird ebenfalls entfernt.
      If fReturn
         *DataReAllocate = ReAllocateMemory(*Data\i, iDataWritePosition - iTermCharLength, #PB_Memory_NoClear)
         If (Not *DataReAllocate)
            fReturn = #False
            *sError\s = CurrentProcedureNameInstance +
                        "Der Speicher mit den empfangenen Daten konnte nicht realloziert werden."
            DebugView::Error("Der Speicher mit den empfangenen Daten konnte nicht realloziert werden.",
                             iInstanceNumber, miThreadedThreadCounter)
            If *Data\i
               FreeMemory(*Data\i)
               *Data\i = #Null
            EndIf
         Else
            *Data\i = *DataReAllocate
            *iTransferred\i = iDataWritePosition - iTermCharLength
         EndIf
      EndIf

      ; Den Request-Speicher freigeben.
      If *RequestBuffer
         FreeMemory(*RequestBuffer)
      EndIf

      ; Gerät wieder freigeben.
      If fDeviceLocked
         UnlockDevice(iInstanceNumber)
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure
   
   
   Procedure.i WaitData(iInstanceNumber.i, *fMessageAvailable.Integer, *iStatusByte.Integer, *sError.String)
      ; Prüft ob eine Antwort bzw. Daten im Bulk-In Endpoint vom Gerät bereitstehen.
      ; Wenn ein '*SRE 16' SCPI-Kommando gesendet und das Flag fWaitDataServiceRequest auf True gesetzt wurde,
      ; wird auf einen Servie Request Interrupt Transer gewartet.
      ; Ist das Flag fWaitDataStatusByte auf True gesetzt, wird dafür das Status Byte per Control- oder Interrupt Transfer
      ; eingelesen. Wenn das MAV-Bit gesetzt ist, stehen Daten für das Abholen per RequestData() bereit.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- fMessageAvailable: True wenn Daten bereitstehen, ansonst False.
      ; <- iStatusByte:       Das empfangene Status Byte.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die Daten gesendet wurden, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, sError.String
      Protected iTag.i, f488InterfaceIs4882.i, f488DeviceIsServiceRequestCapable.i
      Protected fWaitDataServiceRequest.i, fWaitDataStatusByte.i
      Protected uReadStatusByteResponse.uReadStatusByteResponse
      Protected uInterruptTransferResponse.uInterruptTransferResponse
              
      fResult = LockInstance(iInstanceNumber, *sError)
      If fResult
         With mInstanceMap()
            iTag = IncrementControlTransferIdentifier(iInstanceNumber, \iControlTransferIdentifier)
            \iControlTransferIdentifier = iTag
            f488InterfaceIs4882 = \uCapabilities\f488InterfaceIs4882
            f488DeviceIsServiceRequestCapable = \uCapabilities\f488DeviceIsServiceRequestCapable
            fWaitDataServiceRequest = \uThreadData\fWaitDataServiceRequest
            fWaitDataStatusByte = \uThreadData\fWaitDataStatusByte
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)
      
      ; Prüfen ob beide fWaitData-Flags gesetzt sind oder keines.
      fResult = Bool(Not (fWaitDataServiceRequest And fWaitDataStatusByte))
      If (Not fResult)
         *sError\s = "Beide Flags fWaitDataServiceRequest und fWaitDataStatusByte sind auf True gesetzt."
         DebugView::Error("Beide Flags fWaitDataServiceRequest und fWaitDataStatusByte sind auf True gesetzt.",
                          iInstanceNumber, miThreadedThreadCounter)
         ; Abrechen.
         fWaitDataServiceRequest = #False
         fWaitDataStatusByte = #False
      EndIf
      
      ; Einen Service Request Interrupt Transfer abwarten wenn erforderlich.
      If fWaitDataServiceRequest
         If fWaitDataServiceRequest
            ; Prüfen ob das Gerät einen Service Request unterstützt.
            fResult = f488DeviceIsServiceRequestCapable
            If (Not fResult)
               DebugView::Info("Das Gerät unterstützt keinen Interrupt Transfer.",
                                iInstanceNumber, miThreadedThreadCounter)
               ; Erfolg zurückgeben.
               fReturn = #True
            EndIf
      
            ; Das Status Byte über den Interrupt-In Endpoint einlesen.
            If fResult
               fResult = InterruptTransfer(iInstanceNumber, @uInterruptTransferResponse, *sError)
               If (Not fResult)
                  ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
                  DebugView::Error("Das Status Byte konnte nicht vom Interrupt-In Endpoint eingelesen werden (Service Request).",
                                   iInstanceNumber, miThreadedThreadCounter)
               Else
                  ; Prüfen ob eine Anwort von dem Service Request empfangen wurde.
                  ;  Das erste Byte mit dem Wert 0x81 wird in Tabelle 6 der USB488-Dokumentation vorgegeben [USB488: 3.4.1].
                  If Not (uInterruptTransferResponse\iNotify1 = $81)
                     *sError\s = "Das empfangene iNotify1 Byte ist kein Service Request (0x81). " +
                                 "uInterruptTransferResponse\iNotify1 = " + uInterruptTransferResponse\iNotify1
                     DebugView::Error("Das empfangene iNotify1 Byte ist kein Service Request. " + 
                                      "uInterruptTransferResponse\iNotify1 = " + uInterruptTransferResponse\iNotify1,
                                      iInstanceNumber, miThreadedThreadCounter)
                  Else
                     ; Erfolg zurückgeben.
                     fReturn = #True
                     ; Das empfangene Status Byte ausgeben.
                     DebugView::Info("Das empfangene Status Byte von InterruptTransfer() (Service Request). " + 
                                     "uInterruptTransferResponse\iNotify2 = %" +
                                     Bin(uInterruptTransferResponse\iNotify2, #PB_Byte),
                                     iInstanceNumber, miThreadedThreadCounter)
                     ; Nur als Hinweis das RQS-Bit für einen angeforderten Request Service vom Gerät ausgeben.
                     If Not (uInterruptTransferResponse\iNotify2 & #eUsbTmcStatusByteRQS)
                        DebugView::Info("Das Request Service Bit ist nicht im Status Byte gesetzt (Service Request).",
                                        iInstanceNumber, miThreadedThreadCounter)
                     EndIf
                     ; Das MAV-Bit und das Status Byte zurückgeben.
                     *fMessageAvailable\i = Bool(uInterruptTransferResponse\iNotify2 & #eUsbTmcStatusByteMAV)
                     *iStatusByte\i = uInterruptTransferResponse\iNotify2
                  EndIf
               EndIf
            EndIf
         EndIf
      EndIf
            
            
      ; Das Status Byte vom Gerät anforden wenn erforderlich.
      If fWaitDataStatusByte
         fResult = ReadStatusByte(iInstanceNumber, iTag, @uReadStatusByteResponse, *sError)
         If (Not fResult)
            ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
            DebugView::Error("Das Status Byte konnte nicht eingelesen werden.", iInstanceNumber, miThreadedThreadCounter)
         EndIf
         
         ; Prüfen ob das empfangene Tag mit dem gesendeten Tag identisch ist.
         If fResult
            fResult = Bool(iTag = uReadStatusByteResponse\iTag)
            If (Not fResult)
               *sError\s = "Das gesendete Tag stimmt nicht mit dem empfangenen Tag überein. iTag out = " + iTag +
                           " - iTag in = " + uReadStatusByteResponse\iTag
               DebugView::Info("Das gesendete Tag stimmt nicht mit dem empfangenen Tag überein. iTag out = " + iTag +
                               " - iTag in = " + uReadStatusByteResponse\iTag, iInstanceNumber, miThreadedThreadCounter)
            EndIf
         EndIf
         
         ; Den USBTMC-Status prüfen.
         If fResult
            fResult = Bool(uReadStatusByteResponse\eUsbTmcStatus = #eUsbTmcStatusSuccess)
            If (Not fResult)
               If (uReadStatusByteResponse\eUsbTmcStatus = #eUsbTmcStatusInterruptInBusy)
                  *sError\s = CurrentProcedureNameInstance + "Der Interrupt-In Endpoint ist belegt. " +
                              "uReadStatusByteResponse\eUsbTmcStatus = #eUsbTmcStatusInterruptInBusy"
                  DebugView::Error("Der Interrupt-In Endpoint ist belegt. uReadStatusByteResponse\eUsbTmcStatus = " +
                                   "#eUsbTmcStatusInterruptInBusy", iInstanceNumber, miThreadedThreadCounter)
                  ; Den Interrupt Transfer ausführen um den Interrupt-In Endpoint wieder freizugeben.
                  If (Not InterruptTransfer(iInstanceNumber, @uInterruptTransferResponse, sError))
                     DebugView::Error("Der Interrupt Transfer konnte nicht ausgeführt werden (#eUsbTmcStatusInterruptInBusy).",
                                      iInstanceNumber, miThreadedThreadCounter)
                  Else
                     DebugView::Info("Der Interrupt Transfer konnte erfolgreich ausgeführt werden (#eUsbTmcStatusInterruptInBusy).",
                                     iInstanceNumber, miThreadedThreadCounter)
                     ; Abbrechen und die Fehlermeldung mit Interrupt-In Busy zurückgeben.
                  EndIf
               Else
                  ; Ein unbekannter Fehler ist aufgetreten.
                  ; Siehe dazu USBTMC Specification Rev1.0 - PDF Page 25, Table 16 -- USBTMC_status values.
                  *sError\s = CurrentProcedureNameInstance + "Es ist ein unbekannter Fehler aufgetreten. " +
                              "uReadStatusByteResponse\eUsbTmcStatus = " + uReadStatusByteResponse\eUsbTmcStatus
                  DebugView::Error("Es ist ein unbekannter Fehler aufgetreten. uReadStatusByteResponse\eUsbTmcStatus = " +
                                   uReadStatusByteResponse\eUsbTmcStatus, iInstanceNumber, miThreadedThreadCounter)
               EndIf
            EndIf
         EndIf
         
         ; Prüfen ob das Gerät eine Antwort per Interrupt Transfer unterstützt.
         ; Die Schnittstelle muss einen Interrupt-In Endpoint besitzen und das Nachrichtenaustauschprotokoll (MEP) des
         ; Standards IEEE 488.2 unterstützen [USB488: Tabelle 8 und IEEE 488.2: 6].
         If fResult
            fResult = Bool(f488InterfaceIs4882 And (uReadStatusByteResponse\iStatusByte = #Null))
            If (Not fResult)
               DebugView::Info("Das Gerät ünterstützt keinen Interrupt Transfer (Status Byte).",
                               iInstanceNumber, miThreadedThreadCounter)
               ; Erfolg zurückgeben.
               fReturn = #True
               ; Das empfangene Statusbyte ausgeben.
               DebugView::Info("Das empfangene Status Byte von ReadStatusByte(). uReadStatusByteResponse\iStatusByte = %" +
                               Bin(uReadStatusByteResponse\iStatusByte, #PB_Byte), iInstanceNumber, miThreadedThreadCounter)
               ; Das MAV-Bit und das Status Byte zurückgeben.
               *fMessageAvailable\i = Bool(uReadStatusByteResponse\iStatusByte & #eUsbTmcStatusByteMAV)
               *iStatusByte\i = uReadStatusByteResponse\iStatusByte
            EndIf
         EndIf
         
         ; Das Status Byte über den Interrupt-In Endpoint einlesen.
         If fResult
            fResult = InterruptTransfer(iInstanceNumber, @uInterruptTransferResponse, *sError)
            If (Not fResult)
               ; *sError - Fehlerbeschreibung von Prozedur übernehmen.
               DebugView::Error("Das Status Byte konnte nicht vom Interrupt-In Endpoint eingelesen werden (Status Byte).",
                                iInstanceNumber, miThreadedThreadCounter)
            Else
               ; Das empfangene Tag überprüfen.
               If Not ((iTag | $80) = uInterruptTransferResponse\iNotify1)
                  *sError\s = "Das gesendete Tag stimmt nicht mit dem empfangenen Tag überein. iTag|$80 = " + Str(iTag | $80) +
                              " - iNotify1 = " + uInterruptTransferResponse\iNotify1
                  DebugView::Error("Das gesendete Tag stimmt nicht mit dem empfangenen Tag überein. iTag|$80 = " +
                                   Str(iTag | $80) + " - iNotify1 = " + uInterruptTransferResponse\iNotify1,
                                   iInstanceNumber, miThreadedThreadCounter)
               Else
                  ; Erfolg zurückgeben.
                  fReturn = #True
                  ; Das empfangene Status Byte ausgeben.
                  DebugView::Info("Das empfangene Status Byte von InterruptTransfer()  (Status Byte). " +
                                  "uInterruptTransferResponse\iNotify2 = %" +
                                  Bin(uInterruptTransferResponse\iNotify2, #PB_Byte),
                                  iInstanceNumber, miThreadedThreadCounter)
                  ; Das MAV-Bit und das Status Byte zurückgeben.
                  *fMessageAvailable\i = Bool(uInterruptTransferResponse\iNotify2 & #eUsbTmcStatusByteMAV)
                  *iStatusByte\i = uInterruptTransferResponse\iNotify2
               EndIf
            EndIf
         EndIf
      EndIf
      
      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure
   
   ;-
   ;- -- Thread-Prozeduren.

   Procedure.i GetThreadData(iInstanceNumber.i, *uData.uThreadData, *sError.String)
      ; Gibt die Struktur mit den Daten zurück, die mit dem Thread ermittelt wurden.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *uData:            Zeiger auf die Struktur in der die Daten kopiert werden.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True bei Erfolg, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         fReturn = #True
         With mInstanceMap()
            CopyStructure(mInstanceMap()\uThreadData, *uData, uThreadData)
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SetThreadStatus(iInstanceNumber.i, eThreadStatus.i, *sError.String)
      ; Setzt den Status für den Thread.
      ; Wird für das Abbrechen einer laufenden Kommunikation, oder komplette Beenden des Threads verwendet.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> eThreadStatus:     Ein Wert aus der Enumeration eThreadStatus.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True bei Erfolg, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, iThreadNumber.i, iMutex.i
      #iWaitThreadTimeout = 1000   ; Millisekunden.
      
      If (eThreadStatus <= #eThreadStatusMinimum) Or (eThreadStatus >= #eThreadStatusMaximum)
         *sError\s = CurrentProcedureNameInstance + "Ungültiger Wert für den Threadstatus. eThreadStatus = " + eThreadStatus
         DebugView::Error("Ungültiger Wert für den Threadstatus. eThreadStatus = " + eThreadStatus,
                          iInstanceNumber, miThreadedThreadCounter)

      Else
         fResult = LockInstance(iInstanceNumber, *sError)
         If fResult
            With mInstanceMap()
               \eThreadStatus = eThreadStatus
               iThreadNumber = \iThreadNumber
               iMutex = \iMutex
            EndWith
         EndIf
         UnlockInstance(iInstanceNumber)

         ; Den Thread beenden wenn erforderlich.
         If fResult
            fReturn = #True
            If (eThreadStatus = #eThreadStatusEnd)
               If (iThreadNumber And IsThread(iThreadNumber))
                  ResumeThread(iThreadNumber)
                  If (Not WaitThread(iThreadNumber, #iWaitThreadTimeout))
                     KillThread(iThreadNumber)
                     DebugView::Error("Der Thread wurde gekillt. iThreadNumber = " + iThreadNumber,
                                      iInstanceNumber, miThreadedThreadCounter)
                     ; Mutex löschen. Wird wieder automatisch erstellt in LockDevice().
                     ; Wenn der Mutex vom Thread gesperrt war, kann er nicht mehr freigegeben werden.
                     If iMutex
                        FreeMutex(iMutex)
                        If LockInstance(iInstanceNumber, *sError)
                           With mInstanceMap()
                              \iMutex = #Null
                              \iThreadNumber = #Null
                              \eThreadStatus = #eThreadStatusNone
                           EndWith
                        EndIf
                        UnlockInstance(iInstanceNumber)
                     EndIf
                  EndIf
               EndIf
            EndIf
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i GetRequestProgressDataLength(iInstanceNumber.i, *iRequestProgressDataLength.Integer, *sError.String)
      ; Gibt die Länge der jeweils neu zu empfangenden Daten zurück, bei der ein Ereignis für den Fortschritt des
      ; Empfangs der Daten gesendet wird.
      ; -> iInstanceNumber:               Die Nummer der Instanz.
      ; <- *iRequestProgressDataLength:   Die Länge der jeweils neu zu empfangenden Daten in Bytes.
      ; <- sError:                        Fehlerbeschreibung.
      ; Rückgabe:                         True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         If *iRequestProgressDataLength
            *iRequestProgressDataLength\i = mInstanceMap()\iRequestProgressDataLength
            fReturn = #True
         EndIf
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SetRequestProgressDataLength(iInstanceNumber.i, iRequestProgressDataLength.i, *sError.String)
      ; Setzt die Länge der jeweils neu zu empfangenden Daten, bei der ein Ereignis für den Fortschritt des
      ; Empfangs der Daten gesendet werden soll.
      ; -> iInstanceNumber:            Die Nummer der Instanz.
      ; -> iRequestProgressDataLength: Die Länge der jeweils neu zu empfangenden Daten in Bytes.
      ; <- sError:                     Fehlerbeschreibung.
      ; Rückgabe:                      True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         mInstanceMap()\iRequestProgressDataLength = iRequestProgressDataLength
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i GetRequestBufferLength(iInstanceNumber.i, *iRequestBufferLength.Integer, *sError.String)
      ; Gibt die Länge des Empfangsbuffers für die jeweils neu zu empfangenden Daten zurück.
      ; -> iInstanceNumber:         Die Nummer der Instanz.
      ; <- *iRequestBufferLength:   Die Länge des Empfangbuffers in Bytes.
      ; <- sError:                  Fehlerbeschreibung.
      ; Rückgabe:                   True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         If *iRequestBufferLength
            *iRequestBufferLength\i = mInstanceMap()\iRequestBufferLength
            fReturn = #True
         EndIf
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i SetRequestBufferLength(iInstanceNumber.i, iRequestBufferLength.i, *sError.String)
      ; Setzt die Größe des Empfangsbuffers in Bytes. Wenn Geräte Daten in einem Transfer senden und der Empfangsbuffer
      ; größer als die Daten ist, werden keine Ereignisse für den Fortschritt gesendet. Ist der Buffer kleiner als die Daten,
      ; sendet das Gerät in mehreren Transfers und Ereignisse für den Fortschritt können gesendet werden.
      ; Die Ereignisse werden nur bei Verwendung eines Threads gesendet.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> iRequestBufferLength: Die Länge des Empfangsbuffers in Bytes.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True bei Erfolg oder False bei einem Fehler.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i

      If LockInstance(iInstanceNumber, *sError)
         mInstanceMap()\iRequestBufferLength = iRequestBufferLength
         fReturn = #True
      EndIf
      UnlockInstance(iInstanceNumber)

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure ThreadData(iInstanceNumber.i)
      ; Pozedur für die Kommunikation mit dem Gerät in einem Thread.
      ; Es werden die Daten vorbereitet, gesendet und empfangen. Die Ereignisse werden per PostEvent() weitergegeben.
      ; Der Thread pausiert am Ende der Prozedur automatisch wenn eine Threadnummer vorhanden ist,
      ; ansonst wird der Thread beendet.
      ; Für das Beenden des pausierenden Threads, muss dem Feld eThreadStatus der Instanzstruktur,
      ; der Wert #eThreadEnd zugewiesen und der Thread per ResumeThread() wieder gestartet werden.
      ; Der Thread wird beim Beenden der Instanz automatisch beendet, siehe DeleteInstance().
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      Protected fResult.i, fExit.i, sError.String
      Protected eThreadStatus.i, fMessageAvailable.i, iStatusByte.i
      Protected iThreadNumber.i, iEventListenerWindowNumber.i
      Protected uData.uThreadData
      
      If (Not miMutex)
         DebugView::Error("Es ist kein Mutex vorhanden.", iInstanceNumber)
      Else
         LockMutex(miMutex)
         miThreadCounter + 1
         miThreadCounterMax + 1
         miThreadedThreadCounter = miThreadCounterMax ; Für die thread-sichere Debugausgabe.
         UnlockMutex(miMutex)
      EndIf
      DebugView::ThreadIn(iInstanceNumber, miThreadedThreadCounter)
      
      Repeat
         fExit = #False
         ; Prüfen ob die Instanz vorhanden ist, und die Daten kopieren.
         If (Not LockInstance(iInstanceNumber, sError))
            fExit = #True
         Else
            With mInstanceMap()
               fExit = Bool(\eThreadStatus = #eThreadStatusEnd)
               If (Not fExit)
                  \fThreadRunning = #True
                  iEventListenerWindowNumber = \uThreadData\iEventListenerWindowNumber
                  CopyStructure(\uThreadData, uData, uThreadData)
               EndIf
            EndWith
         EndIf
         UnlockInstance(iInstanceNumber)

         If (Not fExit)
            With uData
              ; Die Daten vorbereiten und senden.
               If \fSendData
                  ; Die Daten vorbereiten, Fehler ignorieren, vielleicht sind es leere Daten, und so gewollt.
                  fResult = PrepareData(iInstanceNumber, \sScpi, \SendData, \iSendDataLength, sError)
                  
                  ; Die Daten senden.
                  fResult = SendData(iInstanceNumber, \eSendMessageId, @\iSendTransferred, sError)
                  fExit = Bool(Not fResult)
                  \fSendOk = fResult
                  \fSendFinished = #True
                  uData\sError = sError\s

                  ; Die Daten in der Instanz sichern.
                  If LockInstance(iInstanceNumber, sError)
                     CopyStructure(uData, mInstanceMap()\uThreadData, uThreadData)
                     If (Not fExit)
                        fExit = Bool(mInstanceMap()\eThreadStatus = #eThreadStatusEnd)
                     EndIf
                  EndIf
                  UnlockInstance(iInstanceNumber)
                  
                  ; Das Ereignis an das Fenster senden, damit die Daten verarbeitet werden können.
                  PostEvent(#eCustomEventUsbTmcThreadData, iEventListenerWindowNumber,
                            iInstanceNumber, #eThreadEventSendFinished, #Null)
               EndIf
               
               ; TODO - Funktioniert nicht mit den Rigol Geräten. Status Byte ist im Interrupt Transfer immer Null.
               ; Warten bis die Daten bereitstehen.
               If (\fWaitDataServiceRequest Or \fWaitDataStatusByte)
                  If (Not fExit)
                     If \fRequestData
                        If (\fSendData And \fSendOk)
                           ; Auf die Daten warten.
                           ; TODO - Wartezeit für PIC18F Firmware. Firmware kontrollieren!
                           Delay(10)
                            
                           If (Not WaitData(iInstanceNumber, @fMessageAvailable, @iStatusByte, sError))
                              DebugView::Error("Fehler beim Warten bis die Daten bereitstehen.",
                                               iInstanceNumber, miThreadedThreadCounter)
                              ; Keine weiteren Daten empfangen mit RequestData().
                              ; Fehler ignorieren, möglicherweise hält sich die Firmware des Geräts nicht an den 488-Standard.
                              ;fExit = #True
                              
                           Else
                              If (Not fMessageAvailable)
                                 DebugView::Error("Es stehen keine Daten bereit. Status Byte = %" + Bin(iStatusByte, #PB_Byte),
                                                  iInstanceNumber, miThreadedThreadCounter)
                                 ; Keine weiteren Daten empfangen mit RequestData().
                                 ; Fehler ignorieren, möglicherweise hält sich die Firmware des Geräts nicht an den 488-Standard.
                                 ;
                                 ; INFO - Die Geräte von Rigol wie DSA815, MSO2000, DS1000Z, DG4000 und DP800 antworten sofort
                                 ; INFO - ohne gesetztes MAV-Bit. Auch mehrmaliges Abfragen hilft nicht, das MAV-Bit bleibt Null.
                                 ;fExit = #True
                              Else
                                 ; Die Daten stehen bereit.
                                 DebugView::Info("Die Daten stehen bereit. Status Byte = %" + Bin(iStatusByte, #PB_Byte),
                                                 iInstanceNumber, miThreadedThreadCounter)
                              EndIf
                           EndIf
                           
                        EndIf
                     EndIf
                  EndIf
               EndIf
            
               ; Die Daten empfangen.
               If (Not fExit)
                  If \fRequestData
                     ; Nur Empfangen wenn erfolgreich gesendet wurde, oder überhaupt nicht gesendet wurde.
                     If (\fSendData And \fSendOk) Or (Not \fSendData)
                        fResult = RequestData(iInstanceNumber, @\RequestData, @\iRequestTransferred, \eRequestMessageId, sError)
                        fExit = Bool(Not fResult)
                        \fRequestOk = fResult
                     EndIf
                     \fRequestFinished = #True
                     uData\sError = sError\s

                     ; Die Daten in der Instanz sichern.
                     If LockInstance(iInstanceNumber, sError)
                        CopyStructure(uData, mInstanceMap()\uThreadData, uThreadData)
                     EndIf
                     UnlockInstance(iInstanceNumber)
                     
                     ; Das Ereignis an das Fenster senden, damit die Daten verarbeitet werden können.
                     PostEvent(#eCustomEventUsbTmcThreadData, iEventListenerWindowNumber,
                               iInstanceNumber, #eThreadEventRequestFinished, #Null)
                  EndIf
               EndIf

               ; Der Thread ist fertig.
               If LockInstance(iInstanceNumber, sError)
                  mInstanceMap()\fThreadRunning = #False
                  If (Not fExit)
                     fExit = Bool(mInstanceMap()\eThreadStatus = #eThreadStatusEnd)
                  EndIf
                  If fExit
                     mInstanceMap()\iThreadNumber = #Null
                     mInstanceMap()\eThreadStatus = #eThreadStatusNone
                  EndIf
                  iThreadNumber = mInstanceMap()\iThreadNumber
               EndIf
               UnlockInstance(iInstanceNumber)
               ; Das Ereignis an das Fenster senden, damit die Daten verarbeitet werden können.
               PostEvent(#eCustomEventUsbTmcThreadData, iEventListenerWindowNumber,
                         iInstanceNumber, #eThreadEventThreadFinished, #Null)
            EndWith

            ; Den Thread pausieren, wenn eine Threadnummer vorhanden ist.
            If (Not fExit)
               If iThreadNumber
                  PauseThread(iThreadNumber)
               EndIf
            EndIf
         EndIf
      Until fExit
      
      If (Not miMutex)
         DebugView::Error("Es ist kein Mutex vorhanden.", iInstanceNumber, miThreadedThreadCounter)
      Else
         LockMutex(miMutex)
         miThreadCounter - 1
         UnlockMutex(miMutex)
      EndIf
 
      DebugView::Info("miThreadCounter = " + miThreadCounter, iInstanceNumber, miThreadedThreadCounter)
      DebugView::ThreadOut(iInstanceNumber, miThreadedThreadCounter)
      miThreadedThreadCounter = 0
   EndProcedure


   Procedure.i SendRequestThreadData(iInstanceNumber.i, *uData.uThreadData, *sError.String)
      ; Das Aufbereiten, Senden und Empfangen der Daten wird in einem Thread ausgeführt.
      ; Der Fortschritt der Kommunikation wird per PostEvent() gesendet. Das betreffende Fenster muss dafür das
      ; Event #eCustomEventUsbTmcThreadData abfragen. Der aktuelle Fortschritt wird als Objektnummer übergeben, der ein Wert
      ; aus der Enumeration eThreadStatus ist.
      ; Für das Steuern des Threads gibt es die Prozedur SetThreadStatus(), um z.B. die Kommuikation abzubrechen.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> *uData:            Die Daten für die Kommunikation mit dem Gerät.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn der Thread erstellt wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fError.i, iThreadNumber.i, iEventCounter.i

      ; Die benötigten Daten zum geöffneten Gerät einlesen.
      If (Not LockInstance(iInstanceNumber, *sError))
         fError = #True
         DebugView::Error("Diese Instanz existiert nicht.", iInstanceNumber, miThreadedThreadCounter)

      Else
         With mInstanceMap()
            ; Prüfen ob ein Thread vorhanden und verfügbar ist.
            If (\iThreadNumber And \fThreadRunning)
               fError = #True
               *sError\s = CurrentProcedureNameInstance + "Der Thread ist in Verwendung. iThreadNumber = " + \iThreadNumber
               DebugView::Error("Der Thread ist in Verwendung. iThreadNumber = " + \iThreadNumber,
                                iInstanceNumber, miThreadedThreadCounter)

            Else
               iThreadNumber = \iThreadNumber
               ; Die übergebenen Daten in die Instanz kopieren.
               CopyStructure(*uData, \uThreadData, uThreadData)
               \eThreadStatus = #eThreadStatusRun
            EndIf
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)

      ; Den Thread erneut starten oder einen neuen Thread erstellen und die Threadnummmer sichern.
      If (Not fError)
         If (Not iThreadNumber)
            DebugView::Info("Ein neuer Thread wird erstellt.", iInstanceNumber)
         ElseIf (Not IsThread(iThreadNumber))
            DebugView::Info("Die vorhandene Threadnummer ist ungültig. iThreadNumber = " + iThreadNumber,
                            iInstanceNumber, miThreadedThreadCounter)
            DebugView::Info("Ein neuer Thread wird erstellt.", iInstanceNumber, miThreadedThreadCounter)
            iThreadNumber = #Null
         Else
            DebugView::Info("Der vorhandene Thread wird wieder ausgeführt. iThreadNumber = " + iThreadNumber,
                            iInstanceNumber, miThreadedThreadCounter)
            ResumeThread(iThreadNumber)

            fReturn = #True
         EndIf

         ; Einen neuen Thread erstellen.
         If (Not iThreadNumber)
            iThreadNumber = CreateThread(@ThreadData(), iInstanceNumber)
            If (Not iThreadNumber)
               fError = #True
               *sError\s = CurrentProcedureNameInstance + "Der Thread konnte nicht erstellt werden."
               DebugView::Error("Der Thread konnte nicht erstellt werden.", iInstanceNumber, miThreadedThreadCounter)
            Else
               DebugView::Info("Ein neuer Thread wurde erstellt. iThreadNumber = " + iThreadNumber,
                               iInstanceNumber, miThreadedThreadCounter)
               If LockInstance(iInstanceNumber, *sError)
                  With mInstanceMap()
                     \iThreadNumber = iThreadNumber
                  EndWith
               EndIf
               UnlockInstance(iInstanceNumber)
               
               fReturn = #True
            EndIf
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure

   ;- --

   Procedure.i ControlTransfer(iInstanceNumber.i, eRequestType.i, eDeviceRequest.i, iValue.i, iIndex.i,
                               *Data.Integer, *iTransferred.Integer, *sError.String)
      ; Sendet einen Control Transfer mit dem übergebenen Daten an das Gerät und gibt die empfangenen Daten zurück.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> eRequestType:      Ein Wert oder Kombination von der Enumeration eRequestType.
      ; -> eDeviceRequest:    Ein Wert von der Enumeration eDeviceRequest.
      ; -> iValue:            Die Bedeutung ist abhängig vom Typ des Requests.
      ; -> iIndex:            Wird in vielen Requests verwendet, um einen Endpoint oder ein Interface zu spezifizieren.
      ; <- *Data:             Ein Zeiger auf die empfangenen Daten. Die Daten müssen wieder freigegeben werden.
      ; <> *iTransferred:     Die Anzahl der übertragenen Bytes.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn der Control Transfer durchgeführt wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, fDeviceLocked.i, iTransferred.i
      Protected *uDeviceHandle, iTimeout

      ; Versuchen das Gerät zu sperren.
      fDeviceLocked = LockDevice(iInstanceNumber, *sError)

      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            *uDeviceHandle = \uDevice\uDeviceHandle
            iTimeout = \iTimeoutControlTransferMilliseconds
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)

      If (Not *uDeviceHandle)
         *sError\s = CurrentProcedureNameInstance +
                     "Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0"
         DebugView::Error("Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0",
                          iInstanceNumber, miThreadedThreadCounter)

      Else
         iTransferred = *iTransferred\i
         fResult = LibUsb::ControlTransfer(*uDeviceHandle, eRequestType, eDeviceRequest, iValue, iIndex,
                                           *Data, *iTransferred, iTimeout, *sError)
         If (Not fResult)
            *sError\s = CurrentProcedureNameInstance + *sError\s
            DebugView::Error(*sError\s, iInstanceNumber)

         Else
            If (*iTransferred\i > iTransferred)
               *sError\s = CurrentProcedureNameInstance +
                           "Es wurde eine falsche Anzahl an Bytes empfangen. " +
                           "Erwartet wurden maximal " + iTransferred +
                           " Bytes, empfangen wurden " + *iTransferred\i + " Bytes."
               DebugView::Error("LibUsb::ControlTransfer: Erwartet wurden maximal " + iTransferred + " Bytes, " +
                                "empfangen wurden " + *iTransferred\i + " Bytes.",
                                iInstanceNumber, miThreadedThreadCounter)

            Else
               fReturn = #True
            EndIf
         EndIf
      EndIf

      ; Gerät wieder freigeben.
      If fDeviceLocked
         UnlockDevice(iInstanceNumber)
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i InterruptTransfer(iInstanceNumber.i, *uData.uInterruptTransferResponse, *sError.String)
      ; Sendet einen Interrupt Transfer an das Gerät und gibt die empfangenen Daten zurück.
      ; Der Interrupt-In Endpoint kann gemäß USBTMC-Spezifikation dazu verwendet werden, Mitteilungen an den Host zu senden.
      ; Das Format ist auf 2 Bytes beschränkt, die die Datenfeldnamen bNotify1 und bNotify2 besitzen. USBTMC definiert selbst
      ; keine Interrupt Transfers, sondern verweist auf Unterklassen-Spezifikationen. Die anzuwendende Unterklassen-Spezifikation
      ; ist bisher ausschließlich USB488. Hier wird ausgeführt, dass ein Gerät mit einer 488.2 USB488-Schnittstelle einen
      ; Interrupt-In Endpoint haben muss. Es gibt nur zwei Bedingungen für ein Gerät, Daten über diesen Endpoint zu versenden.
      ; Eine Bedienungsanforderung (Service Request 'SRQ') und bei einem READ_STATUS_BYTE Request.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *uData:            Ein Zeiger auf die empfangenen Daten.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn der Interrupt Transfer durchgeführt wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, fDeviceLocked.i, iEndpoint.i, iLength.i, iTransferred.i
      Protected *uDeviceHandle, iInterface.i, iTimeout
      
      ; Versuchen das Gerät zu sperren.
      fDeviceLocked = LockDevice(iInstanceNumber, *sError)
      
      If LockInstance(iInstanceNumber, *sError)
         With mInstanceMap()
            *uDeviceHandle = \uDevice\uDeviceHandle
            iInterface = \uDevice\iInterfaceNumber
            iEndpoint = \uDevice\iEndpointInterruptIn
            iTimeout = \iTimeoutInterruptTransferMilliseconds
         EndWith
      EndIf
      UnlockInstance(iInstanceNumber)
      
      If (Not *uDeviceHandle)
         *sError\s = CurrentProcedureNameInstance + "Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0"
         DebugView::Error("Das USBTMC-Gerät ist nicht geöffnet. *uInstance\uDevice\uDeviceHandle = 0",
                          iInstanceNumber, miThreadedThreadCounter)
         
      ElseIf (Not iEndpoint)
         *sError\s = CurrentProcedureNameInstance +
                     "Das USBTMC-Gerät besitzt keinen Interrupt-In Endpoint. *uInstance\uDevice\iEndpointInterruptIn = 0"
         DebugView::Error("Das USBTMC-Gerät besitzt keinen Interrupt-In Endpoint. *uInstance\uDevice\iEndpointInterruptIn = 0",
                          iInstanceNumber, miThreadedThreadCounter)
      ElseIf (Not *uData)
         *sError\s = CurrentProcedureNameInstance + "Der Zeiger für die Daten ist Null. *uData = 0"
         DebugView::Error("Der Zeiger für die Daten ist Null. *uData = 0", iInstanceNumber, miThreadedThreadCounter)
         
      Else
         iLength = SizeOf(uInterruptTransferResponse)
         fResult = LibUsb::InterruptTransfer(*uDeviceHandle, iEndpoint, *uData, iLength, @iTransferred, iTimeout, *sError)
         If (Not fResult)
            *sError\s = CurrentProcedureNameInstance + *sError\s
            DebugView::Error(*sError\s, iInstanceNumber)
            
         Else
            If (iTransferred > iLength)
               *sError\s = CurrentProcedureNameInstance +
                           "Es wurde eine falsche Anzahl an Bytes empfangen. " +
                           "Erwartet wurden maximal " + iLength +
                           " Bytes, empfangen wurden " + iTransferred + " Bytes."
               DebugView::Error("LibUsb::ControlTransfer: Erwartet wurden maximal " + iLength + " Bytes, " +
                                "empfangen wurden " + iTransferred + " Bytes.", iInstanceNumber, miThreadedThreadCounter)
               
            Else
               fReturn = #True
            EndIf
         EndIf
      EndIf
      
      ; Gerät wieder freigeben.
      If fDeviceLocked
         UnlockDevice(iInstanceNumber)
      EndIf
      
      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure

   ;-
   ;- -- USBTMC Device Requests.

   Procedure.i GetCapabilities(iInstanceNumber.i, *uCapabilities.uCapabilities, *sError.String)
      ; Mit diesem Request werden die zusätzlichen Eigenschaften und Fähigkeiten des USBTMC Interface ermittelt.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *uCapabilities:    Zeiger auf die reservierte Struktur für die Eigenschaften.
      ;                       Wenn #Null übergeben wird, werden die Eigenschaften direkt in der Instanz gespeichert.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn die Eigenschaften ermittelt wurden, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fInstance.i
      Protected uGetCapabilities.uGetCapabilities, uCapabilities.uCapabilities
      Protected iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                     LibUsb::#eLIBUSB_RECIPIENT_INTERFACE |
                     LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
      iTransferred = SizeOf(uGetCapabilities)
      iValue = 0  ; Reserviert.

      ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
      If LockInstance(iInstanceNumber, *sError)
         iIndex = mInstanceMap()\uDevice\iInterfaceNumber
      EndIf
      UnlockInstance(iInstanceNumber)

      If (Not ControlTransfer(iInstanceNumber, eRequestType, #eGetCapabilities, iValue, iIndex,
                              @uGetCapabilities, @iTransferred, *sError))
         DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

      Else
         ; Die Eigenschaften zurückgeben, oder in der Instanz speichern.
         If (Not *uCapabilities)
            fInstance = #True
            *uCapabilities = @uCapabilities
         EndIf
         With uGetCapabilities
            ; USBTMC-Gerät.
            *uCapabilities\iTmcBcdVersionNumber = \iBcdUsbTmc
            *uCapabilities\fTmcInterfaceAcceptsIndicatorPulseRequest = Bool(\iInterfaceBitmap & %00000100)
            *uCapabilities\fTmcInterfaceIsTalkOnly = Bool(\iInterfaceBitmap & %00000010)
            *uCapabilities\fTmcInterfaceIsListenOnly = Bool(\iInterfaceBitmap & %00000001)
            *uCapabilities\fTmcDeviceSupportsEndingTermChar = Bool(\iDeviceBitmap & %00000001)
            ; USB488-Interface.
            *uCapabilities\i488BcdVersionNumber = \iBcdUsb488
            *uCapabilities\f488InterfaceIs4882 = Bool(\iInterface488Bitmap & %00000100)
            *uCapabilities\f488InterfaceAcceptsRemoteLocalControlRequests = Bool(\iInterface488Bitmap & %00000010)
            *uCapabilities\f488InterfaceAcceptsTriggerMessage = Bool(\iInterface488Bitmap & %00000001)
            *uCapabilities\f488DeviceUnderstandsAllMandatoryScpiCommands = Bool(\iDevice488Bitmap & %00001000)
            *uCapabilities\f488DeviceIsServiceRequestCapable = Bool(\iDevice488Bitmap & %00000100)
            *uCapabilities\f488DeviceIsRemoteLocalCapable = Bool(\iDevice488Bitmap & %00000010)
            *uCapabilities\f488DeviceIsTriggerCapable = Bool(\iDevice488Bitmap & %00000001)
         EndWith
         ; Die Eigenschaften in der Instanz speichern und nicht zurückgeben.
         If fInstance
            If LockInstance(iInstanceNumber, *sError)
               CopyStructure(*uCapabilities, mInstanceMap()\uCapabilities, uCapabilities)
            EndIf
            UnlockInstance(iInstanceNumber)
         EndIf

         fReturn = #True
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i InitiateAbortBulkOut(iInstanceNumber.i, iTransferIdentifier.i, *uData.uInitiateAbortBulkOutInResponse,
                                    *sError.String)
      ; USBTMC Device Requests.
      ; Der Host kann mit diesem Request einen Bulk-Out Transfer abbrechen und die Datenübertragung synchronisieren.
      ; Das darf er aber erst vornehmen, wenn er zuvor die Kommunikation mit dem Bulk-Out Endpoint beendet hat.
      ; Im unteren Byte von iValue muss die Nummer des Tags stehen, mit dem der abzubrechende Transfer gekennzeichnet ist.
      ; Diese Nummer steht im Header, mit dem der abzubrechende Transfer eingeleitet worden ist.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> iTransferIdentifier:  Die Tag-Nummer des abzubrechenden Transfers.
      ; <- *uData:               Die Daten der Anwort vom Gerät.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      If (Not *uData)
         *sError\s = "Der Zeiger für die Daten ist Null. *uData = 0"
         DebugView::Error("Der Zeiger für die Daten ist Null. *uData = 0", iInstanceNumber, miThreadedThreadCounter)

      Else
         eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                        LibUsb::#eLIBUSB_RECIPIENT_ENDPOINT |
                        LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
         iTransferred = SizeOf(uInitiateAbortBulkOutInResponse)
         iValue = iTransferIdentifier

         ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
         If LockInstance(iInstanceNumber, *sError)
            iIndex = mInstanceMap()\uDevice\iEndpointBulkOut
         EndIf
         UnlockInstance(iInstanceNumber)

         fResult = ControlTransfer(iInstanceNumber, eRequestType, #eInitiateAbortBulkOut, iValue, iIndex,
                                   *uData, @iTransferred, *sError)
         If (Not fResult)
            DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

         Else
            fReturn = #True
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i InitiateAbortBulkIn(iInstanceNumber.i, iTransferIdentifier.i, *uData.uInitiateAbortBulkOutInResponse,
                                   *sError.String)
      ; USBTMC Device Requests.
      ; Der Host kann mit diesem Request einen Bulk-In Transfer abbrechen, um die Datenübertragung zu synchronisieren.
      ; Bevor der Host diesen Request sendet, muss er prüfen, ob ein Bulk-Out Transfer im Gange ist, der eine Antwort erwartet.
      ; Wenn dem so ist, dann muss der Host zunächst diesen Transfer beenden.
      ; Im unteren Byte von iValue muss die Nummer des Tags stehen, mit dem der abzubrechende Transfer gekennzeichnet ist.
      ; Diese Nummer steht im Header, mit dem der abzubrechende Transfer eingeleitet worden ist.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; -> iTransferIdentifier:  Die Tag-Nummer des abzubrechenden Transfers.
      ; <- *uData:               Die Daten der Anwort vom Gerät.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      If (Not *uData)
         *sError\s = "Der Zeiger für die Daten ist Null. *uData = 0"
         DebugView::Error("Der Zeiger für die Daten ist Null. *uData = 0", iInstanceNumber, miThreadedThreadCounter)

      Else
         eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                        LibUsb::#eLIBUSB_RECIPIENT_ENDPOINT |
                        LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
         iTransferred = SizeOf(uInitiateAbortBulkOutInResponse)
         iValue = iTransferIdentifier

         ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
         If LockInstance(iInstanceNumber, *sError)
            iIndex = mInstanceMap()\uDevice\iEndpointBulkIn
         EndIf
         UnlockInstance(iInstanceNumber)

         fResult = ControlTransfer(iInstanceNumber, eRequestType, #eInitiateAbortBulkIn, iValue, iIndex,
                                   *uData, @iTransferred, *sError)
         If (Not fResult)
            DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

         Else
            fReturn = #True
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i CheckAbortBulkOutStatus(iInstanceNumber.i, *uData.uCheckAbortBulkOutStatusResponse, *sError.String)
      ; USBTMC Device Requests.
      ; Mit diesem Request überprüft der Host, welchen Status der Bulk-Out Endpoint hat, nachdem das Gerät zuvor
      ; einen INITIATE_ABORT_BULK_OUT Request bearbeitet hat
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; <- *uData:               Die Daten der Anwort vom Gerät.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      If (Not *uData)
         *sError\s = "Der Zeiger für die Daten ist Null. *uData = 0"
         DebugView::Error("Der Zeiger für die Daten ist Null. *uData = 0", iInstanceNumber, miThreadedThreadCounter)

      Else
         eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                        LibUsb::#eLIBUSB_RECIPIENT_ENDPOINT |
                        LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
         iTransferred = SizeOf(uCheckAbortBulkOutStatusResponse)
         iValue = 0  ; Reserviert.

         ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
         If LockInstance(iInstanceNumber, *sError)
            iIndex = mInstanceMap()\uDevice\iEndpointBulkOut
         EndIf
         UnlockInstance(iInstanceNumber)

         fResult = ControlTransfer(iInstanceNumber, eRequestType, #eCheckAbortBulkOutStatus, iValue, iIndex,
                                   *uData, @iTransferred, *sError)
         If (Not fResult)
            DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

         Else
            fReturn = #True
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i CheckAbortBulkInStatus(iInstanceNumber.i, *uData.uCheckAbortBulkInStatusResponse, *sError.String)
      ; USBTMC Device Requests.
      ; Mit diesem Request überprüft der Host, welchen Status der Bulk-In Endpoint hat, nachdem das Gerät zuvor
      ; einen INITIATE_ABORT_BULK_IN Request bearbeitet hat. Der Host soll diesen Request erst senden,
      ; nachdem er ein short packet vom Bulk-In Endpoint empfangen hat.
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; <- *uData:               Die Daten der Anwort vom Gerät.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      If (Not *uData)
         *sError\s = "Der Zeiger für die Daten ist Null. *uData = 0"
         DebugView::Error("Der Zeiger für die Daten ist Null. *uData = 0", iInstanceNumber, miThreadedThreadCounter)

      Else
         eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                        LibUsb::#eLIBUSB_RECIPIENT_ENDPOINT |
                        LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
         iTransferred = SizeOf(uCheckAbortBulkInStatusResponse)
         iValue = 0  ; Reserviert.

         ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
         If LockInstance(iInstanceNumber, *sError)
            iIndex = mInstanceMap()\uDevice\iEndpointBulkIn
         EndIf
         UnlockInstance(iInstanceNumber)

         fResult = ControlTransfer(iInstanceNumber, eRequestType, #eCheckAbortBulkInStatus, iValue, iIndex,
                                   *uData, @iTransferred, *sError)
         If (Not fResult)
            DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

         Else
            fReturn = #True
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i InitiateClear(iInstanceNumber.i, *eUsbTmcStatus.Integer, *sError.String)
      ; USBTMC Device Requests.
      ; Löscht die Inhalte des Bulk-Out und Bulk-In Endpoints.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *Data:             Die Daten der Anwort vom Gerät.
      ;                       Wenn die Daten nicht benötigt werden, kann hier #Null übergeben werden.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iData.a, iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                     LibUsb::#eLIBUSB_RECIPIENT_INTERFACE |
                     LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
      iTransferred = SizeOf(Ascii)
      iValue = 0  ; Reserviert.

      ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
      If LockInstance(iInstanceNumber, *sError)
         iIndex = mInstanceMap()\uDevice\iInterfaceNumber
      EndIf
      UnlockInstance(iInstanceNumber)

      fResult = ControlTransfer(iInstanceNumber, eRequestType, #eInitiateClear, iValue, iIndex,
                                @iData, @iTransferred, *sError)
      If (Not fResult)
         DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

      Else
         ; Die Daten zurückgeben.
         If *eUsbTmcStatus
            *eUsbTmcStatus\i = iData
         EndIf
         fReturn = #True
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i CheckClearStatus(iInstanceNumber.i, *uData.uCheckClearStatusResponse, *sError.String)
      ; USBTMC Device Requests.
      ; Dieser Request wird an das Gerät geschickt, nachdem ein INITIATE_CLEAR Request verarbeitet wurde,
      ; um zu überprüfen, ob alle damit verbundenen Vorgänge abgeschlossen worden sind
      ; -> iInstanceNumber:      Die Nummer der Instanz.
      ; <- *uData:               Die Daten der Anwort vom Gerät.
      ; <- sError:               Fehlerbeschreibung.
      ; Rückgabe:                True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      If (Not *uData)
         *sError\s = "Der Zeiger für die Daten ist Null. *uData = 0"
         DebugView::Error("Der Zeiger für die Daten ist Null. *uData = 0", iInstanceNumber, miThreadedThreadCounter)

      Else
         eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                        LibUsb::#eLIBUSB_RECIPIENT_INTERFACE |
                        LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
         iTransferred = SizeOf(uCheckClearStatusResponse)
         iValue = 0  ; Reserviert.

         ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
         If LockInstance(iInstanceNumber, *sError)
            iIndex = mInstanceMap()\uDevice\iInterfaceNumber
         EndIf
         UnlockInstance(iInstanceNumber)

         fResult = ControlTransfer(iInstanceNumber, eRequestType, #eCheckClearStatus, iValue, iIndex,
                                   *uData, @iTransferred, *sError)
         If (Not fResult)
            DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

         Else
            fReturn = #True
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i IndicatorPulse(iInstanceNumber.i, *eUsbTmcStatus.Integer, *sError.String)
      ; USBTMC Device Requests.
      ; Dieser Request gestattet es dem Host, am Gerät Aktivität zu signalisieren, indem eine LED für den Zeitraum
      ; von 0.5 bis 1 Sekunde (die genaue Zeit innerhalb dieses Intervalls ist dem Hersteller überlassen) eingeschaltet
      ; wird,nachdem dieser Request empfangen wurde. Danach muss sich die LED automatisch ausschalten.
      ; Voraussetzung ist selbstverständlich, dass das Gerät über eine derartige Indikator-LED verfügt und die Fähigkeit
      ; zur Unterstützung dieses Requests im CapabilitiesDescriptor eingetragen ist.
      ; Diese Eigenschaft ist grundsätzlich dazu gedacht, ein Gerät zu identifizieren, jedoch kann sie auch für andere
      ; Zwecke verwendet werden.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *eUsbTmcStatus:    Die Daten der Anwort vom Gerät.
      ;                       Wenn die Daten nicht benötigt werden, kann hier #Null übergeben werden.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iData.a, iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                     LibUsb::#eLIBUSB_RECIPIENT_INTERFACE |
                     LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
      iTransferred = SizeOf(Ascii)
      iValue = 0  ; Reserviert?

      ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
      If LockInstance(iInstanceNumber, *sError)
         iIndex = mInstanceMap()\uDevice\iInterfaceNumber
      EndIf
      UnlockInstance(iInstanceNumber)

      fResult = ControlTransfer(iInstanceNumber, eRequestType, #eIndicatorPulse, iValue, iIndex,
                                @iData, @iTransferred, *sError)
      If (Not fResult)
         DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

      Else
         ; Die Daten zurückgeben.
         If *eUsbTmcStatus
            *eUsbTmcStatus\i = iData
         EndIf
         fReturn = #True
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure

   ;-
   ;- -- USB488 Subclass Device Requests.

   Procedure.i ReadStatusByte(iInstanceNumber.i, iTag.i, *uReadStatusByteResponse.uReadStatusByteResponse, *sError.String)
      ; USB488 Subclass Device Requests.
      ; Mit diesem Request wird das Status Byte des Geräts ausgelesen, sofern die Schnittstelle nicht über einen
      ; Interrupt-In Endpoint verfügt. Wenn ein Interrupt-In Endpoint vorhanden ist, wird mit diesem Request die
      ; Übertragung des Status Byte über den Interrupt-In Endpoint veranlasst.
      ; -> iInstanceNumber:            Die Nummer der Instanz.
      ; -> iTag:                       Wert zwischen 2 und 127, der vom Host erzeugt und mit diesem Request übertragen wird.
      ;                                Der Host sollte bTag für jeden neu gesendeten Request inkrementieren.
      ;                                Der Wert von bTag wird vom Gerät zusammen mit dem Status Byte an den Host zurückgesendet.
      ;                                Damit lassen sich Anfrage und Antwort einander zuordnen, wenn der Interrupt-In Endpoint vom
      ;                                Host gelesen wird.
      ; <- *uReadStatusByteResponse:   Die Daten der Anwort vom Gerät.
      ; <- sError:                     Fehlerbeschreibung.
      ; Rückgabe:                      True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i, sError.String
      Protected iData.l, iTransferred.i, eRequestType.i, iValue.i, iIndex.i, uData.uInterruptTransferResponse

      If (Not *uReadStatusByteResponse)
         *sError\s = "Der Zeiger für die Daten ist Null. *uReadStatusByteResponse = 0"
         DebugView::Error("Der Zeiger für die Daten ist Null. *uReadStatusByteResponse = 0",
                          iInstanceNumber, miThreadedThreadCounter)

      Else
         eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                        LibUsb::#eLIBUSB_RECIPIENT_INTERFACE |
                        LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
         iTransferred = SizeOf(uReadStatusByteResponse)

         If ((iTag < 2) Or (iTag > 127))
            DebugView::Error("Der Wert für iTag ist nicht erlaubt (2 bis 127): iTag = " + iTag,
                             iInstanceNumber, miThreadedThreadCounter)
            iTag = 2
         EndIf
         iValue = iTag  ; bTag Low-Byte.

         ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
         If LockInstance(iInstanceNumber, *sError)
            iIndex = mInstanceMap()\uDevice\iInterfaceNumber
         EndIf
         UnlockInstance(iInstanceNumber)

         fResult = ControlTransfer(iInstanceNumber, eRequestType, #eReadStatusByte, iValue, iIndex,
                                   *uReadStatusByteResponse, @iTransferred, *sError)
         If (Not fResult)
            DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)
         Else
            fReturn = #True
         EndIf
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i RemoteEnableControl(iInstanceNumber.i, fEnable.i, *eUsbTmcStatus.Integer, *sError.String)
      ; USB488 Subclass Device Requests.
      ; Ren Control.
      ; Dieser Request gehört zusammen mit den beiden LocalLockout und GoToLocal zur Steuerung der lokalen, meist manuellen
      ; Gerätefunktionen, und bilden die Remote-Local Schnittstellenfunktionen.
      ; Für diesen Request muss das Gerät über die Fähigkeit RL1 verfügen. RL1 kann mit GetCapabilities abgefragt werden,
      ; und entspicht dem Feld f488InterfaceAcceptsRemoteLocalControlRequests in der zurückgegebenen Struktur.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; -> fEnable:           True für das Aktivieren des Status REN, False für das Deaktivieren.
      ; <- *eUsbTmcStatus:    Die Daten der Anwort vom Gerät.
      ;                       Wenn die Daten nicht benötigt werden, kann hier #Null übergeben werden.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iData.a, iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                     LibUsb::#eLIBUSB_RECIPIENT_INTERFACE |
                     LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
      iTransferred = SizeOf(Ascii)
      iValue = Bool(fEnable)

      ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
      If LockInstance(iInstanceNumber, *sError)
         iIndex = mInstanceMap()\uDevice\iInterfaceNumber
      EndIf
      UnlockInstance(iInstanceNumber)

      fResult = ControlTransfer(iInstanceNumber, eRequestType, #eRenControl, iValue, iIndex,
                                @iData, @iTransferred, *sError)
      If (Not fResult)
         DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

      Else
         ; Die Daten zurückgeben.
         If *eUsbTmcStatus
            *eUsbTmcStatus\i = iData
         EndIf
         fReturn = #True
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i LocalLockout(iInstanceNumber.i, *eUsbTmcStatus.Integer, *sError.String)
      ; USB488 Subclass Device Requests.
      ; Dieser Request gehört zusammen mit den beiden RemoteEnableControl und GoToLocal zur Steuerung der lokalen,
      ;  meist manuellen Gerätefunktionen, und bilden die Remote-Local Schnittstellenfunktionen.
      ; Für diesen Request muss das Gerät über die Fähigkeit RL1 verfügen. RL1 kann mit GetCapabilities abgefragt werden,
      ; und entspicht dem Feld f488InterfaceAcceptsRemoteLocalControlRequests in der zurückgegebenen Struktur.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *eUsbTmcStatus:    Die Daten der Anwort vom Gerät.
      ;                       Wenn die Daten nicht benötigt werden, kann hier #Null übergeben werden.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iData.a, iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                     LibUsb::#eLIBUSB_RECIPIENT_INTERFACE |
                     LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
      iTransferred = SizeOf(Ascii)
      iValue = 0

      ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
      If LockInstance(iInstanceNumber, *sError)
         iIndex = mInstanceMap()\uDevice\iInterfaceNumber
      EndIf
      UnlockInstance(iInstanceNumber)

      fResult = ControlTransfer(iInstanceNumber, eRequestType, #eLocalLockout, iValue, iIndex,
                                @iData, @iTransferred, *sError)
      If (Not fResult)
         DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

      Else
         ; Die Daten zurückgeben.
         If *eUsbTmcStatus
            *eUsbTmcStatus\i = iData
         EndIf
         fReturn = #True
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


   Procedure.i GoToLocal(iInstanceNumber.i, *eUsbTmcStatus.Integer, *sError.String)
      ; USB488 Subclass Device Requests.
      ; Dieser Request gehört zusammen mit den beiden RemoteEnableControl und LocalLockout zur Steuerung der lokalen,
      ;  meist manuellen Gerätefunktionen, und bilden die Remote-Local Schnittstellenfunktionen.
      ; Für diesen Request muss das Gerät über die Fähigkeit RL1 verfügen. RL1 kann mit GetCapabilities abgefragt werden,
      ; und entspicht dem Feld f488InterfaceAcceptsRemoteLocalControlRequests in der zurückgegebenen Struktur.
      ; -> iInstanceNumber:   Die Nummer der Instanz.
      ; <- *eUsbTmcStatus:    Die Daten der Anwort vom Gerät.
      ;                       Wenn die Daten nicht benötigt werden, kann hier #Null übergeben werden.
      ; <- sError:            Fehlerbeschreibung.
      ; Rückgabe:             True wenn der Request erfolgreich gesendet wurde, ansonst False.
      DebugView::ProcIn(iInstanceNumber, miThreadedThreadCounter)
      Protected fReturn.i, fResult.i
      Protected iData.a, iTransferred.i, eRequestType.i, iValue.i, iIndex.i

      eRequestType = LibUsb::#eLIBUSB_REQUEST_DEVICETOHOST |
                     LibUsb::#eLIBUSB_RECIPIENT_INTERFACE |
                     LibUsb::#eLIBUSB_REQUEST_TYPE_CLASS
      iTransferred = SizeOf(Ascii)
      iValue = 0  ; Reserviert.

      ; Fehler ignorieren, bei dem ControlTransfer wird die Instanz erneut geprüft.
      If LockInstance(iInstanceNumber, *sError)
         iIndex = mInstanceMap()\uDevice\iInterfaceNumber
      EndIf
      UnlockInstance(iInstanceNumber)

      fResult = ControlTransfer(iInstanceNumber, eRequestType, #eGoToLocal, iValue, iIndex,
                                @iData, @iTransferred, *sError)
      If (Not fResult)
         DebugView::Error(*sError\s, iInstanceNumber, miThreadedThreadCounter)

      Else
         ; Die Daten zurückgeben.
         If *eUsbTmcStatus
            *eUsbTmcStatus\i = iData
         EndIf
         fReturn = #True
      EndIf

      DebugView::ProcOut(iInstanceNumber, miThreadedThreadCounter)
      ProcedureReturn fReturn
   EndProcedure


EndModule   ; UsbTmc.
